// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Frame Rate Scalability Interface                                 ####
// ####                                                                  ####
// #### Version 1.00  --  September 02, 2000                             ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#ifndef FRAMERATESCALABILITYINTERFACE_H
#define FRAMERATESCALABILITYINTERFACE_H

#include "system.h"


/**
  * This class is an interface for frame rate scalability.
  *
  * @short   Frame Rate Scalability Interface
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class FrameRateScalabilityInterface
{
   // ====== Scalability information methods ================================
   public:
   /**
     * Get name of the frame rate scalability class.
     *
     * @return Frame rate scalability class name.
     */
   virtual const char* getFrameRateScalabilityClass() const = 0;

   /**
     * Check, if frame rate is scalable.
     *
     * @return true, if frame rate is scalable; false otherwise.
     */
   virtual bool isFrameRateScalable() const = 0;


   // ====== Frame rate methods =============================================
   /**
     * Get minimum frame rate.
     *
     * @return Minimum frame rate.
     */
   virtual double getMinFrameRate() const = 0;

   /**
     * Get maximum frame rate.
     *
     * @return Maximum frame rate.
     */
   virtual double getMaxFrameRate() const = 0;

   /**
     * Check, if given frame rate is a valid value.
     *
     * @param frameRate Frame rate to be checked.
     * @return true, if given rate is valid; false otherwise.
     */
   virtual bool isValidFrameRate(const double frameRate) const = 0;

   /**
     * Get nearest lower valid frame rate for given frame rate.
     *
     * @param rate Frame rate.
     * @return Valid frame rate nearest to given rate.
     */
   virtual double getNearestValidFrameRate(const double frameRate) const = 0;

   /**
     * Get next higher valid frame rate for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Next higher valid frame rate.
     */
   virtual double getNextFrameRateForRate(const double frameRate) const = 0;

   /**
     * Get next lower valid frame rate for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Next lower valid frame rate.
     */
   virtual double getPrevFrameRateForRate(const double frameRate) const = 0;


   // ====== Scaling and utilization methods ================================
   /**
     * Get scale factor for given frame rate:
     * (rate - MinFrameRate) / (MaxFrameRate - MinFrameRate)
     *
     * @param frameRate Frame rate.
     * @return Scale factor (out of [0,1])..
     */
   virtual double getFrameRateScaleFactorForRate(const double frameRate) const = 0;

   /**
     * Get utilization for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Utilization (out of [0,1]).
     */
   virtual double getFrameRateUtilizationForRate(const double frameRate) const = 0;
};


#endif


// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Generic Frame Rate Scalability                                   ####
// ####                                                                  ####
// #### Version 1.00  --  September 02, 2000                             ####
// ####                                                                  ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#ifndef GENERICFRAMERATESCALABILITY_H
#define GENERICFRAMERATESCALABILITY_H


#include "system.h"
// #include "frameratescalabilityinterface.h"


/**
  * This class is a generic implementation of FrameRateScalabilityInterface.
  * It provides basic functionality for subclasses. The frame rate utilization
  * here is linear.
  *
  * @short   Generic Frame Rate Scalability
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class GenericFrameRateScalability : virtual public FrameRateScalabilityInterface
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   GenericFrameRateScalability();

   /**
     * Destructor.
     */
   virtual ~GenericFrameRateScalability();


   // ====== Initialization =================================================
   /**
     * Initialize object with new minimum and maximum frame rate.
     *
     * @param minFrameRate Minimum frame rate.
     * @param maxFrameRate Maximum frame rate.
     */
   void initFrameRateScalability(const double minFrameRate,
                                 const double maxFrameRate);


   // ====== Scalability information methods ================================
   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#getFrameRateScalabilityClass
     */
   const char* getFrameRateScalabilityClass() const;

   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#isFrameRateScalable
     */
   bool isFrameRateScalable() const;


   // ====== Frame rate methods =============================================
   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#getMinFrameRate
     */
   double getMinFrameRate() const;

   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#getMaxFrameRate
     */
   double getMaxFrameRate() const;

   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#isValidFrameRate
     */
   bool isValidFrameRate(const double frameRate) const;

   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#getNearestValidFrameRate
     */
   double getNearestValidFrameRate(const double frameRate) const;

   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#getNextFrameRateForRate
     */
   double getNextFrameRateForRate(const double frameRate) const;

   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#getPrevFrameRateForRate
     */
   double getPrevFrameRateForRate(const double frameRate) const;


   // ====== Scaling and utilization methods ================================
   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#getFrameRateScaleFactorForRate
     */
   double getFrameRateScaleFactorForRate(const double frameRate) const;

   /**
     * Implementation of FrameRateScalabilityInterface.
     *
     * @see FrameRateScalabilityInterface#getFrameRateUtilizationForRate
     */
   double getFrameRateUtilizationForRate(const double frameRate) const;


   // ====== Protected data =================================================
   protected:
   double MinFrameRate;
   double MaxFrameRate;
};


#endif



// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Generic Frame Rate Scalability                                   ####
// ####                                                                  ####
// #### Version 1.00  --  September 02, 2000                             ####
// ####                                                                  ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
// #include "genericframeratescalability.h"


// ###### Constructor #######################################################
GenericFrameRateScalability::GenericFrameRateScalability()
{
   MinFrameRate = HUGE_VAL;
   MaxFrameRate = HUGE_VAL;
}


// ###### Destructor ########################################################
GenericFrameRateScalability::~GenericFrameRateScalability()
{
}


// ###### Initialize ########################################################
void GenericFrameRateScalability::initFrameRateScalability(const double minFrameRate,
                                                           const double maxFrameRate)
{
  MinFrameRate  = minFrameRate;
   if(maxFrameRate < minFrameRate) {
      MaxFrameRate = minFrameRate;
   }
   else {
      MaxFrameRate = maxFrameRate;
   }
}


// ###### Get frame rate scalability class name #############################
const char* GenericFrameRateScalability::getFrameRateScalabilityClass() const
{
   return((const char*)"Generic Frame-rate Scalability");
}


// ###### Check, if frame rate is scalable ##################################
bool GenericFrameRateScalability::isFrameRateScalable() const
{
   return(true);
}


// ###### Check, if frame rate is valid #####################################
bool GenericFrameRateScalability::isValidFrameRate(const double frameRate) const
{
   return((frameRate >= MinFrameRate) &&
          (frameRate <= MaxFrameRate) &&
          ((double)floor(frameRate) == frameRate));
}


// ###### Get nearest valid frame rate ######################################
double GenericFrameRateScalability::getNearestValidFrameRate(const double frameRate) const
{
   const double r = floor(frameRate);
   if(r < MinFrameRate) {
      return(MinFrameRate);
   }
   else if(r > MaxFrameRate) {
      return(MaxFrameRate);
   }
   return(r);
}


// ###### Get minimum frame rate ############################################
double GenericFrameRateScalability::getMinFrameRate() const
{
   return(MinFrameRate);
}


// ###### Get maximum frame rate ############################################
double GenericFrameRateScalability::getMaxFrameRate() const
{
   return(MaxFrameRate);
}


// ###### Get next frame rate ###############################################
double GenericFrameRateScalability::getNextFrameRateForRate(const double frameRate) const
{
   if(frameRate < MaxFrameRate) {
      return(frameRate + 1);
   }
   else {
      return(MaxFrameRate);
   }
}


// ###### Get previous frame rate ###########################################
double GenericFrameRateScalability::getPrevFrameRateForRate(const double frameRate) const
{
   if(frameRate > MinFrameRate) {
      return(frameRate - 1);
   }
   else {
      return(MinFrameRate);
   }
}


// ###### Get frame rate scale factor #######################################
double GenericFrameRateScalability::getFrameRateScaleFactorForRate(const double frameRate) const
{
   if((frameRate < MinFrameRate) || (frameRate > MaxFrameRate)) {
      cerr << "WARNING: GenericFrameRateScalability::getFrameRateScaleFactorForRate() - " << endl
           << "Invalid paramter " << frameRate << "!" << endl;
      return(0.0);
   }

   if(MinFrameRate != MaxFrameRate) {
      return((double)(frameRate - MinFrameRate) / (double)(MaxFrameRate - MinFrameRate));
   }
   else {
      return(1.0);
   }
}


// ###### Get frame rate utilization ########################################
double GenericFrameRateScalability::getFrameRateUtilizationForRate(const double frameRate) const
{
   return(getFrameRateScaleFactorForRate(frameRate));
}









// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Frame Size Scalability Interface                                 ####
// ####                                                                  ####
// #### Version 1.00  --  September 02, 2000                             ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#ifndef FRAMESIZESCALABILITYINTERFACE_H
#define FRAMESIZESCALABILITYINTERFACE_H


#include "system.h"


/**
  * This class is an interface for frame size scalability.
  * Important node: All frames sizes in this class are payload frame sizes!
  *
  * @short   Frame Rate Scalability Interface
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class FrameSizeScalabilityInterface
{
   // ====== Scalability information methods ================================
   public:
   /**
     * Get name of the frame size scalability class.
     *
     * @return Frame size scalability class name.
     */
   virtual const char* getFrameSizeScalabilityClass() const = 0;

   /**
     * Check, if frame size is scalable.
     *
     * @return true, if frame size is scalable; false otherwise.
     */
   virtual bool isFrameSizeScalable() const = 0;

   /**
     * Check, if frame size is variable bitrate (frame sizes are different
     * for each frame; the frame size given is the frame size necessary
     * to be reserved for a given buffer delay).
     *
     * @return true, if frame size is variable bitrate; false otherwise.
     */
   virtual bool isVariableBitrate() const = 0;


   // ====== Frame size methods =============================================
   /**
     * Get minimum payload frame size for given buffer delay (in frame rate units).
     *
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @return Minimum payload frame size.
     */
   virtual cardinal getMinPayloadFrameSizeForDelay(
                       const double   frameRate,
                       const cardinal bufferDelay) const = 0;

   /**
     * Get maximum payload frame size for given buffer delay (in frame rate units).
     *
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @return Maximum payload frame size.
     */
   virtual cardinal getMaxPayloadFrameSizeForDelay(
                       const double   frameRate,
                       const cardinal bufferDelay) const = 0;

   /**
     * Get maximum number of frames for given buffer delay (in frame rate units).
     *
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @return Maximum number of frames.
     */
   virtual cardinal getMaxFrameCountForDelay(
                       const double   frameRate,
                       const cardinal bufferDelay) const = 0;

   /**
     * Check, if given payload frame size is a valid value for given buffer delay
     * (in frame rate units).
     *
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @param frameSize Payload frame size to be checked.
     * @return true, if given size is valid; false otherwise.
     */
   virtual bool isValidPayloadFrameSize(
                   const double   frameRate,
                   const cardinal bufferDelay,
                   const cardinal frameSize) const = 0;

   /**
     * Get nearest lower valid payload frame rate for given frame rate for given buffer delay
     * (in frame rate units).
     *
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @param frameSize Payload frame size.
     * @return Valid payload frame size nearest to given size for given buffer delay.
     */
   virtual cardinal getNearestValidPayloadFrameSize(
                       const double   frameRate,
                       const cardinal bufferDelay,
                       const cardinal frameSize) const = 0;


   // ====== Scaling and utilization methods ================================
   /**
     * Get next higher valid payload frame size for given buffer delay
     * (in frame rate units) and payload frame size.
     *
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @param frameSize Payload frame size.
     * @return Next higher valid payload frame size for given buffer delay.
     */
   virtual cardinal getNextPayloadFrameSizeForDelayAndSize(
                       const double   frameRate,
                       const cardinal bufferDelay,
                       const cardinal frameSize) const = 0;

   /**
     * Get next lower valid payload frame size for given buffer delay
     * (in frame rate units) and frame size.
     *
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @param frameSize Payload frame size.
     * @return Next lower valid payload frame size for given buffer delay.
     */
   virtual cardinal getPrevPayloadFrameSizeForDelayAndSize(
                        const double   frameRate,
                        const cardinal bufferDelay,
                        const cardinal frameSize) const = 0;

   /**
     * Get scale factor for given buffer delay
     * (in frame rate units) and payload frame size:
     * (rate - MinFrameSize) / (MaxFrameRate - MinFrameSize)
     *
     * @param frameRate Frame rate.
     * @param frameSize Frame size.
     * @return Scale factor (out of [0,1])..
     */
   virtual double getPayloadFrameSizeScaleFactorForDelayAndSize(
                     const double   frameRate,
                     const cardinal bufferDelay,
                     const cardinal frameSize) const = 0;

   /**
     * Get utilization for given buffer delay
     * (in frame rate units) and payload frame size.
     *
     * @param frameRate Frame rate.
     * @param frameSize Payload frame size.
     * @return Utilization (out of [0,1])..
     */
   virtual double getPayloadFrameSizeUtilizationForDelayAndSize(
                     const double   frameRate,
                     const cardinal bufferDelay,
                     const cardinal frameSize) const = 0;


   // ====== Buffer delay methods ===========================================
   /**
     * Get maximum buffer delay. The *minimum* buffer delay is always 1.
     *
     * @return Maximum buffer delay.
     */
   virtual cardinal getMaxBufferDelay() const = 0;

   /**
     * Get next higher valid buffer delay for given buffer delay.
     *
     * @param bufferDelay Buffer delay in frame rate units.
     * @return Next higher valid buffer delay.
     */
   virtual cardinal getNextBufferDelayForDelay(const cardinal bufferDelay) const = 0;

   /**
     * Get next lower valid buffer delay for given buffer delay.
     *
     * @param bufferDelay Buffer delay in frame rate units.
     * @return Next lower valid buffer delay.
     */
   virtual cardinal getPrevBufferDelayForDelay(const cardinal bufferDelay) const = 0;
};


#endif


// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Abstract Generic Frame Size Scalability                          ####
// ####                                                                  ####
// #### Version 1.00  --  September 02, 2000                             ####
// ####                                                                  ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#ifndef ABSTRACTGENERICFRAMESIZESCALABILITY_H
#define ABSTRACTGENERICFRAMESIZESCALABILITY_H


#include "system.h"
// #include "framesizescalabilityinterface.h"


/**
  * This class is a generic implementation of FrameSizeScalabilityInterface.
  * It provides basic functionality for subclasses. The frame size utilization
  * here is linear.
  * Important node: All frames sizes in this class are payload frame sizes!
  *
  * @short   Generic Frame Size Scalability
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class AbstractGenericFrameSizeScalability : virtual public FrameSizeScalabilityInterface
{
   // ====== Frame size methods =============================================
   public:
   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#isValidPayloadFrameSize
     */
   bool isValidPayloadFrameSize(const double   frameRate,
                                const cardinal bufferDelay,
                                const cardinal frameSize) const;

   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getNearestValidPayloadFrameSize
     */
   cardinal getNearestValidPayloadFrameSize(const double   frameRate,
                                            const cardinal bufferDelay,
                                            const cardinal frameSize) const;

   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getNextFrameSizeForDelayAndSize
     */
   cardinal getNextPayloadFrameSizeForDelayAndSize(const double   frameRate,
                                                   const cardinal bufferDelay,
                                                   const cardinal frameSize) const;
   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getPrevFrameSizeForDelayAndSize
     */
   cardinal getPrevPayloadFrameSizeForDelayAndSize(const double   frameRate,
                                                   const cardinal bufferDelay,
                                                   const cardinal frameSize) const;


   // ====== Scaling and utilization methods ================================
   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getPayloadFrameSizeScaleFactorForDelayAndSize
     */
   double getPayloadFrameSizeScaleFactorForDelayAndSize(const double   frameRate,
                                                        const cardinal bufferDelay,
                                                        const cardinal frameSize) const;

   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getPayloadFrameSizeUtilizationForDelayAndSize
     */
   double getPayloadFrameSizeUtilizationForDelayAndSize(const double   frameRate,
                                                        const cardinal bufferDelay,
                                                        const cardinal frameSize) const;


   // ====== Buffer delay methods ===========================================
   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getNextBufferDelayForDelay
     */
   cardinal getNextBufferDelayForDelay(const cardinal bufferDelay) const;

   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getPrevBufferDelayForDelay
     */
   cardinal getPrevBufferDelayForDelay(const cardinal bufferDelay) const;
};


#endif



// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Abstract Generic Frame Size Scalability                          ####
// ####                                                                  ####
// #### Version 1.00  --  September 02, 2000                             ####
// ####                                                                  ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
// #include "abstractgenericframesizescalability.h"


// ###### Check, if frame size is valid #####################################
bool AbstractGenericFrameSizeScalability::isValidPayloadFrameSize(
                                             const double   frameRate,
                                             const cardinal bufferDelay,
                                             const cardinal frameSize) const
{
   return((frameSize >= getMinPayloadFrameSizeForDelay(frameRate,bufferDelay)) &&
          (frameSize <= getMaxPayloadFrameSizeForDelay(frameRate,bufferDelay)));
}


// ###### Get nearest valid frame size ######################################
cardinal AbstractGenericFrameSizeScalability::getNearestValidPayloadFrameSize(
                                                 const double   frameRate,
                                                 const cardinal bufferDelay,
                                                 const cardinal frameSize) const
{
   const cardinal minFrameSize = getMinPayloadFrameSizeForDelay(frameRate,bufferDelay);
   const cardinal maxFrameSize = getMaxPayloadFrameSizeForDelay(frameRate,bufferDelay);

   if(frameSize < minFrameSize) {
      return(minFrameSize);
   }
   else if(frameSize > maxFrameSize) {
      return(maxFrameSize);
   }
   return(frameSize);
}


// ###### Get next frame size ###############################################
cardinal AbstractGenericFrameSizeScalability::getNextPayloadFrameSizeForDelayAndSize(
                                                 const double   frameRate,
                                                 const cardinal bufferDelay,
                                                 const cardinal frameSize) const
{
   const cardinal maxFrameSize = getMaxPayloadFrameSizeForDelay(frameRate,bufferDelay);

   if(frameSize < maxFrameSize) {
      return(frameSize + 1);
   }
   else {
      return(maxFrameSize);
   }
}


// ###### Get previous frame size ###########################################
cardinal AbstractGenericFrameSizeScalability::getPrevPayloadFrameSizeForDelayAndSize(
                                                 const double   frameRate,
                                                 const cardinal bufferDelay,
                                                 const cardinal frameSize) const
{
   const cardinal minFrameSize = getMinPayloadFrameSizeForDelay(frameRate,bufferDelay);

   if(frameSize > minFrameSize) {
      return(frameSize - 1);
   }
   else {
      return(minFrameSize);
   }
}


// ###### Get frame size scale factor #######################################
double AbstractGenericFrameSizeScalability::getPayloadFrameSizeScaleFactorForDelayAndSize(
                                               const double   frameRate,
                                               const cardinal bufferDelay,
                                               const cardinal frameSize) const
{
   const cardinal minFrameSize = getMinPayloadFrameSizeForDelay(frameRate,bufferDelay);
   const cardinal maxFrameSize = getMaxPayloadFrameSizeForDelay(frameRate,bufferDelay);

   if(frameSize < minFrameSize) {
      cerr << "WARNING: GenericConstantBitrateFrameSizeScalability::getFrameSizeScaleFactorForDelayAndSize() - " << endl
           << "Invalid paramter " << frameSize << "!" << endl;
      return(0.0);
   }

   if(minFrameSize != maxFrameSize) {
      return((double)(min(frameSize,maxFrameSize) - minFrameSize) /
                (double)(maxFrameSize - minFrameSize));
   }
   else {
      return(1.0);
   }
}


// ###### Get frame size utilization ########################################
double AbstractGenericFrameSizeScalability::getPayloadFrameSizeUtilizationForDelayAndSize(
                                               const double   frameRate,
                                               const cardinal bufferDelay,
                                               const cardinal frameSize) const
{
   return(getPayloadFrameSizeScaleFactorForDelayAndSize(frameRate,bufferDelay,frameSize));
}


// ###### Get next buffer delay #############################################
cardinal AbstractGenericFrameSizeScalability::getNextBufferDelayForDelay(
                                                 const cardinal bufferDelay) const
{
   if(bufferDelay < getMaxBufferDelay()) {
      return(bufferDelay + 1);
   }
   return(bufferDelay);
}


// ###### Get previous buffer delay #########################################
cardinal AbstractGenericFrameSizeScalability::getPrevBufferDelayForDelay(
                                                 const cardinal bufferDelay) const
{
   if(bufferDelay > 1) {
      return(bufferDelay - 1);
   }
   return(1);
}





















// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Generic Constant Bitrate Frame Size Scalability                  ####
// ####                                                                  ####
// #### Version 1.00  --  September 02, 2000                             ####
// ####                                                                  ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#ifndef GENERICFRAMESIZESCALABILITY_H
#define GENERICFRAMESIZESCALABILITY_H


#include "system.h"
// #include "framesizescalabilityinterface.h"


/**
  * This class is a generic implementation of FrameSizeScalabilityInterface.
  * It provides basic functionality for subclasses. The frame size utilization
  * here is linear.
  * Important node: All frames sizes in this class are payload frame sizes!
  *
  * @short   Generic Frame Size Scalability
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class GenericConstantBitrateFrameSizeScalability
   : public AbstractGenericFrameSizeScalability
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   GenericConstantBitrateFrameSizeScalability();

   /**
     * Desstructor.
     */
   ~GenericConstantBitrateFrameSizeScalability();


   // ====== Initialization =================================================
   /**
     * Initialize object with new maximum payload frame size and scale factor.
     * MinFrameSize = scaleFactor * MaxFrameSize.
     *
     * @param maxFrameSize Maximum payload frame size.
     * @param scaleFactor Scale factor.
     */
   void initConstantBitratePayloadFrameSizeScalability(
           const cardinal maxFrameSize,
           const double   scaleFactor);


   // ====== Scalability information methods ================================
   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getFrameSizeScalabilityClass
     */
   const char* getFrameSizeScalabilityClass() const;

   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#isFrameSizeScalable
     */
   bool isFrameSizeScalable() const;

   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#isVariableBitRate
     */
   bool isVariableBitrate() const;


   // ====== Frame size methods =============================================
   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getMinPayloadFrameSizeForDelay
     */
   cardinal getMinPayloadFrameSizeForDelay(const double   frameRate,
                                           const cardinal bufferDelay) const;

   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getMaxPayloadFrameSizeForDelay
     */
   cardinal getMaxPayloadFrameSizeForDelay(const double   frameRate,
                                           const cardinal bufferDelay) const;

   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getMaxFrameCountForDelay
     */
   cardinal getMaxFrameCountForDelay(const double   frameRate,
                                     const cardinal bufferDelay) const;


   // ====== Buffer delay methods ===========================================
   /**
     * Implementation of FrameSizeScalabilityInterface.
     *
     * @see FrameSizeScalabilityInterface#getMaxBufferDelay
     */
   cardinal getMaxBufferDelay() const;


   // ====== Protected data =================================================
   protected:
   cardinal MinFrameSize;
   cardinal MaxFrameSize;
};


#endif



// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Generic Constant Bitrate Frame Size Scalability                  ####
// ####                                                                  ####
// #### Version 1.00  --  September 02, 2000                             ####
// ####                                                                  ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
// #include "genericframesizescalability.h"


// ###### Constructor #######################################################
GenericConstantBitrateFrameSizeScalability::GenericConstantBitrateFrameSizeScalability()
{
   MinFrameSize = (cardinal)-1;
   MaxFrameSize = (cardinal)-1;
}


// ###### Desstructor #######################################################
GenericConstantBitrateFrameSizeScalability::~GenericConstantBitrateFrameSizeScalability()
{
}


// ###### Initialize ########################################################
void GenericConstantBitrateFrameSizeScalability::initConstantBitratePayloadFrameSizeScalability(
                                                    const cardinal maxFrameSize,
                                                    const double   scaleFactor)
{
   if((scaleFactor < 0.0) || (scaleFactor > 1.0)) {
      cerr << "ERROR: GenericConstantBitrateFrameSizeScalability::initConstantBitratePayloadFrameSizeScalability() - " << endl
           << "Invalid parameter " << scaleFactor << "!" << endl;
      exit(1);
   }

   MinFrameSize  = (cardinal)ceil(scaleFactor * (double)maxFrameSize);
   MaxFrameSize = maxFrameSize;
}


// ###### Get frame size scalability class name #############################
const char* GenericConstantBitrateFrameSizeScalability::getFrameSizeScalabilityClass() const
{
   return((const char*)"Generic Constant Bitrate Frame-size Scalability");
}


// ###### Check, if frame size is scalable ##################################
bool GenericConstantBitrateFrameSizeScalability::isFrameSizeScalable() const
{
   return(true);
}


// ###### Check, if frame size is variable bit rate #########################
bool GenericConstantBitrateFrameSizeScalability::isVariableBitrate() const
{
   return(false);
}


// ###### Get minimum frame size ############################################
cardinal GenericConstantBitrateFrameSizeScalability::getMinPayloadFrameSizeForDelay(
                                         const double   frameRate,
                                         const cardinal bufferDelay) const
{
   return(MinFrameSize);
}


// ###### Get maxmum frame size ############################################
cardinal GenericConstantBitrateFrameSizeScalability::getMaxPayloadFrameSizeForDelay(
                                         const double   frameRate,
                                         const cardinal bufferDelay) const
{
   return(MaxFrameSize);
}


// ###### Get maximum frame count ###########################################
cardinal GenericConstantBitrateFrameSizeScalability::getMaxFrameCountForDelay(
                                         const double   frameRate,
                                         const cardinal bufferDelay) const
{
   return(bufferDelay);
}


// ###### Get maximum buffer delay ##########################################
cardinal GenericConstantBitrateFrameSizeScalability::getMaxBufferDelay() const
{
   return(1);
}







template<const cardinal a, const cardinal b> class GenericVariableBitrateFrameSizeScalability
   : public AbstractGenericFrameSizeScalability
{
   protected:
   int array1[a];
   int array2[b];
};

/*
// ###### Get next buffer delay #############################################
cardinal GenericConstantBitrateFrameSizeScalability::getNextBufferDelayForDelay(
                                         const cardinal bufferDelay) const
{
   return(1);
}


// ###### Get previous buffer delay #########################################
cardinal GenericConstantBitrateFrameSizeScalability::getPrevBufferDelayForDelay(
                                         const cardinal bufferDelay) const
{
   return(1);
}
*/








#define DEBUG








/*
  * This class contains the layer's QoS requirements.
  * Important node: All frames sizes in this class are *raw* frame sizes, the
  * frames sizes in FrameSizeScalability are payload frame sizes. This class
  * does necessary translation.
  *
  * @short   Abstract Layer Description.
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
*/
class AbstractLayerDescription : virtual public FrameSizeScalabilityInterface
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   AbstractLayerDescription();

   /**
     * Destructor.
     */
   virtual ~AbstractLayerDescription();

   /**
     * Initialize layer description.
     *
     * @param pktHeaderLength Packet header length, e.g. 40 + 8 + 12 (IPv6 + UDP + RTP).
     * @param maxPayloadLength Maximum packet payload length, e.g. 1500.
     * @param bandwidthLimit Bandwidth limit.
     * @param frameSizeWeight Frame size weight for utilization calculation.
     * @param maxTransferDelay Maximum transfer delay in microseconds.
     * @param maxBufferDelay Maximum buffer delay in frame rate units.
     * @param maxLossRate Maximum loss rate (out of [0,1]).
     * @param maxJitter Maximum jitter in microseconds.
     */
   inline void initLayer(const cardinal pktHeaderLength,
                         const cardinal maxPayloadLength,
                         const cardinal bandwidthLimit,
                         const double   frameSizeWeight,
                         const double   maxTransferDelay,
                         const cardinal maxBufferDelay,
                         const double   maxLossRate,
                         const double   maxJitter);


   // ====== Bandwidth and bandwidth quality methods ========================
   /**
     * Get bandwidth.
     *
     * @return Bandwidth.
     */
   inline cardinal getBandwidth() const;

   /**
     * Set bandwidth.
     *
     * @param frameRate Frame rate.
     * @param bandwidth Bandwidth.
     * @return true, if bandwidth is sufficient for minimum requirement.
     */
   inline bool setBandwidth(const double   frameRate,
                            const cardinal bandwidth);

   /**
     * Get bandwidth limit.
     *
     * @return Bandwidth limit.
     */
   inline cardinal getBandwidthLimit() const;

   /**
     * Get packet rate.
     *
     * @param frameRate Frame rate.
     * @return Bandwidth limit.
     */
   virtual cardinal getPacketRate(const double frameRate) const;

   /**
     * Translate frame size into bandwidth.
     *
     * @param frameRate Frame rate.
     * @param frameSize Frame size.
     */
   static inline cardinal frameSizeToBandwidth(const double   frameRate,
                                               const cardinal frameSize);

   /**
     * Translate bandwidth into frame size.
     *
     * @param frameRate Frame rate.
     * @param bandwidth Bandwidth.
     */
   static inline cardinal bandwidthToFrameSize(const double   frameRate,
                                               const cardinal bandwidth);

   /**
     * Get packets per second for given frame size.
     *
     * @param frameRate Frame rate.
     * @param frameSize Frame size.
     * @return Packets per second.
     */
   virtual cardinal frameSizeToPacketRate(const double   frameRate,
                                          const cardinal frameSize) const;

   /**
     * Get maximum transfer delay.
     *
     * @return Maximum transfer delay in microseconds.
     */
   inline double getMaxTransferDelay() const;

   /**
     * Get maximum loss rate.
     *
     * @return Maximum loss rate (out of [0,1]).
     */
   inline double getMaxLossRate() const;

   /**
     * Get maximum jitter.
     *
     * @return Maximum jitter in microseconds.
     */
   inline double getMaxJitter() const;


   // ====== Frame size methods =============================================
   /**
     * Get frame size utilization weight.
     *
     * @return Weight.
     */
   inline double getFrameSizeWeight() const;

   /**
     * Check, if given frame size is valid for given frame rate and buffer delay.
     *
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay.
     * @param size FrameSize.
     * @return true, if frame size is valid; false otherwise.
     */
   inline bool isValidFrameSize(const double   frameRate,
                                const cardinal bufferDelay,
                                const cardinal size) const;

   /**
     * Get nearest lower frame size for given frame rate and buffer delay.
     *
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay.
     * @param size FrameSize.
     * @return Nearest lower frame size.
     */
   inline cardinal getNearestValidFrameSize(const double   frameRate,
                                            const cardinal bufferDelay,
                                            const cardinal size) const;

   /**
     * Translate payload frame size into raw frame size.
     *
     * @param frameRate Frame rate.
     * @param payload Payload frame size.
     * @return Raw frame size.
     */
   virtual cardinal payloadToRaw(const double   frameRate,
                                 const cardinal payload) const;

   /**
     * Translate raw frame size into payload frame size.
     *
     * @param frameRate Frame rate.
     * @param raw Raw frame size.
     * @return Payload frame size.
     */
   virtual cardinal rawToPayload(const double   frameRate,
                                 const cardinal raw) const;


   // ====== Scaling and utilization methods, for current buffer delay ======
   /**
     * Get minimum frame size for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Minimum frame size.
     */
   inline cardinal getMinFrameSize(const double frameRate) const;

   /**
     * Get maximum frame size for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Maximum frame size.
     */
   inline cardinal getMaxFrameSize(const double frameRate) const;

   /**
     * Get peak frame size for given frame rate and size.
     *
     * @param frameRate Frame rate.
     * @return Peak frame size.
     */
   inline cardinal getPeakFrameSizeForSize(const double   frameRate,
                                           const cardinal frameSize) const;

   /**
     * Get number of packets (upper limit) for given frame rate and size.
     *
     * @param frameRate Frame rate.
     * @return Number of packets.
     */
   inline cardinal getPacketCountForSize(const double   frameRate,
                                         const cardinal frameSize) const;

   /**
     * Get next lower frame size for given frame rate and size and size.
     *
     * @param frameRate Frame rate.
     * @param frameSize Frame size.
     * @return Next lower frame size.
     */
   inline double getPrevFrameSizeForSize(const double   frameRate,
                                         const cardinal frameSize) const;

   /**
     * Get next higher frame size for given frame rate and size.
     *
     * @param frameRate Frame rate and size.
     * @param frameSize Frame size.
     * @return Next higher frame size.
     */
   inline double getNextFrameSizeForSize(const double   frameRate,
                                         const cardinal frameSize) const;

   /**
     * Get frame size scale factor for given frame rate and size:
     * (size - MinFrameSize) / (MaxFrameSize - MinFrameSize).
     *
     * @param frameRate Frame rate and size.
     * @param frameSize Frame size.
     * @return Scale factor (out of [0,1]).
     */
   inline double getFrameSizeScaleFactorForSize(const double   frameRate,
                                                const cardinal frameSize) const;

   /**
     * Get frame size utilization for given frame rate and size.
     *
     * @param frameRate Frame rate and size.
     * @param frameSize Frame size.
     * @return Utilization (out of [0,1]).
     */
   inline double getFrameSizeUtilizationForSize(const double   frameRate,
                                                const cardinal frameSize) const;


   // ====== Scaling and utilization methods, for given buffer delay ========
   /**
     * Get minimum frame size for given frame rate and buffer delay.
     *
     * @param frameRate Frame rate.
     * @return Minimum frame size.
     */
   inline cardinal getMinFrameSizeForDelay(const double   frameRate,
                                           const cardinal bufferDelay) const;

   /**
     * Get maximum frame size for given frame rate and buffer delay.
     *
     * @param frameRate Frame rate.
     * @return Maximum frame size.
     */
   inline cardinal getMaxFrameSizeForDelay(const double   frameRate,
                                           const cardinal bufferDelay) const;

   /**
     * Get peak frame size for given frame rate, size and buffer delay.
     *
     * @param frameRate Frame rate.
     * @return Peak frame size.
     */
   inline cardinal getPeakFrameSizeForDelayAndSize(const double   frameRate,
                                                   const cardinal bufferDelay,
                                                   const cardinal frameSize) const;

   /**
     * Get number of packets (upper limit) for given frame rate, size and buffer delay.
     *
     * @param frameRate Frame rate.
     * @return Number of packets.
     */
   inline cardinal getPacketCountForDelayAndSize(const double   frameRate,
                                                 const cardinal bufferDelay,
                                                 const cardinal frameSize) const;

   /**
     * Get next lower frame size for given frame rate, size and buffer delay.
     *
     * @param frameRate Frame rate.
     * @param frameSize Frame size.
     * @return Next lower frame size.
     */
   inline double getPrevFrameSizeForDelayAndSize(const double   frameRate,
                                                 const cardinal bufferDelay,
                                                 const cardinal frameSize) const;
   /**
     * Get next higher frame size for given frame rate, size and buffer delay.
     *
     * @param frameRate Frame rate.
     * @param frameSize Frame size.
     * @return Next higher frame size.
     */
   inline double getNextFrameSizeForDelayAndSize(const double   frameRate,
                                                 const cardinal bufferDelay,
                                                 const cardinal frameSize) const;

   /**
     * Get frame size scale factor for given frame rate, size and buffer delay.
     * (size - MinFrameSize) / (MaxFrameSize - MinFrameSize).
     *
     * @param frameRate Frame rate.
     * @param frameSize Frame size.
     * @return Scale factor (out of [0,1]).
     */
   inline double getFrameSizeScaleFactorForDelayAndSize(const double   frameRate,
                                                        const cardinal bufferDelay,
                                                        const cardinal frameSize) const;

   /**
     * Get frame size utilization for given frame rate, size and buffer delay.
     *
     * @param frameRate Frame rate.
     * @param frameSize Frame size.
     * @return Utilization (out of [0,1]).
     */
   inline double getFrameSizeUtilizationForDelayAndSize(const double   frameRate,
                                                        const cardinal bufferDelay,
                                                        const cardinal frameSize) const;


   // ====== Buffer delay methods ===========================================
   /**
     * Get buffer delay.
     *
     * @return Buffer delay in frame rate units.
     */
   inline cardinal getBufferDelay() const;

   /**
     * Set buffer delay.
     *
     * @param bufferDelay Buffer delay in frame rate units.
     * @return Buffer delay set in frame rate units.
     */
   inline cardinal setBufferDelay(const cardinal bufferDelay);

   /**
     * Get next lower buffer delay.
     *
     * @return Buffer delay in frame rate units.
     */
   inline cardinal getPrevBufferDelay() const;

   /**
     * Get next higher buffer delay.
     *
     * @return Buffer delay in frame rate units.
     */
   inline cardinal getNextBufferDelay() const;



   // ====== Protected data =================================================
   protected:
   cardinal PktHeaderSize;
   cardinal PktMaxSize;

   cardinal Bandwidth;
   cardinal BandwidthLimit;

   double   MaxTransferDelay;
   double   MaxLossRate;
   double   MaxJitter;

   cardinal BufferDelay;
   cardinal MaxBufferDelay;
   double   FrameSizeWeight;
};


// ###### Constructor #######################################################
AbstractLayerDescription::AbstractLayerDescription()
{
   PktHeaderSize        = 28;
   PktMaxSize           = 1500;
   Bandwidth            = 0;
   BandwidthLimit       = (cardinal)-1;
   FrameSizeWeight      = 0.0;
   MaxTransferDelay     = HUGE_VAL;
   MaxLossRate          = HUGE_VAL;
   MaxJitter            = HUGE_VAL;
}


// ###### Desstructor #######################################################
AbstractLayerDescription::~AbstractLayerDescription()
{
}


// ###### Initialization ####################################################
inline void AbstractLayerDescription::initLayer(const cardinal pktHeaderSize,
                                                const cardinal pktMaxSize,
                                                const cardinal bandwidthLimit,
                                                const double   frameSizeWeight,
                                                const double   maxTransferDelay,
                                                const cardinal maxBufferDelay,
                                                const double   maxLossRate,
                                                const double   maxJitter)
{
   PktHeaderSize    = pktHeaderSize;
   PktMaxSize       = pktMaxSize;
   MaxTransferDelay = maxTransferDelay;
   MaxBufferDelay   = maxBufferDelay;
   MaxLossRate      = maxLossRate;
   MaxJitter        = maxJitter;
   BufferDelay      = 1;
   BandwidthLimit   = bandwidthLimit;
   FrameSizeWeight  = frameSizeWeight;
}


// ##########################################################################
// #### Bandwidth and bandwidth quality methods                          ####
// ##########################################################################


// ###### Get packet rate ###################################################
cardinal AbstractLayerDescription::getPacketRate(const double frameRate) const
{
   return((cardinal)ceil(frameRate) *
             getPacketCountForSize(
                frameRate,
                (cardinal)floor((double)Bandwidth / frameRate)));
}


// ###### Get bandwidth #####################################################
inline cardinal AbstractLayerDescription::getBandwidth() const
{
   return(Bandwidth);
}


// ###### Get bandwidth limit ###############################################
inline cardinal AbstractLayerDescription::getBandwidthLimit() const
{
   return(BandwidthLimit);
}


// ###### Set bandwidth #####################################################
inline bool AbstractLayerDescription::setBandwidth(const double   frameRate,
                                                   const cardinal bandwidth)
{
   Bandwidth = bandwidth;
   if(getMinFrameSize(frameRate) > bandwidth) {
      return(false);
   }
   return(true);
}


// ###### Get maximum transfer delay ########################################
inline double AbstractLayerDescription::getMaxTransferDelay() const
{
   return(MaxTransferDelay);
}


// ###### Get maximum loss rate #############################################
inline double AbstractLayerDescription::getMaxLossRate() const
{
   return(MaxLossRate);
}


// ###### Get maximum jitter ################################################
inline double AbstractLayerDescription::getMaxJitter() const
{
   return(MaxJitter);
}



// ##########################################################################
// #### Frame size methods                                               ####
// ##########################################################################


// ###### Get frame size weight #############################################
inline double AbstractLayerDescription::getFrameSizeWeight() const
{
   return(FrameSizeWeight);
}


// ###### Check, if frame size is valid #####################################
inline bool AbstractLayerDescription::isValidFrameSize(
                                        const double   frameRate,
                                        const cardinal bufferDelay,
                                        const cardinal frameSize) const
{
   return(isValidPayloadFrameSize(frameRate,bufferDelay,rawToPayload(frameRate,frameSize)));
}


// ###### Get nearest lower frame size ######################################
inline cardinal AbstractLayerDescription::getNearestValidFrameSize(
                                             const double   frameRate,
                                             const cardinal bufferDelay,
                                             const cardinal frameSize) const
{
   return(payloadToRaw(frameRate,getNearestValidPayloadFrameSize(frameRate,bufferDelay,rawToPayload(frameRate,frameSize))));
}


// ###### Translate raw into payload frame size #############################
cardinal AbstractLayerDescription::rawToPayload(const double   frameRate,
                                                const cardinal raw) const
{
   const cardinal maxFrameCount   = getMaxFrameCountForDelay(frameRate,BufferDelay);
   const cardinal maxPktPayload   = PktMaxSize - PktHeaderSize;
   const cardinal maxFramePackets = (maxFrameCount > 1) ? maxFrameCount - 1 : 0;
   const cardinal frameSize       = raw - (maxFramePackets * PktHeaderSize);
   const cardinal payloadPackets  = (cardinal)ceil((double)frameSize /
                                                      (double)PktMaxSize);

   const cardinal payload = frameSize - (payloadPackets * PktHeaderSize);

/*
   if(raw != payloadToRaw(payload)) {
      cerr << "ERROR: Raw to Payload!" << endl;
      exit(1);
   }
*/
   return(payload);
}


// ###### Translate payload into raw frame size #############################
inline cardinal AbstractLayerDescription::payloadToRaw(
                                             const double   frameRate,
                                             const cardinal payload) const
{
   const cardinal maxFrameCount   = getMaxFrameCountForDelay(frameRate,BufferDelay);
   const cardinal maxPktPayload   = PktMaxSize - PktHeaderSize;
   const cardinal maxFramePackets = (maxFrameCount > 1) ? maxFrameCount - 1 : 0;
   const cardinal payloadPackets  = (cardinal)ceil((double)payload /
                                                      (double)maxPktPayload);

   const cardinal raw = (maxFramePackets + payloadPackets) * PktHeaderSize + payload;

/*
   if(payload != rawToPayload(raw)) {
      cerr << "ERROR: Payload to Raw!" << endl;
      exit(1);
   }
*/
   return(raw);
}


// ###### Translate frame size into bandwidth ###############################
inline cardinal AbstractLayerDescription::frameSizeToBandwidth(
                                             const double   frameRate,
                                             const cardinal frameSize)
{
   return((cardinal)ceil((double)frameSize * frameRate));
}



// ###### Translate bandwidth into frame size ###############################
inline cardinal AbstractLayerDescription::bandwidthToFrameSize(
                                             const double   frameRate,
                                             const cardinal bandwidth)
{
   return((cardinal)floor((double)bandwidth / frameRate));
}


// ###### Calculate packet rate for frame size ##############################
cardinal AbstractLayerDescription::frameSizeToPacketRate(
                                      const double   frameRate,
                                      const cardinal frameSize) const
{
   return((cardinal)ceil((double)getPacketCountForSize(
                                    frameRate, frameSize) * frameRate));
}



// ##########################################################################
// #### Scaling and utilization methods, for current buffer delay        ####
// ##########################################################################



// ###### Get minimum frame size ############################################
inline cardinal AbstractLayerDescription::getMinFrameSize(
                                             const double frameRate) const
{
   return(payloadToRaw(frameRate,getMinPayloadFrameSizeForDelay(frameRate,BufferDelay)));
}


// ###### Get maximum frame size ############################################
inline cardinal AbstractLayerDescription::getMaxFrameSize(
                                             const double frameRate) const
{
   return(payloadToRaw(frameRate,getMaxPayloadFrameSizeForDelay(frameRate,BufferDelay)));
}


// ###### Get peak frame size ###############################################
inline cardinal AbstractLayerDescription::getPeakFrameSizeForSize(
                                             const double   frameRate,
                                             const cardinal frameSize) const
{
   return(getPeakFrameSizeForDelayAndSize(frameRate,BufferDelay,frameSize));
}


// ###### Get packet count ##################################################
inline cardinal AbstractLayerDescription::getPacketCountForSize(
                                             const double   frameRate,
                                             const cardinal frameSize) const
{
   return(getPacketCountForDelayAndSize(frameRate,BufferDelay,frameSize));
}


// ###### Get next lower frame size #########################################
inline double AbstractLayerDescription::getPrevFrameSizeForSize(
                                           const double   frameRate,
                                           const cardinal frameSize) const
{
   return(getPrevFrameSizeForDelayAndSize(frameRate,BufferDelay,frameSize));
}


// ###### Get next higher frame size ########################################
inline double AbstractLayerDescription::getNextFrameSizeForSize(
                                           const double   frameRate,
                                           const cardinal frameSize) const
{
   return(getNextFrameSizeForDelayAndSize(frameRate,BufferDelay,frameSize));
}


// ###### Get frame size scale factor #######################################
inline double AbstractLayerDescription::getFrameSizeScaleFactorForSize(
                                           const double   frameRate,
                                           const cardinal frameSize) const
{
   return(getFrameSizeScaleFactorForDelayAndSize(frameRate,BufferDelay,frameSize));
}


// ###### Get frame size utilization ########################################
inline double AbstractLayerDescription::getFrameSizeUtilizationForSize(
                                           const double   frameRate,
                                           const cardinal frameSize) const
{
   return(getFrameSizeUtilizationForDelayAndSize(frameRate,BufferDelay,frameSize));
}



// ##########################################################################
// #### Scaling and utilization methods, for given buffer delay          ####
// ##########################################################################



// ###### Get minimum frame size ############################################
inline cardinal AbstractLayerDescription::getMinFrameSizeForDelay(
                                             const double   frameRate,
                                             const cardinal bufferDelay) const
{
   return(payloadToRaw(frameRate,getMinPayloadFrameSizeForDelay(frameRate,bufferDelay)));
}


// ###### Get maximum frame size ############################################
inline cardinal AbstractLayerDescription::getMaxFrameSizeForDelay(
                                             const double   frameRate,
                                             const cardinal bufferDelay) const
{
   return(payloadToRaw(frameRate,getMaxPayloadFrameSizeForDelay(frameRate,bufferDelay)));
}


// ###### Get peak frame size ###############################################
inline cardinal AbstractLayerDescription::getPeakFrameSizeForDelayAndSize(
                                             const double   frameRate,
                                             const cardinal bufferDelay,
                                             const cardinal size) const
{
   const double factor = getFrameSizeScaleFactorForDelayAndSize(frameRate,bufferDelay,size);
   const cardinal diff = getMaxFrameSizeForDelay(frameRate,1) - getMinFrameSizeForDelay(frameRate,1);

   return(getMinFrameSizeForDelay(frameRate,1) +
             (cardinal)ceil((double)diff * factor));
}


// ###### Get packet count ##################################################
inline cardinal AbstractLayerDescription::getPacketCountForDelayAndSize(
                                             const double   frameRate,
                                             const cardinal bufferDelay,
                                             const cardinal size) const
{
   const cardinal maxFrameCount   = getMaxFrameCountForDelay(frameRate,bufferDelay);
   const cardinal maxPktPayload   = PktMaxSize - PktHeaderSize;
   const cardinal maxFramePackets = (maxFrameCount > 1) ? maxFrameCount - 1 : 0;
   const cardinal payloadPackets  = (cardinal)ceil((double)rawToPayload(frameRate,size) /
                                                      (double)maxPktPayload);
   return(maxFramePackets + payloadPackets);
}



// ###### Get next higher frame size ########################################
inline double AbstractLayerDescription::getPrevFrameSizeForDelayAndSize(
                                           const double   frameRate,
                                           const cardinal bufferDelay,
                                           const cardinal size) const
{
   return(payloadToRaw(frameRate,
             getPrevPayloadFrameSizeForDelayAndSize(frameRate,bufferDelay,
                rawToPayload(frameRate,size))));
}


// ###### Get next lower frame size #########################################
inline double AbstractLayerDescription::getNextFrameSizeForDelayAndSize(
                                           const double   frameRate,
                                           const cardinal bufferDelay,
                                           const cardinal size) const
{
   return(payloadToRaw(frameRate,
             getNextPayloadFrameSizeForDelayAndSize(frameRate,bufferDelay,
                rawToPayload(frameRate,size))));
}


// ###### Get frame size scale factor #######################################
inline double AbstractLayerDescription::getFrameSizeScaleFactorForDelayAndSize(
                                           const double   frameRate,
                                           const cardinal bufferDelay,
                                           const cardinal size) const
{
   return(getPayloadFrameSizeScaleFactorForDelayAndSize(
             frameRate, bufferDelay, rawToPayload(frameRate,size)));
}


// ###### Get frame size utilization ########################################
inline double AbstractLayerDescription::getFrameSizeUtilizationForDelayAndSize(
                                           const double   frameRate,
                                           const cardinal bufferDelay,
                                           const cardinal size) const
{
   return(getPayloadFrameSizeUtilizationForDelayAndSize(
             frameRate, bufferDelay, rawToPayload(frameRate,size)));
}



// ##########################################################################
// #### Buffer delay methods                                             ####
// ##########################################################################



// ###### Get buffer delay ##################################################
inline cardinal AbstractLayerDescription::getBufferDelay() const
{
   return(BufferDelay);
}


// ###### Set buffer delay ##################################################
cardinal AbstractLayerDescription::setBufferDelay(const cardinal bufferDelay)
{
   if(BufferDelay > getMaxBufferDelay()) {
      BufferDelay = getMaxBufferDelay();
   }
   else if(BufferDelay < 1) {
      BufferDelay = 1;
   }
   else {
      BufferDelay = bufferDelay;
   }
   return(BufferDelay);
}


// ###### Get next higher buffer delay ######################################
inline cardinal AbstractLayerDescription::getNextBufferDelay() const
{
   return(getNextBufferDelayForDelay(BufferDelay));
}


// ###### Get next lower buffer delay #######################################
inline cardinal AbstractLayerDescription::getPrevBufferDelay() const
{
   return(getPrevBufferDelayForDelay(BufferDelay));
}









class DescriptionUpdateInterface
{
   public:
   virtual const char* getDescriptionUpdateClass() const = 0;
   virtual bool isVariableBitrate() const = 0;
};


struct BandwidthInfo
{
   cardinal BufferDelay;
   cardinal BytesPerSecond;
   cardinal PacketsPerSecond;
   double   MaxTransferDelay;
   double   MaxLossRate;
   double   MaxJitter;
};


struct ResourceUtilizationPoint
{
   static const cardinal MaxLayers = 12;

   double   Utilization;
   double   FrameRate;
   double   Cost;
   cardinal TaskID;
   cardinal Bandwidth;
   cardinal Layers;
   BandwidthInfo LayerBandwidthInfo[MaxLayers];
};





class AbstractQoSDescription
   : virtual public FrameRateScalabilityInterface /*,
                               virtual public DescriptionUpdateInterface*/
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   AbstractQoSDescription();

   /**
     * Destructor.
     */
   virtual ~AbstractQoSDescription();

   /**
     * Initialize.
     *
     * @param frameRate Frame rate.
     * @param frameRateWeight Frame rate utilization weight.
     */
   inline void initDescription(const double frameRate,
                               const double frameRateWeight);


   // ====== Frame rate methods =============================================
   /**
     * Get frame rate.
     *
     * @return Frame rate.
     */
   inline double getFrameRate() const;

   /**
     * Set frame rate.
     *
     * @param frameRate Frame rate.
     * @return Frame rate set.
     */
   inline double setFrameRate(const double frameRate);

   /**
     * Get next higher frame rate.
     *
     * @return Frame rate.
     */
   inline double getNextFrameRate() const;

   /**
     * Get next lower frame rate.
     *
     * @return Frame rate.
     */
   inline double getPrevFrameRate() const;

   /**
     * Get frame rate scale factor:
     * (frameRate - MinFrameRate) / (MaxFrameRate - MinFrameRate).
     *
     * @return Frame rate scale factor (out of [0,1]).
     */
   inline double getFrameRateScaleFactor() const;

   /**
     * Get frame utilization.
     *
     * @return Frame rate utilization factor (out of [0,1]).
     */
   inline double getFrameRateUtilization() const;

   /**
     * Get frame weight.
     *
     * @return Frame rate utilization weight.
     */
   inline double getFrameRateWeight() const;


   // ====== Total bandwidth methods ========================================
   /**
     * Get minimum required total bandwidth.
     *
     * @return Minimum total bandwidth.
     */
   cardinal getMinBandwidth() const;

   /**
     * Get maximum required total bandwidth.
     *
     * @return Maximum total bandwidth.
     */
   cardinal getMaxBandwidth() const;


   // ====== Abstract layer methods =========================================
   /**
     * Get number of layers.
     *
     * @return Number of layers.
     */
   virtual cardinal getLayers() const = 0;

   /**
     * Get layer.
     *
     * @param layer Layer number.
     * @return Layer.
     */
   virtual AbstractLayerDescription* getLayer(const cardinal layer) const = 0;


   // ====== Get/set resources, calculate utilization for bandwidths ========
   /**
     * Get resources.
     *
     * @param rup ResourceUtilizationPoint reference to store resources.
     * @return Utilization.
     */
   double getResources(ResourceUtilizationPoint& rup) const;

   /**
     * Set resources.
     *
     * @param rup ResourceUtilizationPoint reference containing resources.
     * @return Utilization.
     */
   double setResources(const ResourceUtilizationPoint& rup);

   /**
     * Set resources with buffer delay optimization.
     *
     * @param rup ResourceUtilizationPoint reference containing resources.
     * @return Utilization.
     */
   double setResourcesOptimized(ResourceUtilizationPoint& rup);

   /**
     * Calculate utilization for given frame rate and layers bandwidths.
     *
     * @param frameRate Frame rate.
     * @param layers Number of layers.
     * @param bandwidth Bandwidth array with entry for each layer.
     * @return Utilization.
     */
   virtual double calculateUtilizationForLayerBandwidths(
                     const double    frameRate,
                     const cardinal  layers,
                     const cardinal* bandwidth) const;


   // ====== Resource/Utilization calculation ===============================
   /**
     * Calculate resource utilization list.
     *
     * @param rup ResourceUtilizationPoint array capable of storing maxPoints entries.
     * @param bwThreshold Bandwidth threshold.
     * @param utThreshold Utilization threshold.
     * @param maxPoints Maximum number of ResourceUtilizationPoint to generate.
     */
   virtual cardinal calculateResourceUtilizationList(
                       ResourceUtilizationPoint* rup,
                       const cardinal            bwThreshold,
                       const double              utThreshold,
                       const cardinal            maxPoints) const;

   /**
     * Calculate maximum utilization for given bandwidth.
     *
     * @param totalBandwidth Total bandwidth.
     * @param rup ResourceUtilizationPoint reference to store result.
     * @return Utilization.
     */
   virtual double calculateMaxUtilizationForBandwidth(
                     const cardinal            totalBandwidth,
                     ResourceUtilizationPoint& rup) const = 0;


   // ====== Protected data =================================================
   protected:
   double FrameRate;
   double FrameRateWeight;


   // ====== Private data ===================================================
   private:
   void doResourceUtilizationIteration(
           ResourceUtilizationPoint* rup,
           const cardinal            bwThreshold,
           const double              utThreshold,
           double*                   utilizationCache,
           cardinal*                 bandwidthCache,
           const cardinal            maxPoints,
           const cardinal            maxCachePoints,
           const cardinal            start,
           const cardinal            end,
           const cardinal            startBandwidth,
           const cardinal            endBandwidth,
           const cardinal            level,
           const cardinal            maxLevel,
           cardinal&                 count) const;

   void calculateBandwidthInfo(const cardinal layer,
                               BandwidthInfo& bandwidthInfo) const;
};


// ###### Constructor #######################################################
AbstractQoSDescription::AbstractQoSDescription()
{
   FrameRate       = HUGE_VAL;
   FrameRateWeight = 0.0;
}


// ###### Destructor ########################################################
AbstractQoSDescription::~AbstractQoSDescription()
{
}


// ###### Initialization ####################################################
inline void AbstractQoSDescription::initDescription(const double frameRate,
                                             const double frameRateWeight)
{
   FrameRateWeight = frameRateWeight;
   setFrameRate(frameRate);
}



// ##########################################################################
// #### Frame rate methods                                               ####
// ##########################################################################



// ###### Get frame rate ####################################################
inline double AbstractQoSDescription::getFrameRate() const
{
   return(FrameRate);
}


// ###### Set frame rate ####################################################
inline double AbstractQoSDescription::setFrameRate(const double rate)
{
   if(rate < getMinFrameRate()) {
      FrameRate = getMinFrameRate();
   }
   else if(rate > getMaxFrameRate()) {
      FrameRate = getMaxFrameRate();
   }
   else {
      FrameRate = getNearestValidFrameRate(rate);
   }
   return(FrameRate);
}


// ###### Get next higher frame rate ########################################
inline double AbstractQoSDescription::getNextFrameRate() const
{
   return(getNextFrameRateForRate(FrameRate));
}


// ###### Get next lower frame rate #########################################
inline double AbstractQoSDescription::getPrevFrameRate() const
{
   return(getPrevFrameRateForRate(FrameRate));
}


// ###### Get frame rate scale factor #######################################
inline double AbstractQoSDescription::getFrameRateScaleFactor() const
{
   return(getFrameRateScaleFactorForRate(FrameRate));
}


// ###### Get frame rate utilization ########################################
inline double AbstractQoSDescription::getFrameRateUtilization() const
{
   return(getFrameRateUtilizationForRate(FrameRate));
}


// ###### Get frame rate weight #############################################
inline double AbstractQoSDescription::getFrameRateWeight() const
{
   return(FrameRateWeight);
}


// ###### Get minimum required total bandwidth ##############################
cardinal AbstractQoSDescription::getMinBandwidth() const
{
   cardinal minBandwidth     = 0;
   const cardinal layers     = getLayers();
   const double minFrameRate = getMinFrameRate();
   for(cardinal i = 0;i < layers;i++) {
      const AbstractLayerDescription* layer = getLayer(i);
      minBandwidth += layer->frameSizeToBandwidth(
                         minFrameRate, layer->getMinFrameSize(minFrameRate));
   }
   return(minBandwidth);
}


// ###### Get maximum required total bandwidth ##############################
cardinal AbstractQoSDescription::getMaxBandwidth() const
{
   cardinal maxBandwidth     = 0;
   const cardinal layers     = getLayers();
   const double maxFrameRate = getMaxFrameRate();
   for(cardinal i = 0;i < layers;i++) {
      const AbstractLayerDescription* layer = getLayer(i);
      maxBandwidth += layer->frameSizeToBandwidth(
                         maxFrameRate, layer->getMaxFrameSize(maxFrameRate));
   }
   return(maxBandwidth);
}



// ###########################################################################
// #### Get/set resources, calculate utilization for bandwidths           ####
// ###########################################################################



// ###### Get resources ######################################################
double AbstractQoSDescription::getResources(ResourceUtilizationPoint& rup) const
{
   rup.Cost        = 0.0;
   rup.TaskID      = 0;
   rup.FrameRate   = FrameRate;
   rup.Bandwidth   = 0;
   rup.Layers      = getLayers();

   cardinal bandwidth[rup.Layers];
   for(cardinal i = 0;i < rup.Layers;i++) {
      const AbstractLayerDescription* ald = getLayer(i);
      calculateBandwidthInfo(i,rup.LayerBandwidthInfo[i]);
      bandwidth[i] = rup.LayerBandwidthInfo[i].BytesPerSecond;
      rup.Bandwidth += bandwidth[i];
   }

   rup.Utilization = calculateUtilizationForLayerBandwidths(FrameRate,rup.Layers,(cardinal*)bandwidth);
   return(rup.Utilization);
}


// ###### Set resources ######################################################
double AbstractQoSDescription::setResources(const ResourceUtilizationPoint& rup)
{
   bool ok = (setFrameRate(rup.FrameRate) == rup.FrameRate);

   cardinal bandwidth[rup.Layers];
   for(cardinal i = 0;i < min(rup.Layers,getLayers());i++) {
      AbstractLayerDescription* ald = getLayer(i);
      ald->setBandwidth(FrameRate,rup.LayerBandwidthInfo[i].BytesPerSecond);
      bandwidth[i] = ald->getBandwidth();
   }

   return(calculateUtilizationForLayerBandwidths(FrameRate,rup.Layers,(cardinal*)bandwidth));
}


// ###### Set resources, do buffer delay optimization #######################
double AbstractQoSDescription::setResourcesOptimized(ResourceUtilizationPoint& rup)
{

cout << "NOT IMPL!" << endl;
exit(0);

   return(getResources(rup));
}


// ###### Calculate bandwidth info ##########################################
void AbstractQoSDescription::calculateBandwidthInfo(const cardinal layer,
                                                    BandwidthInfo& bandwidthInfo) const
{
   if(layer >= getLayers()) {
      cerr << "WARNING: AbstractQoSDescription::calculateBandwidthInfo() - " << endl
           << "Invalid parameter " << layer << "!" << endl;
      return;
   }

   const AbstractLayerDescription* ald = getLayer(layer);
   bandwidthInfo.BufferDelay      = ald->getBufferDelay();
   bandwidthInfo.BytesPerSecond   = ald->getBandwidth();
   bandwidthInfo.PacketsPerSecond = ald->getPacketRate(FrameRate);

   bandwidthInfo.MaxTransferDelay = ald->getMaxTransferDelay();
   bandwidthInfo.MaxLossRate      = ald->getMaxLossRate();
   bandwidthInfo.MaxJitter        = ald->getMaxJitter();
}


// ###### Calculate utilization for layer bandwidths ########################
double AbstractQoSDescription::calculateUtilizationForLayerBandwidths(
          const double   frameRate,
          const cardinal layers,
          const cardinal* bandwidth) const
{
   // ====== Initialize =====================================================
   if(layers < getLayers()) {
      return(-1.0);
   }
   double utilization = 0.0;
   double weight      = 0.0;
   bool   ok          = true;

   // ====== Calculate utilization for each layer ===========================
   for(cardinal i = 0;i < getLayers();i++) {
      const AbstractLayerDescription* ald = getLayer(i);

      const cardinal frameSize =ald->bandwidthToFrameSize(frameRate,bandwidth[i]);
      if(frameSize < ald->getMinFrameSize(frameRate)) {
         ok = false;
      }
      else {
         weight      += ald->getFrameSizeWeight();
         utilization += ald->getFrameSizeWeight() *
            ald->getFrameSizeUtilizationForSize(frameRate,frameSize);
      }
   }

   // ====== Calculate global utilization ===================================
   if(ok == true) {
      weight      += FrameRateWeight;
      utilization += getFrameRateWeight() * getFrameRateUtilizationForRate(frameRate);
      utilization /= weight;
   }
   else {
      utilization = -1.0;
   }
   return(utilization);
}


// ###########################################################################
// #### Resource/Utilization calculation                                  ####
// ###########################################################################



// ###### Recursive ResourceUtilizationPoint calculation method #############
void AbstractQoSDescription::doResourceUtilizationIteration(
          ResourceUtilizationPoint* rup,
          const cardinal            bwThreshold,
          const double              utThreshold,
          double*                   utilizationCache,
          cardinal*                 bandwidthCache,
          const cardinal            maxPoints,
          const cardinal            maxCachePoints,
          const cardinal            start,
          const cardinal            end,
          const cardinal            startBandwidth,
          const cardinal            endBandwidth,
          const cardinal            level,
          const cardinal            maxLevel,
          cardinal&                 count) const
{
   // ====== Calculate middle cache position and bandwidth ==================
   const cardinal m            = (start + end) / 2;
   const cardinal newBandwidth = (startBandwidth + endBandwidth) / 2;


   // ====== If level is lowest: Calculate new ResourceUtilizationPoint =====
   if(level == maxLevel) {
      utilizationCache[m] = calculateMaxUtilizationForBandwidth(newBandwidth,rup[count]);
      bandwidthCache[m]   = rup[count].Bandwidth;


      // ====== Add point, if utilization and bandwidth is useful ===========
      if(utilizationCache[m] >= 0.0) {
         // ====== Get nearest lower point's utilization ====================
         double uLeft = utilizationCache[m];
         for(integer i = (integer)m - 1;i >= 0;i--) {
            if(utilizationCache[i] != -HUGE_VAL) {
               uLeft = utilizationCache[i];
               break;
            }
         }

         // ====== Get nearest higher point's utilization ===================
         double uRight = utilizationCache[m];
         for(cardinal i = m + 1;i < maxCachePoints;i++) {
            if(utilizationCache[i] != -HUGE_VAL) {
               uRight = utilizationCache[i];
               break;
            }
         }


         // Add point, if utilization or bandwidth difference is greater than
         // threshold.
         if( ((fabs(utilizationCache[m] - uLeft) >= utThreshold) &&
              (fabs(uRight - utilizationCache[m]) >= utThreshold)) ||
             ((bandwidthCache[m] >= bandwidthCache[start] + bwThreshold) &&
              (bandwidthCache[m] + bwThreshold < bandwidthCache[end])) ) {
            count++;
         }
         else {
            utilizationCache[m] = -HUGE_VAL;
         }
      }
      else {
         utilizationCache[m] = -HUGE_VAL;
      }
   }


   // ====== Go into next recursion level ===================================
   else {
      if(count < maxPoints) {
         if(start + 1 < m) {
            doResourceUtilizationIteration(rup,
                                           bwThreshold, utThreshold,
                                           utilizationCache, bandwidthCache,
                                           maxPoints,
                                           maxCachePoints,
                                           start, m,
                                           startBandwidth, newBandwidth,
                                           level + 1, maxLevel, count);
         }
      }
      if(count < maxPoints) {
         if(m + 1 < end) {
            doResourceUtilizationIteration(rup,
                                           bwThreshold, utThreshold,
                                           utilizationCache, bandwidthCache,
                                           maxPoints,
                                           maxCachePoints,
                                           m, end,
                                           newBandwidth, endBandwidth,
                                           level + 1, maxLevel, count);
         }
      }
   }
}


// ###### Calculate ResourceUtilization list ################################
cardinal AbstractQoSDescription::calculateResourceUtilizationList(
            ResourceUtilizationPoint* rup,
            const cardinal            bwThreshold,
            const double              utThreshold,
            const cardinal            maxPoints) const
{
   // ====== Get minimum and maximum RU points ==============================
   cardinal minBandwidth = getMinBandwidth();
   cardinal maxBandwidth = getMaxBandwidth();
   if(calculateMaxUtilizationForBandwidth(minBandwidth,rup[0]) < 0.0) {
      cerr << "ERROR: Emptry RU list?!" << endl;
      return(0);
   }
   if(calculateMaxUtilizationForBandwidth(maxBandwidth,rup[1]) < 0.0) {
      cerr << "ERROR: Emptry RU list?!" << endl;
      return(0);
   }


   // ====== Initialize utilization and bandwidth caches ====================
   const cardinal additionalDepth = 2;
   const cardinal iterations =
      additionalDepth + (cardinal)ceil(log((double)maxPoints) / log(2.0));
   const cardinal maxCachePoints =  (1 << (iterations + 1));

   double utilizationCache[maxCachePoints];
   cardinal bandwidthCache[maxCachePoints];
   for(cardinal i = 2;i < maxCachePoints;i++) {
      utilizationCache[i] = -HUGE_VAL;
      bandwidthCache[i]   = (cardinal)-1;
   }
   utilizationCache[0]                  = rup[0].Utilization;
   utilizationCache[maxCachePoints - 1] = rup[1].Utilization;
   bandwidthCache[0]                    = rup[0].Bandwidth;
   bandwidthCache[maxCachePoints - 1]   = rup[1].Bandwidth;
   cardinal count = 2;


   // ====== Execute recursive algorithm ====================================
   for(cardinal i = 0;count < maxPoints;i++) {
      const cardinal oldCount = count;
      doResourceUtilizationIteration(rup,
         bwThreshold, utThreshold,
         (double*)&utilizationCache, (cardinal*)&bandwidthCache,
         maxPoints,
         maxCachePoints,
         0, maxCachePoints - 1,
         minBandwidth, maxBandwidth,
         0, i, count);

      if(i >= iterations) {
         break;
      }
   }

   return(count);
}








class MPEGLayerDescription : // virtual public FrameSizeScalabilityInterface,
                             public AbstractLayerDescription,
                             public GenericConstantBitrateFrameSizeScalability
{
   public:
   MPEGLayerDescription();

   const char* getFrameSizeScalabilityClass() const;
};

MPEGLayerDescription::MPEGLayerDescription()
{
}

const char* MPEGLayerDescription::getFrameSizeScalabilityClass() const
{
   return((const char*)"MPEG Video Frame-size Scalability");
}




class MPEGQoSDescription : // virtual public FrameRateScalabilityInterface,
                           public GenericFrameRateScalability,
                           public AbstractQoSDescription
{
   public:
   MPEGQoSDescription();

   const char* getFrameRateScalabilityClass() const;


   cardinal getLayers() const;
   AbstractLayerDescription* getLayer(const cardinal layer) const;

   double calculateMaxUtilizationForBandwidth(
             const cardinal totalBandwidth,
             ResourceUtilizationPoint& rup) const;

   double calculateUtilizationForLayerBandwidths(const double    frameRate,
                                                 const cardinal  layers,
                                                 const cardinal* bandwidth) const;


   protected:
   static const cardinal MaxLayers = 9;

   cardinal Layers;
   MPEGLayerDescription Layer[MaxLayers];
};

MPEGQoSDescription::MPEGQoSDescription()
{
   Layers = 3;
   Layer[0].initConstantBitratePayloadFrameSizeScalability(20000,0.5);
   Layer[0].initLayer(68,1500,(cardinal)-1,6.0,1024.1,20,0.05,50.0);
   Layer[1].initConstantBitratePayloadFrameSizeScalability(5000,0.75);
   Layer[1].initLayer(68,1500,(cardinal)-1,6.0,1024.1,20,0.15,100.0);
   Layer[2].initConstantBitratePayloadFrameSizeScalability(256,1.0);
   Layer[2].initLayer(68,1500,(cardinal)-1,6.0,1024.1,20,0.15,100.0);

   initFrameRateScalability(1,30);
   initDescription(20,5.0 * Layers);
}

const char* MPEGQoSDescription::getFrameRateScalabilityClass() const
{
   return((const char*)"MPEG Video Frame-rate Scalability");
}


cardinal MPEGQoSDescription::getLayers() const
{
   return(Layers);
}

AbstractLayerDescription* MPEGQoSDescription::getLayer(const cardinal layer) const
{
   if(layer >= Layers) {
      cerr << "ERROR: AbstractQoSDescription::getLayer() - " << endl
           << "Invalid parameter " << layer << "!" << endl;
   }
   return((AbstractLayerDescription*)&Layer[layer]);
}


double MPEGQoSDescription::calculateMaxUtilizationForBandwidth(
                              const cardinal            totalBandwidth,
                              ResourceUtilizationPoint& rup) const
{
   // ====== Initialize =====================================================
   rup.Layers      = getLayers();
   rup.Utilization = 0;
   rup.Cost        = 0.0;
   rup.TaskID      = 0;
   for(cardinal i = 0;i < rup.Layers;i++) {
      rup.Bandwidth = 0;
      rup.LayerBandwidthInfo[i].MaxTransferDelay = Layer[i].getMaxTransferDelay();
      rup.LayerBandwidthInfo[i].MaxLossRate      = Layer[i].getMaxLossRate();
      rup.LayerBandwidthInfo[i].MaxJitter        = Layer[i].getMaxJitter();
   }


   // ====== Get minimum bandwidth required =================================
   for(double frameRate = getMinFrameRate();;frameRate = getNextFrameRateForRate(frameRate)) {

      // ====== Initialize iteration ========================================
      cardinal available = totalBandwidth;
      cardinal bandwidth[rup.Layers];
      for(cardinal i = 0;i < rup.Layers;i++) {
         bandwidth[i] = 0;
      }
      double utilization = 0.0;


      // ====== First, use MPEG base layer, then extension layers ===========
      for(cardinal level = 0;level < rup.Layers;level += 3) {


         // ====== Try maximum allocation ===================================
         cardinal required = 0;
         for(cardinal i = 0;i < min(3,rup.Layers);i++) {
            bandwidth[level + i] =
               AbstractLayerDescription::frameSizeToBandwidth(
                  frameRate,Layer[level + i].getMaxFrameSize(frameRate));
            required += bandwidth[level + i];
         }


         // ====== Bandwidth sufficient for maximum allocation ==============
         if(required <= available) {
            for(cardinal i = 0;i < min(3,rup.Layers);i++) {
               available -= bandwidth[level + i];
            }
         }

         // ====== Bandwidth too low for maximum allocation =================
         else {
            // ====== Try minimum allocation ================================
            required = 0;
            cardinal minBandwidth[3];
            for(cardinal i = 0;i < min(3,rup.Layers);i++) {
               bandwidth[level + i] = 0;
               minBandwidth[i] =
                  AbstractLayerDescription::frameSizeToBandwidth(
                     frameRate,Layer[level + i].getMinFrameSize(frameRate));
               required += minBandwidth[i];
            }


            // ====== Check, if bandwidth is too low ========================
            if(required > available) {
               for(cardinal i = 0;i < min(3,rup.Layers);i++) {
                  bandwidth[level + i] = 0;
               }
               if(level == 0) {
                  utilization = -1.0;
               }
               break;
            }


            for(cardinal i = 0;i < min(3,rup.Layers);i++) {
               available -= minBandwidth[i];
            }


            // ====== Divide up remaining bandwidth =========================
            const cardinal size =
               AbstractLayerDescription::bandwidthToFrameSize(frameRate,available);

            const cardinal l0 = Layer[level + 0].getMaxFrameSize(frameRate) - Layer[level + 0].getMinFrameSize(frameRate);
            const cardinal l1 = Layer[level + 1].getMaxFrameSize(frameRate) - Layer[level + 1].getMinFrameSize(frameRate);
            const cardinal l2 = Layer[level + 2].getMaxFrameSize(frameRate) - Layer[level + 2].getMinFrameSize(frameRate);
            const double l  = (double)(l0 + l1 + l2);
            const double d0 = (double)l0 / l;
            const double d1 = (double)l1 / l;
            const double d2 = (double)l2 / l;

            const cardinal f0 = Layer[level + 0].getNearestValidFrameSize(
                                   frameRate, Layer[level + 0].getBufferDelay(),
                                      Layer[level + 0].getMinFrameSize(frameRate) +
                                         (cardinal)floor((cardinal)size * d0));
            const cardinal f1 = Layer[level + 1].getNearestValidFrameSize(
                                   frameRate, Layer[level + 1].getBufferDelay(),
                                      Layer[level + 1].getMinFrameSize(frameRate) +
                                         (cardinal)floor((cardinal)size * d0));
            const cardinal f2 = Layer[level + 2].getNearestValidFrameSize(
                                   frameRate, Layer[level + 2].getBufferDelay(),
                                      Layer[level + 2].getMinFrameSize(frameRate) +
                                         (cardinal)floor((cardinal)size * d0));
            const cardinal b0 = AbstractLayerDescription::frameSizeToBandwidth(
                                   frameRate, f0);
            const cardinal b1 = AbstractLayerDescription::frameSizeToBandwidth(
                                   frameRate, f1);
            const cardinal b2 = AbstractLayerDescription::frameSizeToBandwidth(
                                   frameRate, f2);

            bandwidth[level + 0] += b0;
            bandwidth[level + 1] += b1;
            bandwidth[level + 2] += b2;
            available -= (b0 + b1 + b2);

            break;   // No more bandwidth for extension layers available
         }
      }


      // ====== Calculate utilization =======================================
      if(utilization >= 0.0) {
         utilization = calculateUtilizationForLayerBandwidths(frameRate,rup.Layers,(cardinal*)&bandwidth);

         // ====== Update RUP, if utilization is highest ====================
         if(utilization > rup.Utilization) {
            rup.Utilization = utilization;
            rup.FrameRate   = frameRate;
            rup.Bandwidth   = 0;
            for(cardinal i = 0;i < rup.Layers;i++) {
               rup.Bandwidth += bandwidth[i];
               rup.LayerBandwidthInfo[i].BytesPerSecond   = bandwidth[i];
               rup.LayerBandwidthInfo[i].PacketsPerSecond =
                  Layer[i].frameSizeToPacketRate(frameRate,bandwidth[i]);
            }
         }
      }


      // ====== Check, if frame rate is maximum -> exit =====================
      if(frameRate == getMaxFrameRate()) {
         break;
      }
   }

   return(rup.Utilization);
}


double MPEGQoSDescription::calculateUtilizationForLayerBandwidths(
          const double    frameRate,
          const cardinal  layers,
          const cardinal* bandwidth) const
{
   // return(AbstractQoSDescription::calculateUtilizationForLayerBandwidths(frameRate,layers,bandwidth));

   double u = AbstractQoSDescription::calculateUtilizationForLayerBandwidths(frameRate,layers,bandwidth);

   u *= frameRate;
   u /= 30;
//   cout << u << endl;

   return(u);
}




ostream& operator<<(ostream& os, const AbstractQoSDescription& aqd)
{
   const double frameRate = aqd.getFrameRate() ;
   ResourceUtilizationPoint rup;
   aqd.getResources(rup);

   os << "Bandwidth        = " << aqd.getMinBandwidth() << " <= "
                               << rup.Bandwidth << " <= "
                               << aqd.getMaxBandwidth()   << endl;
   os << "TotalUtilization = " << rup.Utilization * 100.0 << " [%]" << endl;
   os << "Layers           = " << aqd.getLayers() << endl;

   os << "Frame Rate:" << endl;
   os << "   FrameRate        = " << aqd.getMinFrameRate() << " <= "
                               << frameRate             << " <= "
                               << aqd.getMaxFrameRate() << endl;
   os << "   Scalability Cls. = " << aqd.getFrameRateScalabilityClass() << endl;
   os << "   Properties       = ";
   if(aqd.isFrameRateScalable()) os << "<Scalable> "; else os << "<Non-scalable> ";
// ???? VBRInterface   if(aqd.isVariableBitrate())   os << "<VBR> ";      else os << "<CBR> ";
   os << endl;
   os << "   Scale Factor     = " << aqd.getFrameRateScaleFactor() * 100.0
                               << " [%]" << endl;
   os << "   Utilization      = " << aqd.getFrameRateUtilization() * 100.0
                               << " [%]" << endl;
   os << "   Weight           = " << aqd.getFrameRateWeight() << endl;
   os << "   PrevFrameRate    = " << aqd.getPrevFrameRate() << " ("
                               << aqd.getFrameRateUtilizationForRate(aqd.getPrevFrameRate()) * 100.0 << " [%])" << endl;
   os << "   NextFrameRate    = " << aqd.getNextFrameRate() << " ("
                               << aqd.getFrameRateUtilizationForRate(aqd.getNextFrameRate()) * 100.0 << " [%])" << endl;

   BandwidthInfo bi;
   for(cardinal i = 0;i < aqd.getLayers();i++) {
      os << "Layer #" << i << " Frame Size:" << endl;
      const AbstractLayerDescription* ald = aqd.getLayer(i);
      const cardinal frameSize = ald->bandwidthToFrameSize(frameRate,ald->getBandwidth());
//(cardinal)floor((double)ald->getBandwidth() / frameRate);

      os << "   FrameSize        = " << ald->getMinFrameSize(frameRate)  << " <= "
                                     << frameSize << " <= "
                                     << ald->getMaxFrameSize(frameRate)  << "   (Payload: "
                                     << ald->getMinPayloadFrameSizeForDelay(frameRate,ald->getBufferDelay()) << " <= "
                                     << ald->rawToPayload(frameRate,frameSize)    << " <= "
                                     << ald->getMaxPayloadFrameSizeForDelay(frameRate,ald->getBufferDelay()) << ")"
                                     << endl;
      os << "   BytesPerSecond   = " << ald->getBandwidth()           << endl;
      os << "   PacketsPerSecond = " << ald->getPacketRate(frameRate) << " ("
                                     << ald->getPacketCountForSize(frameRate,frameSize)
                                     << " per frame)" << endl;
      os << "   Scalability Cls. = " << ald->getFrameSizeScalabilityClass() << endl;
      os << "   Properties       = ";
      if(ald->isFrameSizeScalable()) os << "<Scalable> "; else os << "<Non-scalable> ";
      if(ald->isVariableBitrate())   os << "<VBR> ";      else os << "<CBR> ";
      os << endl;
      os << "   Scale Factor     = " << ald->getFrameSizeScaleFactorForSize(frameRate,frameSize) * 100.0
                                     << " [%]" << endl;
      os << "   Utilization      = " << ald->getFrameSizeUtilizationForSize(frameRate,frameSize) * 100.0
                                     << " [%]" << endl;
      os << "   Weight           = " << ald->getFrameSizeWeight()     << endl;
      os << "   PrevFrameSize    = " << ald->getPrevFrameSizeForSize(frameRate,frameSize) << " (Payload: "
                                     << ald->rawToPayload(frameRate,ald->getPrevFrameSizeForSize(frameRate,frameSize)) << ", "
                                     << ald->getPayloadFrameSizeUtilizationForDelayAndSize(frameRate,ald->getBufferDelay(),ald->rawToPayload(frameRate,ald->getPrevFrameSizeForDelayAndSize(frameRate,ald->getBufferDelay(),frameSize))) * 100.0
                                     << " [%])" << endl;
      os << "   NextFrameSize    = " << ald->getNextFrameSizeForSize(frameRate,frameSize) << " (Payload: "
                                     << ald->rawToPayload(frameRate,ald->getNextFrameSizeForSize(frameRate,frameSize)) << ", "
                                     << ald->getPayloadFrameSizeUtilizationForDelayAndSize(frameRate,ald->getBufferDelay(),ald->rawToPayload(frameRate,ald->getNextFrameSizeForDelayAndSize(frameRate,ald->getBufferDelay(),frameSize))) * 100.0
                                     << " [%])" << endl;
      os << "   PeakFrameSize    = " << ald->getPeakFrameSizeForSize(frameRate,frameSize) << " (Payload: "
                                     << ald->rawToPayload(frameRate,ald->getPeakFrameSizeForSize(frameRate,frameSize))
                                     << ")" << endl;

      os << "   BufferDelay      = " << ald->getBufferDelay()   << " [Frame(s)]" << endl;
      os << "   MaxTransferDelay = " << ald->getMaxTransferDelay()    << " [ms]" << endl;
      os << "   MaxLossRate      = " << ald->getMaxLossRate() * 100.0 << " [%]"  << endl;
      os << "   MaxJitter        = " << ald->getMaxJitter()           << " [ms]" << endl;
   }
   return(os);
}


ostream& operator<<(ostream& os, const ResourceUtilizationPoint& rup)
{
   char str[300 + (ResourceUtilizationPoint::MaxLayers * 64)];
   char lstr[64];

   // ====== Print header ===================================================
   const cardinal frameRate = (cardinal)rint(rup.FrameRate * 10.0);
   snprintf((char*)&str, 256,
            "Task #%02d: U=%1.2f C=%6f   FPS=%2d.%d BPS=%7d   (",
            rup.TaskID,
            rup.Utilization, rup.Cost,
            frameRate / 10, frameRate % 10,
            rup.Bandwidth);

   // ====== Print each layer's bandwidth ===================================
   for(cardinal i = 0;i < rup.Layers;i++) {
      snprintf((char*)&lstr, sizeof(lstr) - 2, "L%d=%6d ",
               i,
               rup.LayerBandwidthInfo[i].BytesPerSecond);
      if(i < (rup.Layers - 1)) {
         strcat((char*)&lstr,", ");
      }
      strcat((char*)&str,(char*)&lstr);
   }

   os << str << ")";
   return(os);
}





inline swapResourceUtilizationPoints(ResourceUtilizationPoint& a,
                                     ResourceUtilizationPoint& b)
{
   ResourceUtilizationPoint tmp = a;
   a = b;
   b = tmp;
}


void sortResourceUtilizationList(ResourceUtilizationPoint* rup,
                                 const integer            start,
                                 const integer            end)
{
   ResourceUtilizationPoint tmp;
   const double v = rup[(start + end) / 2].Utilization;
   integer i      = start;
   integer j      = end;

   do {
      while(rup[i].Utilization < v) i++;
      while(rup[j].Utilization > v) j--;
      if(i <= j) {
         swapResourceUtilizationPoints(rup[i],rup[j]);
         i++;
         j--;
      }
   } while(j >= i);

   if(start < j) {
      sortResourceUtilizationList(rup,start,j);
   }
   if(i < end) {
      sortResourceUtilizationList(rup,i,end);
   }
}

inline integer ccw(const ResourceUtilizationPoint& p0,
                   const ResourceUtilizationPoint& p1,
                   const ResourceUtilizationPoint& p2)
{
   // ResourceUtilizationPoint:
   //    x = Resource (Bandwidth)
   //    y = Utilization

   const double dx1 = (double)p1.Bandwidth - (double)p0.Bandwidth;
   const double dy1 = p1.Utilization - p0.Utilization;
   const double dx2 = (double)p2.Bandwidth - (double)p0.Bandwidth;
   const double dy2 = p2.Utilization - p0.Utilization;

   if(dx1 * dy2 > dy1 * dx2) {
      return(+1);
   }
   if(dx1 * dy2 < dy1 * dx2) {
      return(-1);
   }
   if((dx1 * dx2 < 0) || (dy1 * dy2 < 0)) {
      return(-1);
   }
   if((dx1 * dy1 + dy1 * dy1) < (dx2 * dx2 + dy2 * dy2)) {
      return(+1);
   }
   return(0);
}

cardinal grahamScanResourceUtilizationList(ResourceUtilizationPoint* rup,
                                           const cardinal            count)
{
   // ResourceUtilizationPoint:
   //    x = Resource (Bandwidth)
   //    y = Utilization

   cardinal i;
   cardinal min;
   cardinal M;

   for(min = 0, i = 1;i < count;i++) {
      if(rup[i].Utilization < rup[min].Utilization) {
         min = i;
      }
   }

   for(i = 0;i < count;i++) {
      if(rup[i].Utilization == rup[min].Utilization) {
         if(rup[i].Bandwidth > rup[min].Bandwidth) {
            min = i;
         }
      }
   }
   swapResourceUtilizationPoints(rup[0],rup[min]);
   sortResourceUtilizationList(rup,0,count - 1);

   for(M = 2, i = 3; i < count;i++) {
      while(ccw(rup[M - 1],rup[M],rup[i]) >= 0) {
              // ????????????????? M, M - 1, i    >= 0 ???????????????
         M--;
         if(M == 0) {
            break;
         }
      }
      M++;
      swapResourceUtilizationPoints(rup[i],rup[M]);
   }
   return(M + 1);
}




cardinal mergeResourceUtilizationLists(ResourceUtilizationPoint*  destination,
                                       ResourceUtilizationPoint** listArray,
                                       const cardinal*            listSizeArray,
                                       const cardinal             listCount)
{
   cardinal pos = 0;
   for(cardinal i = 0;i < listCount;i++) {
      for(cardinal j = 0;j < listSizeArray[i];j++) {
         destination[pos] = listArray[i][j];
         pos++;
      }
   }

   sortResourceUtilizationList(destination,0,pos - 1);
   return(pos);
}

int main()
{
   MPEGQoSDescription q0;

   cardinal bwMin = 30000;
   double   utMin = 0.1;

   ResourceUtilizationPoint* drup = new ResourceUtilizationPoint[1000];
   ResourceUtilizationPoint* rup = new ResourceUtilizationPoint[100];
   ResourceUtilizationPoint* rup1 = new ResourceUtilizationPoint[100];
   ResourceUtilizationPoint* rup2 = new ResourceUtilizationPoint[100];
   ResourceUtilizationPoint* rup3 = new ResourceUtilizationPoint[100];

   cardinal s = q0.calculateResourceUtilizationList(rup, bwMin, utMin, 32);
   sortResourceUtilizationList(rup,0,s-1);

//   sortResourceUtilizationList(rup,0,s - 1);

   for(cardinal i = 0;i < s;i++) {
      cout << rup[i] << endl;

//      cerr << rup[i].Bandwidth << " " << rup[i].Utilization << endl;
   }


   cout << "GrahamScan:" << endl;
   cardinal M = grahamScanResourceUtilizationList(rup,s);
   for(cardinal i = 0;i < M;i++) {
      cout << rup[i] << endl;
//cerr << rup[i].Bandwidth << " " << rup[i].Utilization << endl;
   }

/*
   cout << "Merge:" << endl;
   memcpy(rup1,rup,sizeof(ResourceUtilizationPoint) * M);
   memcpy(rup2,rup,sizeof(ResourceUtilizationPoint) * M);
   memcpy(rup3,rup,sizeof(ResourceUtilizationPoint) * M);
   for(cardinal i = 0;i < M;i++) {
      rup1[i].TaskID = 1;
      rup2[i].TaskID = 2;
      rup3[i].TaskID = 3;
   }


   ResourceUtilizationPoint* list[4];
   list[0] = rup;
   list[1] = rup1;
   list[2] = rup2;
   list[3] = rup3;
   cardinal size[4];
   size[0] = M;
   size[1] = M;
   size[2] = M;
   size[3] = M;

   M = mergeResourceUtilizationLists(drup,(ResourceUtilizationPoint**)&list,
                                         (cardinal*)&size,4);

   for(cardinal i = 0;i < M;i++) {
      cout << drup[i] << endl;
   }
*/


/*
   double ut = q0.calculateMaxUtilizationForBandwidth(750000,rup[0]);
   cout << rup[0] << endl;
*/

/*
   q0.getResources(rup[0]);
   cout << "CURRENT=" << rup[0] << endl;
   q0.setResources(rup[5]);
   q0.getResources(rup[0]);
   cout << "CURRENT=" << rup[0] << endl;
*/
   return(0);
}
