#include "system.h"
#include <fstream.h>

#include <sys/mman.h>
#include <fcntl.h>
#include <sys/time.h>

inline card16 translate16(const card16 x)
{
#ifdef USE_BYTESWAP
   return(bswap_16(x));
#else
   return( (x << 8) | (x >> 8) );
#endif
}

inline card32 translate32(const card32 x)
{
#ifdef USE_BYTESWAP
   return(bswap_32(x));
#else
   return( (x << 24)               |
           ((x & 0x0000ff00) << 8) |
           ((x & 0x00ff0000) >> 8) |
           (x >> 24) );
#endif
}

inline card64 translate64(const card64 x)
{
#ifdef USE_BYTESWAP
   return(bswap_64(x));
#else
   const card32 a = (card32)(x >> 32);
   const card32 b = (card32)(x & 0xffffffff);
   return( (card64)((a << 24) | ((a & 0x0000ff00) << 8) |
                    ((a & 0x00ff0000) >> 8) | (a >> 24))
        |  ((card64)((b << 24) | ((b & 0x0000ff00) << 8) |
                    ((b & 0x00ff0000) >> 8) | (b >> 24)) << 32) );
#endif
}

// ###### Calculate the current time in milliseconds from 01-Jan-1970 #######
card64 getMicroTime()
{
  struct timeval tv;
  gettimeofday(&tv,NULL);
  return(((card64)tv.tv_sec * (card64)1000000) + (card64)tv.tv_usec);
}




// ###### Translate double to 64-bit binary value (32/32 bit separation) ####
inline card64 translateToBinary(const double x)
{
   const double shift = (double)((card64)1 << 32);
   return((card64)((int64)(x * shift)));
}


// ###### Translate 64-bit binary value (32/32 bit separation) to double ####
inline double translateToDouble(const card64 x)
{
   const double shift = (double)((card64)1 << 32);
   return(((double)((int64)x)) / shift);
}




struct TDTFPrefix
{
   char   ID[4];
   card32 Version;


   const static card32   CurrentVersion     = 0x74660001;
   const static cardinal MaxTitleLength     = 64;
   const static cardinal MaxCopyrightLength = 64;

   card64 TimeStamp;
   char   Title[MaxTitleLength];
   char   Copyright[MaxCopyrightLength];
};


struct TDTFSuffix
{
   card64 TimeStamp;
   card32 MainIndex;
   char   ID[4];
};








struct TraceLayerConfiguration
{
   double   CostFactor;
   double   FrameSizeUtilizationA;
   double   FrameSizeUtilizationB;
   double   FrameSizeUtilizationC;
   cardinal EmpiricalEnvelopePairs;
   cardinal FrameCountPairs;
};

struct TraceConfiguration
{
   cardinal FrameRate;

   cardinal ExtLayers;
   double   FakeE1;
   double   FakeE2;

   cardinal MaxBufferDelay;
   double   RemappingCost;

   double   FrameRateUtilizationA;
   double   FrameRateUtilizationB;
   double   FrameRateUtilizationC;

   char   Title[TDTFPrefix::MaxTitleLength];
   char   Copyright[TDTFPrefix::MaxCopyrightLength];

   static const cardinal MaxLayers = 9;
   TraceLayerConfiguration Layer[MaxLayers];
};

ostream& operator<<(ostream& os, const TraceConfiguration config)
{
   cout << "FrameRate            = " << config.FrameRate      << endl;
   cout << "ExtensionLayers      = " << config.ExtLayers      << endl;
   cout << "Fake 1st Ext. Layer  = " << config.FakeE1         << endl;
   cout << "Fake 2nd Ext. Layer  = " << config.FakeE2         << endl;
   cout << "MaxBufferDelay       = " << config.MaxBufferDelay << endl;
   cout << "RemappingCost        = " << config.RemappingCost  << endl;
   cout << "FrameRateUtilization = " << config.FrameRateUtilizationA << ", "
                                     << config.FrameRateUtilizationB << ", "
                                     << config.FrameRateUtilizationC << endl;
   cout << "Title                = " << config.Title          << endl;
   cout << "Copyright            = " << config.Copyright      << endl;
   for(cardinal i = 0;i < config.MaxLayers;i++) {
      cout << "Layer #" << i << ":" << endl;
      cout << "   CostFactor             = " << config.Layer[i].CostFactor             << endl;
      cout << "   EmpiricalEnvelopePairs = " << config.Layer[i].EmpiricalEnvelopePairs << endl;
      cout << "   FrameCountPairs        = " << config.Layer[i].FrameCountPairs        << endl;
      cout << "   FrameSizeUtilization   = " << config.Layer[i].FrameSizeUtilizationA << ", "
                                             << config.Layer[i].FrameSizeUtilizationB << ", "
                                             << config.Layer[i].FrameSizeUtilizationC << endl;
   }
   return(os);
};

TraceConfiguration gConfig = {
   25,

   0,   // ExtLayers
   0.0, // FakeE1
   0.0, // FakeE2

   30,
   3000000.0,

   1.0,2.0,3.0,

   {"Test #1"},
   {"Copyright 2000, Thomas Dreibholz"},

   {
      {
         3.0,
         1.9, 1.8, 1.3,
         12,
         4
      },
      {
         2.0,
         1.7, 1.6, 1.2,
         12,
         4
      },
      {
         1.0,
         1.5, 1.4, 1.1,
         12,
         4
      },

      {
         1.0,
         1.9, 1.8, 1.3,
         12,
         4
      },
      {
         1.0,
         1.7, 1.6, 1.2,
         12,
         4
      },
      {
         1.0,
         1.5, 1.4, 1.1,
         12,
         4
      },

      {
         1.0,
         1.9, 1.8, 1.3,
         12,
         4
      },
      {
         1.0,
         1.7, 1.6, 1.2,
         12,
         4
      },
      {
         1.0,
         1.5, 1.4, 1.1,
         12,
         4
      }
   }
};




#include "strings.h"


bool loadTraceConfiguration(const char*         fileName,
                            TraceConfiguration& config)
{
   // ====== Open file ======================================================
   FILE* inputFD = fopen(fileName,"r");
   if(inputFD == NULL) {
      cerr << "WARNING: Unable to open input file <" << fileName << ">!" << endl;
      return(false);
   }

   // ====== Read file names and initialize AudioReaders ====================
   char str[256];
   char* result = fgets((char*)&str,256,inputFD);
   cardinal line = 0;
   while(!feof(inputFD)) {
      line++;
      const cardinal inputLength = strlen((char*)&str);
      if(inputLength > 1) {
         str[inputLength - 1] = 0x00;
         switch(str[0]) {
            // ====== Line is a comment =====================================
            case '#':
             break;

            // ====== Line is a setting =====================================
            default:
               const String input(str);
               String name;
               String value;
               if(input.scanSetting(name,value)) {
                  if(name == "FRAME RATE UTILIZATION") {
                     double a,b,c;
                     if(sscanf(value.getData(),"%lf %lf %lf",&a,&b,&c) != 3) {
                        cerr << "ERROR: Bad frame rate utilization setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Frame Rate Utilization = <a> <b> <c>" << endl;
                        exit(1);
                     }
                     config.FrameRateUtilizationA = a;
                     config.FrameRateUtilizationB = b;
                     config.FrameRateUtilizationC = c;
                  }
                  else if(name == "LAYER FRAME SIZE UTILIZATION") {
                     unsigned int layer;
                     double        a,b,c;
                     if(sscanf(value.getData(),"#%d %lf %lf %lf",&layer,&a,&b,&c) != 4) {
                        cerr << "ERROR: Bad frame size utilization setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Layer Frame Size Utilization = #<layer> <a> <b> <c>" << endl;
                        exit(1);
                     }
                     if(layer >= config.MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;
                        exit(1);
                     }
                     config.Layer[layer].FrameSizeUtilizationA = a;
                     config.Layer[layer].FrameSizeUtilizationB = b;
                     config.Layer[layer].FrameSizeUtilizationC = c;
                  }
                  else if(name == "LAYER COST FACTOR") {
                     unsigned int layer;
                     double        cost;
                     if(sscanf(value.getData(),"#%d %lf",&layer,&cost) != 2) {
                        cerr << "ERROR: Bad layer cost factor setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Layer Cost Factor = #<layer> <cost>" << endl;
                        exit(1);
                     }
                     if(layer >= config.MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;

                        exit(1);
                     }
                     config.Layer[layer].CostFactor = cost;
                  }
                  else if(name == "REMAPPING COST") {
                     double cost;
                     if(sscanf(value.getData(),"%lf",&cost) != 1) {
                        cerr << "ERROR: Bad remapping cost setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Remapping Cost = <cost>" << endl;
                        exit(1);
                     }
                     config.RemappingCost = cost;
                  }
                  else if(name == "BYTERATE EMPIRICAL ENVELOPE PAIRS") {
                     unsigned int layer;
                     unsigned int pairs;
                     if(sscanf(value.getData(),"#%d %d",&layer,&pairs) != 2) {
                        cerr << "ERROR: Bad byterate empirical envelope pairs setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Byterate Empirical Envelope Pairs = #<layer> <pairs>" << endl;
                        exit(1);
                     }
                     if(layer >= config.MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;

                        exit(1);
                     }
                     if(pairs < 2) {
                        pairs = 2;
                     }
                     if(pairs > 256) {
                        pairs = 256;
                     }
                     config.Layer[layer].EmpiricalEnvelopePairs = pairs;
                  }
                  else if(name == "FRAME COUNT PAIRS") {
                     unsigned int layer;
                     unsigned int pairs;
                     if(sscanf(value.getData(),"#%d %d",&layer,&pairs) != 2) {
                        cerr << "ERROR: Bad frame count pairs setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Frame Count Pairs = #<layer> <pairs>" << endl;
                        exit(1);
                     }
                     if(layer >= config.MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;

                        exit(1);
                     }
                     if(pairs < 2) {
                        pairs = 2;
                     }
                     if(pairs > 256) {
                        pairs = 256;
                     }
                     config.Layer[layer].FrameCountPairs = pairs;
                  }
                  else if(name == "MAX BUFFER DELAY") {
                     unsigned int delay;
                     if(sscanf(value.getData(),"%d",&delay) != 1) {
                        cerr << "ERROR: Bad max buffer delay setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Max Buffer Delay = <delay>" << endl;
                        exit(1);
                     }
                     if(delay < 1) {
                        delay = 1;
                     }
                     if(delay > 1024) {
                        delay = 1024;
                     }
                     config.MaxBufferDelay = delay;
                  }
                  else if(name == "EXTENSION LAYERS") {
                     unsigned int layers;
                     if(sscanf(value.getData(),"%d",&layers) != 1) {
                        cerr << "ERROR: Bad extension layers setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Extension Layers = <0|1|2>" << endl;
                        exit(1);
                     }
                     if(layers > 2) {
                        layers = 2;
                     }
                     config.ExtLayers = layers;
                  }
                  else if(name == "FAKE EXTENSION LAYER") {
                     unsigned int layer;
                     double       factor;
                     if(sscanf(value.getData(),"E%d %lf",&layer,&factor) != 2) {
                        cerr << "ERROR: Bad fake extension layer setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Fake Extension Layer = E<layer> <factor>" << endl;
                        exit(1);
                     }
                     if(factor < 0.0) {
                        factor = 0.0;
                     }
                     if(layer <= 1) {
                        config.FakeE1 = factor;
                     }
                     if(layer >= 2) {
                        config.FakeE2 = factor;
                     }
                  }
                  else if(name == "TITLE") {
                     snprintf((char*)&config.Title,sizeof(config.Title),"%s",
                              value.getData());
                  }
                  else if(name == "COPYRIGHT") {
                     snprintf((char*)&config.Copyright,sizeof(config.Copyright),"%s",
                              value.getData());
                  }
                  else {
                     cerr << "ERROR: Unknown option <"
                          << name << " = " << value << ">, "
                             "line " << line << "!" << endl;
                     exit(1);
                  }
               }
             break;
         }
      }
      result = fgets((char*)&str,256,inputFD);
   }
   fclose(inputFD);
}




struct EmpiricalEnvelopePair
{
   card16 Interval;
   card32 Sum;
};

struct EmpiricalEnvelope
{
   card16                Pairs;
   EmpiricalEnvelopePair Pair[0];
};







struct FrameDescription
{
   card32 ID;
   card32 Size;
};

struct TraceHeader
{
   card64   FrameRate;
   card32   Frames;
   card8    Layers;

   card8    pad01;
   card16   pad02;

 card32 Test;



   /**
     * 0      .. Frames - 1       Layer #0 trace
     * Frames .. 2 * Frames - 1   Layer #1 trace
     * ...
     */
   FrameDescription Frame[0];
};

struct IntervalHeader
{
card32 Test;
   card32 Position;
   card32 Length;

   card64 FrameRateUtilizationA;
   card64 FrameRateUtilizationB;
   card64 FrameRateUtilizationC;

   card8  Layers;
   card8  pad01;
   card16 IntervalDescriptionSize;

   /**
     * 0:        Layer header fr Layer #0
     * 1:        Empirical Envelope header fr Layer #0
     * 2:        Frame count Empirical Envelope header for layer #1
     * 3:        Layer header fr Layer #1
     * 4:        Empirical Envelope header fr Layer #1
     * 5:        Frame count Empirical Envelope header for layer #1
     * ...
     */
   card16 Offset[0];
};


struct LayerHeader
{
card32 Test;
   card64 FrameSizeUtilizationA;
   card64 FrameSizeUtilizationB;
   card64 FrameSizeUtilizationC;
};



struct IndexPositionLengthInterval
{
   card32 Position;
   card32 Length;
   card32 Interval;
};


struct IndexPositionLengthIntervalHeader
{
   card32                      Entries;
   IndexPositionLengthInterval Entry[0];
};


struct MainIndexEntry
{
   card64 FrameRate;
   card32 Trace;
   card32 Intervals;
};


struct MainIndexHeader
{
   card32 Test;
   card32 Entries;
   MainIndexEntry Entry[0];
};






ostream& operator<<(ostream& os, const EmpiricalEnvelope& ee)
{
   char str[256];

   for(cardinal i = 0;i < ee.Pairs;i++) {
      snprintf((char*)&str,sizeof(str),"%2d -> %8d   (%7d per Interval)",
               ee.Pair[i].Interval,
               ee.Pair[i].Sum,
               (cardinal)ceil((double)ee.Pair[i].Sum / (double)ee.Pair[i].Interval));
      os << str << endl;
   }
   return(os);
}






class TraceArray
{
   public:
   TraceArray(const TraceConfiguration& config);
//const card32 maxLayers,
//              const card32 maxFrames);
   virtual ~TraceArray();

   void init(const card32 maxLayers,
             const card32 maxFrames);


   void calculateEmpiricalEnvelope(const cardinal     layer,
                                   const cardinal     start,
                                   const cardinal     end,
                                   const cardinal     traceStart,
                                   const cardinal     traceEnd,
                                   EmpiricalEnvelope* ee,
                                   const bool         frameCount = false) const;

   void TraceArray::calculateOptimalIntervals(
                       double*   costArray,
                       cardinal* lengthArray) const;


   inline double TraceArray::calculateTrafficCost(
                                const cardinal position,
                                const cardinal length) const;


   virtual TraceArray* decreaseFrameRate() const;



   public:
   card32 Layers;
   card32 Frames;
   card32 MaxLayers;
   card32 MaxFrames;
   card32 FrameRate;

   struct Trace {
      card32           Frames;
      FrameDescription Frame[0];
   }** LayerTrace;




   inline cardinal TraceArray::eeSize(const cardinal eePairs) const;
   EmpiricalEnvelope* TraceArray::initEmpiricalEnvelope(
                                     char*          buffer,
                                     const cardinal eePairs,
                                     const bool     frameCount = false) const;

   protected:
   TraceConfiguration Config;

   private:
   cardinal calculateEmpiricalEnvelopePoint(const cardinal layer,
                                            const cardinal start,
                                            const cardinal end,
                                            const cardinal traceStart,
                                            const cardinal traceEnd,
                                            const cardinal delay,
                                            const bool     frameCount = false) const;
};


// ###### Constructor #######################################################
TraceArray::TraceArray(const TraceConfiguration& config)
{
   LayerTrace = NULL;
   Config     = config;
   Layers     = 0;
   Frames     = 0;
   MaxLayers  = 0;
   MaxFrames  = 0;
}


// ###### Destructor ########################################################
TraceArray::~TraceArray()
{
   if(LayerTrace != NULL) {
      for(cardinal i = 0;i < MaxLayers;i++) {
         if(LayerTrace[i] != NULL) {
            delete LayerTrace[i];
         }
      }
      delete LayerTrace;
   }
   LayerTrace = NULL;
}


// ###### Initialize ########################################################
void TraceArray::init(const cardinal maxLayers,
                      const cardinal maxFrames)
{
   // ====== Remove old trace ===============================================
   if(LayerTrace != NULL) {
      for(cardinal i = 0;i < MaxLayers;i++) {
         if(LayerTrace[i] != NULL) {
            delete LayerTrace[i];
         }
      }
      delete LayerTrace;
   }

   // ====== Initialize =====================================================
   Layers    = maxLayers;
   Frames    = 0;
   MaxLayers = maxLayers;
   MaxFrames = maxFrames;

   // ====== Allocate memory for new trace ==================================
   LayerTrace = new Trace*[MaxLayers];
   if(LayerTrace == NULL) {
      cerr << "ERROR: Out of memory!" << endl;
      exit(1);
   }
   for(cardinal i = 0;i < MaxLayers;i++) {
      LayerTrace[i] = NULL;
   }

   for(cardinal i = 0;i < MaxLayers;i++) {
      LayerTrace[i] = (Trace*)new char[sizeof(Trace) + maxFrames * sizeof(FrameDescription)];
      if(LayerTrace[i] == NULL) {
         cerr << "ERROR: Out of memory!" << endl;
         exit(1);
      }
   }
}


// ###### Decrease frame rate by 1 frame/s ##################################
TraceArray* TraceArray::decreaseFrameRate() const
{
   return(NULL);
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const TraceArray& traceArray)
{
   char str[256];

   printf("L=%d   %d\n",traceArray.Layers,traceArray.Frames);


   for(cardinal i = 0;i < traceArray.Frames;i++) {
      snprintf((char*)&str,sizeof(str),"#%06d:   ",i);
      os << str;
      for(cardinal j = 0;j < traceArray.Layers;j++) {
         snprintf((char*)&str,sizeof(str),"%7d  ",traceArray.LayerTrace[j]->Frame[i].Size);
         os << str;
      }
      os << endl;
   }
   return(os);
}


// ###### Calculate one empirical envelope point ############################
cardinal TraceArray::calculateEmpiricalEnvelopePoint(
                        const cardinal layer,
                        const cardinal start,
                        const cardinal end,
                        const cardinal traceStart,
                        const cardinal traceEnd,
                        const cardinal delay,
                        const bool     frameCount = false) const
{
   cardinal istart;
   cardinal iend;

   if(start - traceStart >= delay - 1) {
      istart = traceStart + start - (delay - 1);
   }
   else {
      istart = traceStart;
   }

   if(end + delay - 1 < traceEnd) {
       iend = end + delay - 1;
   }
   else {
       iend = traceEnd;
   }

   cardinal max = 0;
   for(cardinal i = istart;i <= iend;i++) {
      cardinal sum = 0;
      for(cardinal j = i;j <= min(i + (delay - 1),iend);j++) {
         if(!frameCount) {
            sum += LayerTrace[layer]->Frame[j].Size;
         }
         else {
            if(LayerTrace[layer]->Frame[j].Size > 0) {
               sum++;
            }
         }
      }
      if(sum > max) {
         max = sum;
      }
   }

   return(max);
}


// ###### Calculate empirical envelope ######################################
void TraceArray::calculateEmpiricalEnvelope(
                    const cardinal     layer,
                    const cardinal     start,
                    const cardinal     end,
                    const cardinal     traceStart,
                    const cardinal     traceEnd,
                    EmpiricalEnvelope* ee,
                    const bool         frameCount = false) const
{
   for(cardinal i = 0;i < ee->Pairs;i++) {
      ee->Pair[i].Sum = calculateEmpiricalEnvelopePoint(
                           layer,
                           start, end,
                           traceStart, traceEnd,
                           ee->Pair[i].Interval,
                           frameCount);
   }
}


// ###### Calculate size for empirical envelope #############################
inline cardinal TraceArray::eeSize(const cardinal eePairs) const
{
  return(sizeof(EmpiricalEnvelope) + eePairs * sizeof(EmpiricalEnvelopePair));
}


// ###### Initialize intervals of empirical envelope ########################
EmpiricalEnvelope* TraceArray::initEmpiricalEnvelope(
                                  char*          buffer,
                                  const cardinal eePairs,
                                  const bool     frameCount = false) const
{
   // ====== Initialize =====================================================
   const double step = (double)Config.MaxBufferDelay / (double)(eePairs - 1);
   EmpiricalEnvelope* ee = (EmpiricalEnvelope*)buffer;
   ee->Pairs = eePairs;

   // ====== Calculate intervals using MaxBufferDelay =======================
   ee->Pair[0].Interval = 1;
   for(cardinal i = 1;i < ee->Pairs - 1;i++) {
      ee->Pair[i].Interval = (card16)ceil((double)i * step);
   }
   ee->Pair[eePairs - 1].Interval = Config.MaxBufferDelay;

   return(ee);
}


// ###### Calculate cost for given interval #################################
inline double TraceArray::calculateTrafficCost(
                             const cardinal position,
                             const cardinal length) const
{
   double cost = 0.0;

   for(cardinal i = 0;i < Layers;i++) {
      char eeBuffer[eeSize(Config.Layer[i].EmpiricalEnvelopePairs)];
      EmpiricalEnvelope* ee = initEmpiricalEnvelope(
                                 eeBuffer,
                                 Config.Layer[i].EmpiricalEnvelopePairs);

      calculateEmpiricalEnvelope(i,
                                 position, position + length - 1,
                                 0, Frames - 1,
                                 ee);

      double layerCost = 0.0;
      for(cardinal j = 0;j < ee->Pairs;j++) {
         layerCost += ceil((double)ee->Pair[j].Sum /
                              (double)ee->Pair[j].Interval);
      }
      cost += Config.Layer[i].CostFactor * layerCost;
   }

   cost *= (double)length;
   return(cost);
}


// ###### Calculate optimal intervals using Gumbrich's algorithm ############
void TraceArray::calculateOptimalIntervals(
                    double*   costArray,
                    cardinal* lengthArray) const
{
   // ====== Initialize =====================================================
   char str[256];
   for(cardinal i = 0;i < Frames;i++) {
      costArray[i]   = HUGE_VAL;
      lengthArray[i] = 0;
   }
   costArray[Frames - 1]   = calculateTrafficCost(Frames - 1,1);
   lengthArray[Frames - 1] = 1;


   // ====== Execute Gumbrich's algorithm ([Gumb98]) ========================
   cardinal count = 0;
   cerr << "        0 -> ";
   for(integer i = (integer)Frames - 1;i >= 0;i--) {
      for(cardinal length = 1;(cardinal)i + length < Frames;length++) {
         // ====== Compare cost =============================================
         const double cost =
            Config.RemappingCost           +
            calculateTrafficCost(i,length) +
            costArray[i + length];

         if(cost < costArray[i]) {
            costArray[i]   = cost;
            lengthArray[i] = length;
         }


// ????????????????????????????
//if(length >= 10) {  break;  }
// ????????????????????????????


      }

      // ====== Print progress ==============================================
      count++;
      if((count % 10) == 0) {
         cerr << ".";
         if((count % 100) == 0) {
            if((count % 500) != 0) {
               cerr << " ";
            }
            else {
               snprintf((char*)&str,sizeof(str)," [%1.1f%%]",
                           (double)count * 100.0 / Frames);
               cerr << str << endl;
               snprintf((char*)&str,sizeof(str),"   %6d -> ",count);
               cerr << str;
            }
         }
      }
   }

   cerr << endl;

/*
   for(cardinal i = 0;i < Frames;i++) {
      printf("I%02d: L=%d  C=%lf\n",i,lengthArray[i],costArray[i]);
   }
   exit(0);
*/
}







class MPEGTraceArray : public TraceArray
{
   // ====== Constructor ====================================================
   public:
   /**
     * Constructor.
     */
   MPEGTraceArray(const TraceConfiguration& config);


   // ====== MPEG methods ===================================================
   /**
     * Load MPEG trace.
     *
     * @param name File name.
     * @param framePattern Frame pattern of the trace file.
     * @param frameRate Frame rate of the trace file.
     * @param extLayers Number of extension layers (0, 1 or 2).
     * @param fakeE1 Fake 1st extension layer: FrameSizeE1 = fakeE1 * FrameSizeBase.
     * @param fakeE2 Fake 2nd extension layer: FrameSizeE2 = fakeE2 * FrameSizeBase.
     * @return true, if load has been successful; false otherwise.
     */
   bool load(const char*    name,
             const char*    framePattern,
             const cardinal frameRate,
             const cardinal extLayers,
             const double   fakeE1,
             const double   fakeE2);

   /**
     * Reimplementation of TraceArray's decreaseFrameRate() method.
     *
     * @see TraceArray#decreaseFrameRate
     */
   TraceArray* MPEGTraceArray::decreaseFrameRate() const;


   // ====== MPEG layer numbers =============================================
   /**
     * MPEG-I/II I-frames layer.
     */
   static const cardinal LayerMPEGBaseI = 0;

   /**
     * MPEG-I/II P-frames layer.
     */
   static const cardinal LayerMPEGBaseP = 1;

   /**
     * MPEG-I/II B-frames layer.
     */
   static const cardinal LayerMPEGBaseB = 2;

   /**
     * MPEG-II I-frames 1st extension layer.
     */
   static const cardinal LayerMPEGExtI1 = 3;

   /**
     * MPEG-II P-frames 1st extension layer.
     */
   static const cardinal LayerMPEGExtP1 = 4;

   /**
     * MPEG-II B-frames 1st extension layer.
     */
   static const cardinal LayerMPEGExtB1 = 5;

   /**
     * MPEG-II I-frames 2nd extension layer.
     */
   static const cardinal LayerMPEGExtI2 = 6;

   /**
     * MPEG-II P-frames 2nd extension layer.
     */
   static const cardinal LayerMPEGExtP2 = 7;

   /**
     * MPEG-II B-frames 2nd extension layer.
     */
   static const cardinal LayerMPEGExtB2 = 8;
};


MPEGTraceArray::MPEGTraceArray(const TraceConfiguration& config)
   : TraceArray(config)
{
}


// ###### Load MPEG trace ###################################################
bool MPEGTraceArray::load(const char*    name,
                          const char*    framePattern,
                          const cardinal frameRate,
                          const cardinal extLayers,
                          const double   fakeE1,
                          const double   fakeE2)
{
   // ====== Check pattern ==================================================
   const cardinal patternSize = strlen(framePattern);
   if(patternSize < 1) {
      cerr << "ERROR: MPEGTraceArray::load() - Empty frame pattern!" << endl;
      return(false);
   }
   cardinal       pattern[patternSize + 1];
   const cardinal layers = 3 + (3 * extLayers);

printf("L=%d\n",layers);

   for(cardinal i = 0;i < patternSize;i++) {
      switch(toupper(framePattern[i])) {
         case 'I':
            pattern[i] = LayerMPEGBaseI;
          break;
         case 'P':
            pattern[i] = LayerMPEGBaseP;
          break;
         case 'B':
            pattern[i] = LayerMPEGBaseB;
          break;
         default:
            cerr << "ERROR: Illegal frame type " << pattern[i] << "!" << endl;
            return(false);
          break;
      }
   }
   pattern[patternSize] = (cardinal)-1;


   // ====== Open trace file, allocate memory ===============================
   ifstream is(name);
   if(is.good()) {
      // ====== Count frames ================================================
      cardinal frameCount = 0;
      cardinal id         = 0;
      cardinal frameSizeB;
      cardinal frameSizeE1;
      cardinal frameSizeE2;
      cardinal idB;
      cardinal idE1;
      cardinal idE2;
      while(!is.eof()) {
         is >> frameSizeB;
         if(extLayers > 0) {
            if(fakeE1 == 0.0) {
               is >> frameSizeE1;
            }
            if(extLayers > 1) {
               if(fakeE2 == 0.0) {
                  is >> frameSizeE1;
               }
            }
         }
         frameCount++;
      }
      is.close();
      is.open(name);

      // ====== Initialize ==================================================
      cardinal patternPos = 0;
      init(layers,frameCount);
      FrameRate = frameRate;

      // ====== Read frame sizes ============================================
      for(;;) {
         if(Frames >= MaxFrames) {
            cerr << "ERROR: MPEGTraceArray::load() - Too many frames!" << endl;
            return(false);
         }

         // ====== Read frame size ==========================================
         is >> frameSizeB;
         if(is.eof()) {
            break;
         }
         idB = id++;
         if(extLayers > 0) {
            if(fakeE1 != 0.0) {
               frameSizeE1 = (cardinal)ceil(fakeE1 * (double)frameSizeB);
            }
            else {
               is >> frameSizeE1;
               if(is.eof()) {
                  break;
               }
            }
            idE1 = id++;
            if(extLayers > 1) {
               if(fakeE2 != 0.0) {
                  frameSizeE2 = (cardinal)ceil(fakeE2 * (double)frameSizeB);
               }
               else {
                  is >> frameSizeE2;
                  if(is.eof()) {
                     break;
                  }
               }
               idE2 = id++;
            }
         }

         // ====== Determine frame type =====================================
         if(pattern[patternPos] == (cardinal)-1) {
            patternPos = 0;
         }
         const cardinal frameLayer = (cardinal)pattern[patternPos];

         // ====== Write trace into array ===================================
         for(cardinal i = 0;i < MaxLayers;i++) {
            LayerTrace[i]->Frame[Frames].ID   = (cardinal)-1;
            LayerTrace[i]->Frame[Frames].Size = 0;
            LayerTrace[i]->Frames++;
         }
         LayerTrace[frameLayer]->Frame[Frames].ID   = idB;
         LayerTrace[frameLayer]->Frame[Frames].Size = frameSizeB;
         if(extLayers > 0) {
            LayerTrace[frameLayer + 3]->Frame[Frames].ID   = idE1;
            LayerTrace[frameLayer + 3]->Frame[Frames].Size = frameSizeE1;
            if(extLayers > 1) {
               LayerTrace[frameLayer + 6]->Frame[Frames].ID   = idE2;
               LayerTrace[frameLayer + 6]->Frame[Frames].Size = frameSizeE2;
            }
         }

         Frames++;
         patternPos++;
      }

      // ====== Ensure correct alignment ====================================
      const cardinal diff = Frames % FrameRate;
      Frames -= diff;
      return(Frames > 0);
   }
   else {
      cerr << "ERROR: MPEGTraceArray::load() - Unable to load trace " << name << "!" << endl;
      return(false);
   }
}


// ###### Decrease frame rate of MPEG trace by 1 frame/s ####################
TraceArray* MPEGTraceArray::decreaseFrameRate() const
{
   // ====== Initialize =====================================================
   const cardinal frameRate = (cardinal)rint(FrameRate);
   if(frameRate <= 1.0) {
      return(NULL);
   }
   const cardinal frames    = Frames;
   const cardinal layers    = Layers;
   bool skipArray[frames];
   for(cardinal i = 0;i < frames;i++) {
      skipArray[i] = false;
   }


   // ====== Step 1: Drop B-frames ==========================================
   cardinal newFrames = frames;
   for(cardinal i = 0;i < frames;i += frameRate) {
      for(cardinal j = i;j < min(frames,i + frameRate);j++) {
         if(LayerTrace[LayerMPEGBaseB]->Frame[j].Size > 0) {
            skipArray[j] = true;
            newFrames--;
            break;
         }
      }
   }

   // ====== Step 1b: Not enough B-frames -> Drop random P-frames ===========
   if(newFrames < frames) {
      const cardinal calculatedFrames = frames - (frames / frameRate);
      integer diff = (integer)newFrames - (integer)calculatedFrames;
      if(diff > 0) {
         cardinal index[frames];
         cardinal counter  = 0;
         cardinal nextI    = (cardinal)-1;
         bool     newBlock = false;
         for(integer i = (integer)newFrames - 1;i >= 0;i--) {
            if(LayerTrace[LayerMPEGBaseI]->Frame[i].Size > 0) {
               nextI = i;
            }
            if(((i + 1) % frameRate) == 0) {
               newBlock = true;
            }
            if(newBlock == true) {
               if(LayerTrace[LayerMPEGBaseP]->Frame[i].Size > 0) {
                  index[counter] = i;
                  counter++;
                  newBlock = false;
               }
            }
         }

         do {
            cardinal r = random() % counter;
            if(skipArray[index[r]] == false) {
               skipArray[index[r]] = true;
               newFrames--;
            }
         } while(newFrames > calculatedFrames);
      }
   }


   // ====== Step 2: Drop P-frames ==========================================
   else {
      cardinal nextI    = (cardinal)-1;
      bool     newBlock = false;

      for(integer i = (integer)frames - 1;i >= 0;i--) {
         if(LayerTrace[LayerMPEGBaseI]->Frame[i].Size > 0) {
            nextI = i;
         }
         if(((i + 1) % frameRate) == 0) {
            newBlock = true;
         }

         if(newBlock == true) {
            if(LayerTrace[LayerMPEGBaseP]->Frame[i].Size > 0) {
               skipArray[i] = true;
               newFrames--;
               newBlock     = false;
            }
         }
      }


      // ====== Step 2b: Not enough P-frames -> Drop random I-frames ========
      if(newFrames < frames) {
         const cardinal calculatedFrames = frames - (frames / frameRate);
         integer diff = (integer)newFrames - (integer)calculatedFrames;
         if(diff > 0) {
            cardinal index[frames];
            cardinal counter = 0;
            for(cardinal i = 0;i < frames;i++) {
               if(LayerTrace[LayerMPEGBaseI]->Frame[i].Size > 0) {
                  index[counter] = i;
                  counter++;
               }
            }

            do {
               cardinal r = random() % counter;
               if(skipArray[index[r]] == false) {
                  skipArray[index[r]] = true;
                  newFrames--;
               }
            } while(newFrames > calculatedFrames);
         }
      }


      // ====== Step 3: Drop I-frames =======================================
      else if(newFrames == frames) {
         for(cardinal i = 0;i < frames;i += frameRate) {
            for(cardinal j = i;j < min(frames,i + frameRate);j++) {
               if(LayerTrace[LayerMPEGBaseI]->Frame[j].Size > 0) {
                  skipArray[j] = true;
                  newFrames--;
                  break;
               }
            }
         }
      }
   }


   // ====== Allocate trace array ===========================================
   TraceArray* newTraceArray = new MPEGTraceArray(Config);
   if(newTraceArray == NULL) {
      cerr << "ERROR: Out of memory!" << endl;
      exit(1);
   }
   newTraceArray->init(layers,newFrames);
   newTraceArray->Layers    = layers;
   newTraceArray->Frames    = newFrames;
   newTraceArray->FrameRate = frameRate - 1;


   // ====== Copy selected elements =========================================
   cardinal j = 0;
   for(cardinal i = 0;i < frames;i++) {
      if(skipArray[i] == false) {
         for(cardinal k = 0;k < layers;k++) {
            newTraceArray->LayerTrace[k]->Frame[j].ID   = LayerTrace[k]->Frame[i].ID;
            newTraceArray->LayerTrace[k]->Frame[j].Size = LayerTrace[k]->Frame[i].Size;
         }
         j++;
      }
   }

   return(newTraceArray);
}














// FILE* OutputFile = NULL;

class TraceWriter
{
   public:
   TraceWriter();
   ~TraceWriter();


   bool generate(const char*               name,
                 TraceArray*               originalTraceArray,
                 const TraceConfiguration& config);


   private:
   cardinal writeTrace();
   cardinal writeIntervals();

   FILE*              OutputFile;
   TraceArray*        TArray;
   TraceConfiguration Config;
};


TraceWriter::TraceWriter()
{
}

TraceWriter::~TraceWriter()
{
}

cardinal TraceWriter::writeTrace()
{
   cardinal filePosition = ftell(OutputFile);

   // ====== Write trace header =============================================
   cerr << "   Writing trace..." << endl;
   TraceHeader th;
   th.FrameRate = translateToBinary(TArray->FrameRate);
   th.Frames    = translate32(TArray->Frames);
   th.Layers    = (card8)TArray->Layers;
   th.pad01     = 0;
   th.pad02     = 0;
th.Test = 0x14338833;
   if(fwrite((char*)&th,sizeof(th),1,OutputFile) != 1) {
      cerr << "ERROR: Write error!" << endl;
      exit(1);
   }


   // ====== Write trace ====================================================
   for(cardinal j = 0;j < TArray->Layers;j++) {
      FrameDescription frames[TArray->Frames];
      for(cardinal i = 0;i < TArray->Frames;i++) {
         frames[i].ID   = translate32(TArray->LayerTrace[j]->Frame[i].ID);
         frames[i].Size = translate32(TArray->LayerTrace[j]->Frame[i].Size);
      }
      if(fwrite((char*)&frames,sizeof(frames),1,OutputFile) != 1) {
         cerr << "ERROR: Write error!" << endl;
         exit(1);
      }
   }

   return(filePosition);
}



cardinal TraceWriter::writeIntervals()
{
   // ====== Compute intervals ==============================================
   cerr << "   Computing intervals..." << endl;
   double   costArray[TArray->Frames];
   cardinal lengthArray[TArray->Frames];
   TArray->calculateOptimalIntervals((double*)&costArray,(cardinal*)&lengthArray);


   // ====== Count intervals ================================================
   cerr << "   Writing intervals..." << endl;
   cardinal pos       = 0;
   cardinal intervals = 0;
   while(pos < TArray->Frames) {
      pos += lengthArray[pos];
      intervals++;
   }


   // ====== Generate traffic description ===================================
   IndexPositionLengthInterval index[TArray->Frames];
   cardinal indexPosition = 0;
   pos = 0;
   while(pos < TArray->Frames) {
      // ====== Add entry to index ==========================================
      index[indexPosition].Position = translate32(pos);
      index[indexPosition].Length   = translate32(lengthArray[pos]);
      index[indexPosition].Interval = translate32(ftell(OutputFile));
      indexPosition++;


      // ====== Write interval header =======================================
      char ihBuffer[sizeof(IntervalHeader) + (3 * TArray->Layers * sizeof(card16))];

      IntervalHeader* ih = (IntervalHeader*)&ihBuffer;
ih->Test = 0x87654321;
      ih->FrameRateUtilizationA   = translateToBinary(Config.FrameRateUtilizationA);
      ih->FrameRateUtilizationB   = translateToBinary(Config.FrameRateUtilizationB);
      ih->FrameRateUtilizationC   = translateToBinary(Config.FrameRateUtilizationC);
      ih->Position                = translate32(pos);
      ih->Length                  = translate32(lengthArray[pos]);
      ih->Layers                  = (card8)TArray->Layers;
      ih->pad01                   = 0;
      const cardinal indexIntervalHeader = ftell(OutputFile);
      /* Important note:
         This header will be overwritten with updated IntervalHeader
         including offsets for Layer and EmpiricalEnvelope headers!
      */
      ih->IntervalDescriptionSize = sizeof(ihBuffer);
      if(fwrite(ihBuffer,sizeof(ihBuffer),1,OutputFile) != 1) {
         cerr << "ERROR: Write error!" << endl;
         exit(1);
      }


      for(cardinal i = 0;i < TArray->Layers;i++) {
         // ====== Write layer header =======================================
         LayerHeader lh;
   lh.Test = 0x12345678;
         lh.FrameSizeUtilizationA = translateToBinary(Config.Layer[i].FrameSizeUtilizationA);
         lh.FrameSizeUtilizationB = translateToBinary(Config.Layer[i].FrameSizeUtilizationB);
         lh.FrameSizeUtilizationC = translateToBinary(Config.Layer[i].FrameSizeUtilizationC);
         if(fwrite((char*)&lh,sizeof(lh),1,OutputFile) != 1) {
            cerr << "ERROR: Write error!" << endl;
            exit(1);
         }
         ih->Offset[(i * 3) + 0] = translate16(ih->IntervalDescriptionSize);
         ih->IntervalDescriptionSize += sizeof(lh);


         // ====== Calculate empirical envelope =============================
         char eeBuffer[TArray->eeSize(Config.Layer[i].EmpiricalEnvelopePairs)];
         EmpiricalEnvelope* ee =
            TArray->initEmpiricalEnvelope(eeBuffer,Config.Layer[i].EmpiricalEnvelopePairs,false);
         TArray->calculateEmpiricalEnvelope(
                        i, pos, pos + lengthArray[pos] - 1,
                        0, TArray->Frames - 1, ee,
                        false);


         // ====== Write empirical envelope =================================
         for(cardinal j = 0;j < ee->Pairs;j++) {
            ee->Pair[j].Sum      = translate32(ee->Pair[j].Sum);
            ee->Pair[j].Interval = translate16(ee->Pair[j].Interval);
         }
         ee->Pairs = translate16(ee->Pairs);
         if(fwrite((char*)&eeBuffer,sizeof(eeBuffer),1,OutputFile) != 1) {
            cerr << "ERROR: Write error!" << endl;
            exit(1);
         }
         ih->Offset[(i * 3) + 1] = translate16(ih->IntervalDescriptionSize);
         ih->IntervalDescriptionSize += sizeof(eeBuffer);


         // ====== Calculate frame counts ===================================
         char fcBuffer[TArray->eeSize(Config.Layer[i].FrameCountPairs)];
         EmpiricalEnvelope* fc =
            TArray->initEmpiricalEnvelope(fcBuffer,Config.Layer[i].FrameCountPairs,true);
         TArray->calculateEmpiricalEnvelope(
                        i, pos, pos + lengthArray[pos] - 1,
                        0, TArray->Frames - 1, fc,
                        true);

         // ====== Write frame counts =======================================
         for(cardinal j = 0;j < fc->Pairs;j++) {
            fc->Pair[j].Sum      = translate32(fc->Pair[j].Sum);
            fc->Pair[j].Interval = translate16(fc->Pair[j].Interval);
         }
         fc->Pairs = translate16(fc->Pairs);
         if(fwrite((char*)&fcBuffer,sizeof(fcBuffer),1,OutputFile) != 1) {
            cerr << "ERROR: Write error!" << endl;
            exit(1);
         }
         ih->Offset[(i * 3) + 2] = translate16(ih->IntervalDescriptionSize);
         ih->IntervalDescriptionSize += sizeof(fcBuffer);

      }


      // ====== Update header ============================================
      const cardinal oldPosition = ftell(OutputFile);
      fseek(OutputFile,indexIntervalHeader,SEEK_SET);
      ih->IntervalDescriptionSize = translate16(ih->IntervalDescriptionSize);
      if(fwrite(ihBuffer,sizeof(ihBuffer),1,OutputFile) != 1) {
         cerr << "ERROR: Write error!" << endl;
         exit(1);
      }
      fseek(OutputFile,oldPosition,SEEK_SET);


      pos += lengthArray[pos];
   }


   // ====== Write index ====================================================
   cardinal filePosition = ftell(OutputFile);
   IndexPositionLengthIntervalHeader iplih;
   iplih.Entries = translate32(indexPosition);
   if(fwrite((char*)&iplih,sizeof(iplih),1,OutputFile) != 1) {
      cerr << "ERROR: Write error!" << endl;
      exit(1);
   }
   if(fwrite((char*)&index,sizeof(IndexPositionLengthInterval),indexPosition,OutputFile) != indexPosition) {
      cerr << "ERROR: Write error!" << endl;
      exit(1);
   }


/*
for(cardinal i = 0;i < indexPosition;i++) {
      printf("P=%d  L=%d\n",
         translate32(index[i].Position),
         translate32(index[i].Length));
}
exit(0);
*/
   return(filePosition);
}


bool TraceWriter::generate(const char*               name,
                           TraceArray*               originalTraceArray,
                           const TraceConfiguration& config)
{
   // ====== Create file, initialize ========================================
   OutputFile=fopen(name,"w+");
   if(OutputFile == NULL) {
      cerr << "ERROR: TraceWriter::generator() - Unable to create file <"
           << name << ">!" << endl;
      return(false);
   }
   TArray = originalTraceArray;
   Config = config;


   // ====== Writing prefix =================================================
   TDTFPrefix tp;
   tp.ID[0]     = 'T';
   tp.ID[1]     = 'D';
   tp.ID[2]     = 'T';
   tp.ID[3]     = 'F';
   tp.Version   = translate32(TDTFPrefix::CurrentVersion);
   tp.TimeStamp = translate64(getMicroTime());
   strncpy((char*)&tp.Title,(char*)&config.Title,TDTFPrefix::MaxTitleLength);
   strncpy((char*)&tp.Copyright,(char*)&config.Copyright,TDTFPrefix::MaxCopyrightLength);
   for(cardinal i = strlen(tp.Title);i < TDTFPrefix::MaxTitleLength;i++) {
      tp.Title[i] = 0x00;
   }
   for(cardinal i = strlen(tp.Copyright);i < TDTFPrefix::MaxCopyrightLength;i++) {
      tp.Copyright[i] = 0x00;
   }
   if(fwrite((char*)&tp,sizeof(tp),1,OutputFile) != 1) {
      cerr << "ERROR: Write error!" << endl;
      exit(1);
   }


   // ====== Write trace and intervals for each frame rate ==================
   // Limit of possible frame rates: 256
   const cardinal maxFrameRates = 256;
   double   indexFrameRate[maxFrameRates];
   cardinal indexTrace[maxFrameRates];
   cardinal indexIntervals[maxFrameRates];
   cardinal frameRates = 0;

   do {
      cerr << "Computing intervals for frame rate " << TArray->FrameRate
           << "." << endl;

      indexFrameRate[frameRates] = TArray->FrameRate;
      indexTrace[frameRates]     = writeTrace();
      indexIntervals[frameRates] = writeIntervals();
      frameRates++;

      TArray = TArray->decreaseFrameRate();
   } while((TArray != NULL) && (frameRates < maxFrameRates));


   // ====== Write main index ===============================================
   cerr << "Writing index..." << endl;
   const cardinal indexMainIndex = ftell(OutputFile);
   MainIndexHeader ih;
ih.Test = 0x33445566;
   ih.Entries = translate32(frameRates);
   if(fwrite((char*)&ih,sizeof(ih),1,OutputFile) != 1) {
      cerr << "ERROR: Write error!" << endl;
      exit(1);
   }
   for(cardinal i = 0;i < frameRates;i++) {
      MainIndexEntry ie;
      ie.FrameRate = translateToBinary(indexFrameRate[i]);
      ie.Trace     = translate32(indexTrace[i]);
      ie.Intervals = translate32(indexIntervals[i]);
      if(fwrite((char*)&ie,sizeof(ie),1,OutputFile) != 1) {
         cerr << "ERROR: Write error!" << endl;
         exit(1);
      }
   }


   // ====== Write suffix ===================================================
   TDTFSuffix ts;
   ts.ID[0] = 'T';
   ts.ID[1] = 'D';
   ts.ID[2] = 'T';
   ts.ID[3] = 'F';
   ts.MainIndex = translate32(indexMainIndex);
   ts.TimeStamp = translate64(getMicroTime());
   if(fwrite((char*)&ts,sizeof(ts),1,OutputFile) != 1) {
      cerr << "ERROR: Write error!" << endl;
      exit(1);
   }

   fclose(OutputFile);
   OutputFile = NULL;
   cerr << endl << "Congratulations! Your trace file is complete!" << endl;
}








/**
  * This class is a reader for a TDTF trace file. The file will be mapped
  * to memory using mmap() to provide easy access to trace and interval data.
  *
  * @short   Trace Reader
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class TraceReader
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   TraceReader();

   /**
     * Destructor.
     */
   ~TraceReader();


   // ====== Open/close TDTF trace file =====================================
   /**
     * Open TDTF trace file.
     *
     * @param name Name of TDTF trace file to open.
     * @return true, if open operation has been successful; false otherwise.
     */
   bool open(const char* name);

   /**
     * Close TDTF trace file.
     */
   void close();


   // ====== Frame rate methods =============================================
   /**
     * Check, if given frame rate is a valid value.
     *
     * @param frameRate Frame rate to be checked.
     * @return true, if given rate is valid; false otherwise.
     */
   bool isValidFrameRate(const double frameRate) const;

   /**
     * Get nearest lower valid frame rate for given frame rate.
     *
     * @param rate Frame rate.
     * @return Valid frame rate nearest to given rate.
     */
   double getNearestValidFrameRate(const double frameRate) const;

   /**
     * Get minimum frame rate.
     *
     * @return Minimum frame rate.
     */
   inline double getMinFrameRate() const;

   /**
     * Get maximum frame rate.
     *
     * @return Maximum frame rate.
     */
   inline double getMaxFrameRate() const;


   /**
     * Get next higher valid frame rate for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Next higher valid frame rate.
     */
   double getNextFrameRateForRate(const double frameRate) const;

   /**
     * Get next lower valid frame rate for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Next lower valid frame rate.
     */
   double getPrevFrameRateForRate(const double frameRate) const;


   // ====== Frame size methods =============================================
   /**
     * Get maximum number of bytes for given buffer delay (in frame rate units).
     *
     * @param position Position within media.
     * @param layer Layer number.
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @return Maximum number of bytes.
     */
    cardinal getMaxByteCountForDelay(const cardinal position,
                                     const cardinal layer,
                                     const double   frameRate,
                                     const cardinal delay) const;

   /**
     * Get maximum number of frames for given buffer delay (in frame rate units).
     *
     * @param position Position within media.
     * @param layer Layer number.
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @return Maximum number of frames.
     */
    cardinal getMaxFrameCountForDelay(const cardinal position,
                                      const cardinal layer,
                                      const double   frameRate,
                                      const cardinal delay) const;

   /**
     * Get payload frame size for given buffer delay (in frame rate units).
     *
     * @param position Position within media.
     * @param layer Layer number.
     * @param frameRate Frame rate.
     * @param bufferDelay Buffer delay in frame rate units.
     * @return Payload frame size.
     */
   inline cardinal getPayloadFrameSizeForDelay(
                      const cardinal position,
                      const cardinal layer,
                      const double   frameRate,
                      const cardinal bufferDelay) const;

   /**
     * Get payload frame size for given position, layer and frame rate from trace.
     *
     * @param position Position within media.
     * @param layer Layer number.
     * @param frameRate Frame rate.
     * @return Frame size.
     */
   cardinal getTraceFrameSize(const cardinal position,
                              const cardinal layer,
                              const double   frameRate) const;

   /**
     * Get frame ID for given position, layer and frame rate from trace.
     *
     * @param position Position within media.
     * @param layer Layer number.
     * @param frameRate Frame rate.
     * @return Frame ID.
     */
   cardinal getTraceFrameID(const cardinal position,
                            const cardinal layer,
                            const double   frameRate) const;

   /**
     * Get number of frames for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Number of frames.
     */
   cardinal getFrames(const double frameRate) const;

   /**
     * Get number of layers for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Number of layers.
     */
   cardinal getLayers(const double frameRate) const;


   // ====== Buffer delay methods ===========================================
   /**
     * Get maximum buffer delay for given position and frame rate.
     *
     * @param position Position within media.
     * @param frameRate Frame rate.
     * @return Maximum buffer delay
     */
   cardinal getMaxBufferDelay(const cardinal position,
                              const double   frameRate) const;


   // ====== TDTF trace file access =========================================
   /**
     * Check, if access to trace file at given position for given length
     * is within valid range.
     *
     * @param position File Position.
     * @param length Access length.
     * @return true, is access is valid; false otherwise.
     */
   bool checkAccess(const cardinal position, const cardinal length) const;

   /**
     * Get TDTF prefix.
     *
     * @return Pointer to TDTF prefix.
     */
   const TDTFPrefix* TraceReader::getTDTFPrefix() const;

   /**
     * Get TDTF suffix.
     *
     * @return Pointer to TDTF suffix.
     */
   const TDTFSuffix* TraceReader::getTDTFSuffix() const;

   /**
     * Get entry for given frame rate from main index.
     *
     * @param frameRate Frame rate.
     * @return Pointer to entry.
     */
   const MainIndexEntry* getMainIndexEntry(const double frameRate) const;

   /**
     * Get trace header for given frame rate.
     *
     * @param frameRate Frame rate.
     * @return Pointer to trace header.
     */
   const TraceHeader* getTraceHeader(const double frameRate) const;

   /**
     * Get entry for given frame rate from Position/Length/Interval index.
     *
     * @param frameRate Frame rate.
     * @return Pointer to entry.
     */
   const IndexPositionLengthIntervalHeader* getIPLIHeader(
                                               const double frameRate) const;

   /**
     * Get interval header for given frame rate and position.
     *
     * @param position Position.
     * @param frameRate Frame rate.
     * @return Pointer to trace header.
     */
   const IntervalHeader* getIntervalHeader(const cardinal position,
                                           const double frameRate) const;

   /**
     * Get layer header for given frame rate and position and layer.
     *
     * @param position Position.
     * @param layer Layer.
     * @param frameRate Frame rate.
     * @return Pointer to layer header.
     */
   const LayerHeader* getLayerHeader(const cardinal position,
                                     const cardinal layer,
                                     const double   frameRate) const;

   /**
     * Get empirical envelope for byterate or frame count for given
     * frame rate, layer and position.
     *
     * @param position Position.
     * @param layer Layer.
     * @param frameRate Frame rate.
     * @param frameCount true to get empirical envelope for frameCount; false for byterate.
     * @return Pointer to empirical envelope header.
     */
   const EmpiricalEnvelope* getEmpiricalEnvelope(const cardinal position,
                                                 const cardinal layer,
                                                 const double   frameRate,
                                                 const bool     frameCount = false) const;


   // ====== Utilization methods ============================================
   /**
     * Get frame size utilization constants for given frame rate, layer and
     * position.
     *
     * @param position Position.
     * @param layer Layer.
     * @param frameRate Frame rate.
     * @param a Reference to store constant A.
     * @param b Reference to store constant B.
     * @param c Reference to store constant C.
     * @return true on success; false otherwise.
     */
   bool getFrameSizeUtilizationConstants(
           const cardinal position,
           const cardinal layer,
           const double   frameRate,
           double&        a,
           double&        b,
           double&        c) const;

   /**
     * Get frame rate utilization constants for given frame rate, layer and
     * position.
     *
     * @param position Position.
     * @param frameRate Frame rate.
     * @param a Reference to store constant A.
     * @param b Reference to store constant B.
     * @param c Reference to store constant C.
     * @return true on success; false otherwise.
     */
   bool getFrameRateUtilizationConstants(
           const cardinal position,
           const double   frameRate,
           double&        a,
           double&        b,
           double&        c) const;


   // ====== Printing methods ===============================================
   /**
     * Print empirical envelopes for all layers for byterate or
     * frame count for given frame rate and position.
     *
     * @param os Output stream.
     * @param position Position.
     * @param frameRate Frame rate.
     * @param compact true, to print *without* EE approximations; false otherwise.
     * @param frameCount true to get empirical envelope for frameCount; false for byterate.
     */
   void printEmpiricalEnvelope(ostream&       os,
                               const double   frameRate,
                               const cardinal position,
                               const bool     frameCount,
                               const bool     compact) const;

   /**
     * Print complete TDTF trace for given frame rate.
     *
     * @param os Output stream.
     * @param frameRate Frame rate.
     * @param printEE true, to print empirical envelope; false otherwise.
     * @param compactEE true, to print Empirical envelope *without* approximations; false otherwise.
     */
   void print(ostream&     os,
              const double frameRate,
              const bool   printEE,
              const bool   compactEE) const;


   // ====== Protected data =================================================
   protected:
   char*            InputMemory;
   cardinal         InputLength;

   MainIndexHeader* MainIndex;
   cardinal         MainIndexEntries;

   double           MinFrameRate;
   double           MaxFrameRate;


   // ====== Private data ===================================================
   private:
   int              InputFile;
};


// ###### Constructor #######################################################
TraceReader::TraceReader()
{
   InputFile   = -1;
   InputMemory = NULL;

   // Use close() method to reset attributes!
   close();
}


// ###### Destructor ########################################################
TraceReader::~TraceReader()
{
   close();
}


// ###### Check, if trace file access is within valid range #################
bool TraceReader::checkAccess(const cardinal position, const cardinal length) const
{
   if(position + length <= InputLength) {
      return(true);
   }

   cerr << "WARNING: TraceReader::checkAccess() - "
           "Illegal access to " << position << ", length " << length << "!" << endl;
   return(false);
}


// ###### Close TDTF trace file #############################################
void TraceReader::close()
{
   if(InputFile != -1) {
      ::close(InputFile);
      InputFile = -1;
   }
   if(InputMemory != NULL) {
      munmap(InputMemory,InputLength);
      InputMemory = NULL;
   }
   InputLength      = 0;
   MainIndex        = NULL;
   MainIndexEntries = 0;
   MinFrameRate     = 0.0;
   MaxFrameRate     = 0.0;
}


// ###### Open TDTF trace file ##############################################
bool TraceReader::open(const char* name)
{
   // ====== Open file ======================================================
   close();
   InputFile = ::open(name,O_RDONLY);
   if(InputFile <= 0) {
      cerr << "ERROR: TraceReader::open() - Unable to open <"
           << name << ">!" << endl;
      return(false);
   }
   off_t len = lseek(InputFile,0,SEEK_END);
   if(len < 0) {
      cerr << "ERROR: TraceReader::open() - lseek() failed!" << endl;
      return(false);
   }
   InputLength = (cardinal)len;


   // ====== Map file to memory =============================================
   InputMemory = (char*)mmap(NULL,InputLength,PROT_READ,MAP_SHARED,InputFile,0);
   if(InputMemory == NULL) {
      cerr << "ERROR: TraceReader::open() - mmap() failed!" << endl;
      return(false);
   }


   // ====== Check format ===================================================
   const TDTFPrefix* tp = (TDTFPrefix*)&InputMemory[0];
   if((tp->ID[0] != 'T') && (tp->ID[1] != 'D') && (tp->ID[2] != 'T') && (tp->ID[3] != 'F')) {
      cerr << "ERROR: TraceReader::open() - No TDTF header!" << endl;
      return(false);
   }
   if(translate32(tp->Version) != TDTFPrefix::CurrentVersion) {
      cerr << "ERROR: TraceReader::open() - Incompatible trace version!" << endl;
      return(false);
   }
   const TDTFSuffix* ts = (TDTFSuffix*)&InputMemory[InputLength - sizeof(TDTFSuffix)];
   if((ts->ID[0] != 'T') && (ts->ID[1] != 'D') && (ts->ID[2] != 'T') && (ts->ID[3] != 'F')) {
      cerr << "ERROR: TraceReader::open() - Corrupted TDTF file -- Missing TDTF suffix!" << endl;
      return(false);
   }


   // ====== Read main index ================================================
   const cardinal indexMainIndex = translate32(ts->MainIndex);
   if(checkAccess(translate32(ts->MainIndex),sizeof(MainIndexHeader)) == false) {
      return(false);
   }
   MainIndex        = (MainIndexHeader*)&InputMemory[indexMainIndex];
   MainIndexEntries = translate32(MainIndex->Entries);
   if(checkAccess(indexMainIndex,MainIndexEntries * sizeof(MainIndexEntry)) == false) {
      cerr << "ERROR: TraceReader::open() - Bad main index!" << endl;
      return(false);
   }


   // ====== Initialize =====================================================
   // Note: Frame rates are in *de*creasing order!
   MinFrameRate = translateToDouble(MainIndex->Entry[MainIndexEntries - 1].FrameRate);
   MaxFrameRate = translateToDouble(MainIndex->Entry[0].FrameRate);


   // ====== Do some integrity checks =======================================
   for(cardinal i = 0;i < MainIndexEntries;i++) {
      if(checkAccess(translate32(MainIndex->Entry[i].Trace),sizeof(TraceHeader)) == false) {
         cerr << "ERROR: TraceReader::open() - Bad main index reference to trace!" << endl;
         return(false);
      }
      if(checkAccess(translate32(MainIndex->Entry[i].Intervals),sizeof(IndexPositionLengthIntervalHeader)) == false) {
         cerr << "ERROR: TraceReader::open() - Bad main index reference to interval!" << endl;
         return(false);
      }
   }

   return(true);
}


// ###### Get minimum frame rate ############################################
inline double TraceReader::getMinFrameRate() const
{
   return(MinFrameRate);
}


// ###### Get maximum frame rate ############################################
inline double TraceReader::getMaxFrameRate() const
{
   return(MaxFrameRate);
}


// ###### Check, if frame rate is valid #####################################
bool TraceReader::isValidFrameRate(const double frameRate) const
{
   for(cardinal i = 0;i < MainIndexEntries;i++) {
      const double f = translateToDouble(MainIndex->Entry[i].FrameRate);
      if(f == frameRate) {
         return(true);
      }
   }
   return(false);
}


// ###### Get nearest lower valid frame rate ################################
double TraceReader::getNearestValidFrameRate(const double frameRate) const
{
   // ====== Check limits ===================================================
   if(frameRate <= MinFrameRate) {
      return(MinFrameRate);
   }
   else if(frameRate >= MaxFrameRate) {
      return(MaxFrameRate);
   }

   // ====== Search for frame rate ==========================================
   // Note: Frame rates are in *de*creasing order!
   for(cardinal i = 0;i < MainIndexEntries;i++) {
      const double f = translateToDouble(MainIndex->Entry[i].FrameRate);
      if(f <= frameRate) {
         return(f);
      }
   }
   return(MinFrameRate);
}


// ###### Get next higher frame rate ########################################
double TraceReader::getNextFrameRateForRate(const double frameRate) const
{
   // ====== Check limits ===================================================
   if(frameRate >= MaxFrameRate) {
      return(MaxFrameRate);
   }
   else if(frameRate < MinFrameRate) {
      return(MinFrameRate);
   }

   // ====== Search for frame rate ==========================================
   // Note: Frame rates are in *de*creasing order!
   for(integer i = (integer)MainIndexEntries - 1;i >= 0;i--) {
      const double f = translateToDouble(MainIndex->Entry[i].FrameRate);
      if(f > frameRate) {
         return(f);
      }
   }
   return(MaxFrameRate);
}


// ###### Get next lower frame rate #########################################
double TraceReader::getPrevFrameRateForRate(const double frameRate) const
{
   // ====== Check limits ===================================================
   if(frameRate > MaxFrameRate) {
      return(MaxFrameRate);
   }
   else if(frameRate <= MinFrameRate) {
      return(MinFrameRate);
   }

   // ====== Search for frame rate ==========================================
   // Note: Frame rates are in *de*creasing order!
   for(cardinal i = 0;i < MainIndexEntries;i++) {
      const double f = translateToDouble(MainIndex->Entry[i].FrameRate);
      if(f < frameRate) {
         return(f);
      }
   }
   return(MinFrameRate);
}



// ###### Get main index entry ##############################################
const MainIndexEntry* TraceReader::getMainIndexEntry(const double frameRate) const
{
   cardinal start = 0;
   cardinal end   = MainIndexEntries - 1;

   // ====== Check limits ===================================================
   if(frameRate <= translateToDouble(MainIndex->Entry[end].FrameRate)) {
      return(&MainIndex->Entry[end]);
   }
   else if(frameRate >= translateToDouble(MainIndex->Entry[0].FrameRate)) {
      return(&MainIndex->Entry[0]);
   }

   // ====== Do binary search ===============================================
   else {
      cardinal m = 0;
      do {
         m = (start + end) / 2;
         const double f = translateToDouble(MainIndex->Entry[m].FrameRate);
         if(frameRate > f) {
            end = m;
         }
         else if(frameRate < f) {
            start = m;
         }
         else {
            return(&MainIndex->Entry[m]);
         }
      } while(end - start > 1);
   }
   return(NULL);
}


// ###### Get TDTF prefix ###################################################
const TDTFPrefix* TraceReader::getTDTFPrefix() const
{
   return((TDTFPrefix*)InputMemory);
}


// ###### Get TDTF suffix ###################################################
const TDTFSuffix* TraceReader::getTDTFSuffix() const
{
   return((TDTFSuffix*)&InputMemory[InputLength - sizeof(TDTFSuffix)]);
}


// ###### Get trace header ##################################################
const TraceHeader* TraceReader::getTraceHeader(const double frameRate) const
{
   // ====== Get main index entry ===========================================
   const MainIndexEntry* ie = getMainIndexEntry(frameRate);
   if(ie == NULL) {
      return(NULL);
   }
   const cardinal indexTrace = translate32(ie->Trace);

   // ====== Check reference integrity ======================================
   const TraceHeader* th = (const TraceHeader*)&InputMemory[indexTrace];
   if(checkAccess(indexTrace,sizeof(TraceHeader)) == false) {
      return(NULL);
   }
   if(checkAccess(indexTrace,sizeof(TraceHeader) +
                     translate32(th->Frames) * th->Layers * sizeof(FrameDescription)) == false) {
      return(NULL);
   }
   return(th);
}


// ###### Get PositionLengthInterval index header ###########################
const IndexPositionLengthIntervalHeader* TraceReader::getIPLIHeader(
                                            const double frameRate) const
{
   // ====== Get main index entry ===========================================
   const MainIndexEntry* ie = getMainIndexEntry(frameRate);
   if(ie == NULL) {
      return(NULL);
   }
   const cardinal indexIPLI = translate32(ie->Intervals);

   // ====== Check reference integrity ======================================
   const IndexPositionLengthIntervalHeader* ipli = (const IndexPositionLengthIntervalHeader*)&InputMemory[indexIPLI];
   if(checkAccess(indexIPLI,sizeof(IndexPositionLengthIntervalHeader)) == false) {
      return(NULL);
   }
   if(checkAccess(indexIPLI,sizeof(IndexPositionLengthIntervalHeader) +
                     translate32(ipli->Entries) * sizeof(IndexPositionLengthInterval)) == false) {
      return(NULL);
   }
   return(ipli);
}


// ###### Get interval header ###############################################
const IntervalHeader* TraceReader::getIntervalHeader(
                                      const cardinal position,
                                      const double   frameRate) const
{
   // ====== Get PositionLengthInterval index header ========================
   const IndexPositionLengthIntervalHeader* ipli = getIPLIHeader(frameRate);
   if(ipli == NULL) {
      return(NULL);
   }

   cardinal indexInterval = (cardinal)-1;
   cardinal start         = 0;
   cardinal end           = translate32(ipli->Entries) - 1;


   // ====== Check limits ===================================================
   if(position <= translate32(ipli->Entry[0].Position)) {
      indexInterval = translate32(ipli->Entry[0].Interval);
   }
   else if(position >= translate32(ipli->Entry[end].Position)) {
      indexInterval = translate32(ipli->Entry[end].Interval);
   }


   // ====== Do binary search ===============================================
   else {
      cardinal m = 0;
      do {
         m = (start + end) / 2;
         const cardinal p0 = translate32(ipli->Entry[m].Position);
         const cardinal p1 = p0 + translate32(ipli->Entry[m].Length) - 1;
         if(position < p0) {
            end = m;
         }
         else if(position > p1) {
            start = m;
         }
         else {
            start = m;
            end   = m;
         }
      } while(end - start > 1);
      indexInterval = translate32(ipli->Entry[start].Interval);
   }


   // ====== Check reference integrity ================================
   const IntervalHeader* ih = (const IntervalHeader*)&InputMemory[indexInterval];
   if(checkAccess(indexInterval,sizeof(IntervalHeader)) == false) {
      return(NULL);
   }
   if(checkAccess(indexInterval,sizeof(IndexPositionLengthIntervalHeader) +
                                   translate16(ih->IntervalDescriptionSize)) == false) {
      return(NULL);
   }
   if((position < translate32(ih->Position)) ||
      (position > translate32(ih->Position) + translate32(ih->Length) - 1)) {
      cerr << "ERROR: TraceReader::getIntervalHeader() - Position check failed!" << endl;
      return(NULL);
   }
   return(ih);
}


// ###### Get layer header ##################################################
const LayerHeader* TraceReader::getLayerHeader(
                                   const cardinal position,
                                   const cardinal layer,
                                   const double   frameRate) const
{
   // ====== Get Interval index header ======================================
   const IntervalHeader* ih = getIntervalHeader(position,frameRate);
   if(ih == NULL) {
      cerr << "ERROR: getLayerHeader() - Invalid parameters "
           << "P=" << position << ", FR=" << frameRate << "!" << endl;
      return(NULL);
   }


   // ====== Get layer header ===============================================
   if(layer >= ih->Layers) {
      cerr << "ERROR: getLayerHeader() - Invalid parameter "
           << "L=" << layer << endl;
      return(NULL);
   }
   cardinal offset = (layer * 3) + 0;
   const cardinal lhOffset = (cardinal)translate16(ih->Offset[offset]);
   const LayerHeader* lh   = (const LayerHeader*)((long)ih + (long)lhOffset);

   // ====== Check reference integrity ================================
   if(checkAccess((long)lh - (long)InputMemory,sizeof(LayerHeader)) == false) {
      return(NULL);
   }

   return(lh);
}


// ###### Get empirical envelope ############################################
const EmpiricalEnvelope* TraceReader::getEmpiricalEnvelope(
                                         const cardinal position,
                                         const cardinal layer,
                                         const double   frameRate,
                                         const bool     frameCount) const
{
   // ====== Get Interval index header ======================================
   const IntervalHeader* ih = getIntervalHeader(position,frameRate);
   if(ih == NULL) {
      cerr << "ERROR: getEmpiricalEnvelope() - Invalid parameters "
           << "P=" << position << ", FR=" << frameRate << "!" << endl;
      return(NULL);
   }


   // ====== Get empirical envelope =========================================
   if(layer >= ih->Layers) {
      cerr << "ERROR: getEmpiricalEnvelope() - Invalid parameter "
           << "L=" << layer << endl;
      return(NULL);
   }
   cardinal offset = (layer * 3) + 1;
   if(frameCount) {
      offset++;
   }
   const cardinal eeOffset = (cardinal)translate16(ih->Offset[offset]);
   const EmpiricalEnvelope* ee = (const EmpiricalEnvelope*)((long)ih + (long)eeOffset);


   // ====== Check reference integrity ================================
   if(checkAccess((long)ee - (long)InputMemory,sizeof(EmpiricalEnvelope)) == false) {
      return(NULL);
   }
   if(checkAccess((long)ee - (long)InputMemory,sizeof(EmpiricalEnvelope) +
                                translate16(ee->Pairs) * sizeof(EmpiricalEnvelopePair)) == false) {
      return(NULL);
   }
   return(ee);
}


// ###### Get maximum payload frame size ####################################
inline cardinal TraceReader::getPayloadFrameSizeForDelay(
                         const cardinal position,
                         const cardinal layer,
                         const double   frameRate,
                         const cardinal bufferDelay) const
{
   const cardinal count = getMaxByteCountForDelay(position,layer,frameRate,bufferDelay);
   return((cardinal)ceil((double)count / (double)bufferDelay));

}


// ###### Get maximum byte count ############################################
cardinal TraceReader::getMaxByteCountForDelay(
                         const cardinal position,
                         const cardinal layer,
                         const double   frameRate,
                         const cardinal bufferDelay) const
{
   // ====== Get empirical envelope =========================================
   const EmpiricalEnvelope* ee = getEmpiricalEnvelope(position,layer,frameRate);
   if(ee == NULL) {
      return(0);
   }
   const cardinal pairs = translate16(ee->Pairs);


   // ====== No approximation necessary if interval is in table =============
   for(cardinal i = 0;i < pairs;i++) {
      if(translate16(ee->Pair[i].Interval) == bufferDelay) {
         return(translate32(ee->Pair[i].Sum));
      }
   }


   // ====== Calculate indices for D-BIND approx. calculation ([KZ97]) ======
   integer k0 = 0;
   for(integer i = 0;i < pairs;i++) {
      const cardinal interval = translate16(ee->Pair[i].Interval);
      if(interval >= bufferDelay) {
         k0 = max(i - 1,0);
         break;
      }
   }
   if(k0 >= pairs) {
      k0 = pairs - 1;
   }
   integer k1 = (integer)pairs - 1;
   for(integer i = k1;i >= 0;i--) {
      const cardinal interval = translate16(ee->Pair[i].Interval);
      if(interval <= bufferDelay) {
         k1 = min(i + 1,pairs - 1);
         break;
      }
   }

   // ====== Do D-BIND approximation ([KZ97]) ===============================
// ????????????????????????????????????????????
   printf("   [%d %d]   ",k0,k1);
   fflush(stdout);

   const double i0 = (double)translate16(ee->Pair[k0].Interval);
   const double i1 = (double)translate16(ee->Pair[k1].Interval);
   const double r0 = (double)translate32(ee->Pair[k0].Sum);
   const double r1 = (double)translate32(ee->Pair[k1].Sum);
   const double n = ((r1 * i1) - (r0 * i0)) * (double)(bufferDelay - i1);
   const double m = i1 - i0;
   return((cardinal)ceil( (n / m) + (r1 * i1) ));
}


// ###### Get maximum frame count ###########################################
cardinal TraceReader::getMaxFrameCountForDelay(const cardinal position,
                                               const cardinal layer,
                                               const double   frameRate,
                                               const cardinal delay) const
{
   // ====== Get empirical envelope =========================================
   const EmpiricalEnvelope* fc = getEmpiricalEnvelope(position,layer,frameRate,true);
   if(fc == NULL) {
      return(0);
   }
   const cardinal pairs = translate16(fc->Pairs);


   // ====== Get maximum frame count for delay ==============================
   cardinal count = delay;
   for(cardinal i = 0;i < pairs;i++) {
      const cardinal interval = translate16(fc->Pair[i].Interval);
      const cardinal sum      = translate32(fc->Pair[i].Sum);
      if(interval >= delay) {
         count = sum;
         break;
      }
   }

   return(count);
}


// ###### Get number of frames ##############################################
cardinal TraceReader::getFrames(const double frameRate) const
{
   const TraceHeader* trace = getTraceHeader(frameRate);
   if(trace != NULL) {
      return(translate32(trace->Frames));
   }
   return(0);
}


// ###### Get number of layers ##############################################
cardinal TraceReader::getLayers(const double frameRate) const
{
   const TraceHeader* trace = getTraceHeader(frameRate);
   if(trace != NULL) {
      return(trace->Layers);
   }
   return(0);
}


// ###### Get frame size for position from trace ############################
cardinal TraceReader::getTraceFrameSize(const cardinal position,
                                        const cardinal layer,
                                        const double   frameRate) const
{
   const TraceHeader* trace = getTraceHeader(frameRate);
   if(trace != NULL) {
      const cardinal frames = translate32(trace->Frames);
      if(position < frames) {
         return(translate32(trace->Frame[layer * frames + position].Size));
      }
   }
   return(0);
}


// ###### Get frame ID for position from trace ############################
cardinal TraceReader::getTraceFrameID(const cardinal position,
                                      const cardinal layer,
                                      const double   frameRate) const
{
   const TraceHeader* trace = getTraceHeader(frameRate);
   if(trace != NULL) {
      const cardinal frames = translate32(trace->Frames);
      if(position < frames) {
         return(translate32(trace->Frame[layer * frames + position].ID));
      }
   }
   return(0);
}


// ###### Get maximum buffer delay ##########################################
cardinal TraceReader::getMaxBufferDelay(const cardinal position,
                                        const double   frameRate) const
{
   // ====== Get empirical envelope for each layer ==========================
   const cardinal layers = getLayers(frameRate);
   cardinal minDelay = (cardinal)-1;
   for(cardinal j = 0;j < layers;j++) {
      const EmpiricalEnvelope* ee = getEmpiricalEnvelope(position,j,frameRate);
      if(ee == NULL) {
         cerr << "ERROR: TraceReader::getMaxBufferDelay() - Empirical envelope header not found!" << endl;
         exit(1);
      }
      const cardinal pairs = translate16(ee->Pairs);
      if(translate16(ee->Pair[pairs - 1].Interval) < minDelay) {
         minDelay = translate16(ee->Pair[pairs - 1].Interval);
      }
   }
   return(minDelay);
}


// ###### Get utilization constants for frame size ##########################
bool TraceReader::getFrameSizeUtilizationConstants(
                     const cardinal position,
                     const cardinal layer,
                     const double   frameRate,
                     double&        a,
                     double&        b,
                     double&        c) const
{
   // ====== Get layer header ===============================================
   const LayerHeader* lh = getLayerHeader(position,layer,frameRate);
   if(lh == NULL) {
      return(false);
   }

   // ====== Get constants ==================================================
   a = translateToDouble(lh->FrameSizeUtilizationA);
   b = translateToDouble(lh->FrameSizeUtilizationB);
   c = translateToDouble(lh->FrameSizeUtilizationC);
   return(true);
}


// ###### Get utilization constants for frame rate ##########################
bool TraceReader::getFrameRateUtilizationConstants(
                     const cardinal position,
                     const double   frameRate,
                     double&        a,
                     double&        b,
                     double&        c) const
{
   // ====== Get interval header ============================================
   const IntervalHeader* ih = getIntervalHeader(position,frameRate);
   if(ih == NULL) {
      return(false);
   }

   // ====== Get constants ==================================================
   a = translateToDouble(ih->FrameRateUtilizationA);
   b = translateToDouble(ih->FrameRateUtilizationB);
   c = translateToDouble(ih->FrameRateUtilizationC);
   return(true);
}


// ###### Print empirical envelope data for given frame rate ################
void TraceReader::printEmpiricalEnvelope(ostream&       os,
                                         const double   frameRate,
                                         const cardinal position,
                                         const bool     frameCount,
                                         const bool     compact) const
{
   const cardinal           layers = getLayers(frameRate);
   const EmpiricalEnvelope* ee[layers];
   char                     str[256];


   // ====== Get empirical envelope for each layer ==========================
   cardinal maxPairs       = 0;
   cardinal maxBufferDelay = 0;
   for(cardinal j = 0;j < layers;j++) {
      ee[j] = getEmpiricalEnvelope(position,j,frameRate,frameCount);
      if(ee[j] == NULL) {
         cerr << "ERROR: TraceReader::print() - Empirical envelope header not found!" << endl;
         exit(1);
      }
      const cardinal pairs = translate16(ee[j]->Pairs);
      if(pairs > maxPairs) {
         maxPairs = pairs;
      }
      if(maxBufferDelay < translate16(ee[j]->Pair[pairs - 1].Interval)) {
         maxBufferDelay = translate16(ee[j]->Pair[pairs - 1].Interval);
      }
   }


   // ====== Print empirical envelope for each layer ========================
   os << " ";
   for(cardinal j = 0;j < layers;j++) {
      snprintf((char*)&str,sizeof(str),"Layer #%d       ",j);
      os << str;
   }
   os << endl;


   // ====== Full view, with EE approximations ==============================
   if(!compact) {
      for(cardinal j = 1;j <= maxBufferDelay;j++) {
         os << " ";
         for(cardinal k = 0;k < layers;k++) {
            cardinal sum;
            if(frameCount == false) {
               sum = getMaxByteCountForDelay(position,k,frameRate,j);
            }
            else {
               sum = getMaxFrameCountForDelay(position,k,frameRate,j);
            }

            bool checked = false;
            for(cardinal i = 0;i < translate16(ee[k]->Pairs);i++) {
               if(j == translate16(ee[k]->Pair[i].Interval)) {
                  checked = true;
                  if(sum != translate32(ee[k]->Pair[i].Sum)) {
                     cerr << "INTERNAL ERROR: TraceReader::printEmpiricalEnvelope() - Verification assertion!"
                          << endl;
                     exit(1);
                  }
                  break;
               }
            }

            if(checked == false) {
               snprintf((char*)&str,sizeof(str),"%2d -> %6d   ",j,sum);
            }
            else {
               snprintf((char*)&str,sizeof(str),"%2d +> %6d   ",j,sum);
            }
            os << str;
         }
         os << endl;
      }
   }

   // ====== Compact view, without EE approximations ===========================
   else {
      for(cardinal j = 0;j < maxPairs;j++) {
         os << " ";
         for(cardinal k = 0;k < layers;k++) {
            // ====== Check, if there are pairs to print =======================
            if(j < translate16(ee[k]->Pairs)) {
               // ====== Print entry ===========================================
               const cardinal interval = translate16(ee[k]->Pair[j].Interval);
               const cardinal sum      = translate32(ee[k]->Pair[j].Sum);
               snprintf((char*)&str,sizeof(str),"%2d +> %6d   ",interval,sum);
               os << str;

               // ====== Do verification =======================================
               if(frameCount) {
                  const cardinal s2 = getMaxFrameCountForDelay(position,k,frameRate,interval);
                  if(s2 != sum) {
                     cerr << "INTERNAL ERROR: TraceReader::printEmpiricalEnvelope() - Verification assertion for frame count!"
                          << endl;
                     exit(1);
                  }
               }
               else {
                  const cardinal s2 = getPayloadFrameSizeForDelay(position,k,frameRate,interval);
                  const cardinal s1 = (cardinal)ceil((double)sum / (double)interval);
                  if(s2 != s1) {
                     cerr << "INTERNAL ERROR: TraceReader::printEmpiricalEnvelope() - Verification assertion for frame size!"
                          << endl;
                     exit(1);
                  }
               }
            }

            // ====== No more pairs to print in this layer =====================
            else {
               os << "               ";
            }
         }
         os << endl;
      }
   }
}


// ###### Print TDTF trace for given frame rate #############################
void TraceReader::print(ostream&     os,
                        const double frameRate,
                        const bool   printEE,
                        const bool   compactEE) const
{
   // ====== Initialize =====================================================
   const IntervalHeader* ih;
   const cardinal frames    = getFrames(frameRate);
   const cardinal layers    = getLayers(frameRate);
   cardinal       intLength = 0;
   cardinal       intCount  = 0;
   char           str[256];


   // ====== Get interval count =============================================
   const IndexPositionLengthIntervalHeader* ipli = getIPLIHeader(frameRate);
   if(ipli == NULL) {
      cerr << "ERROR: TraceReader::print() - PositionLengthInterval header not found!" << endl;
      exit(1);
   }
   const cardinal intMaxCount = translate32(ipli->Entries);


   // ====== Print all frames ===============================================
   for(cardinal i = 0;i < frames;i++) {
      // ====== Update interval description =================================
      if(intLength <= 0) {
         // ====== Get interval header ======================================
         ih = getIntervalHeader(i,frameRate);
         if(ih == NULL) {
            cerr << "ERROR: TraceReader::print() - Interval header not found!" << endl;
            exit(1);
         }
         intLength = translate32(ih->Length);
         intCount++;

         snprintf((char*)&str,sizeof(str),
                  "==== Interval %05d of %05d, length %04d =============",
                  intCount,intMaxCount,intLength);
         os << endl << str;
         for(cardinal j = 3;j < layers;j++) {
            os << "==============";
         }
         os << endl;


         // ===== Print empirical envelopes =================================
         if(printEE == true) {
            os << "Byterate Empirical Envelope:" << endl;
            printEmpiricalEnvelope(os,frameRate,i,false,compactEE);
            os << "Frame Count Empirical Envelope:" << endl;
            printEmpiricalEnvelope(os,frameRate,i,true,compactEE);
         }
      }
      intLength--;


      // ====== Print trace =================================================
      snprintf((char*)&str,sizeof(str),"#%06d: ",i);
      os << str;
      cardinal size[layers];
      for(cardinal j = 0;j < layers;j++) {
         size[j] = getTraceFrameSize(i,j,frameRate);
         snprintf((char*)&str,sizeof(str),"%6d ",size[j]);
         os << str;
      }
      os << "   (";
      for(cardinal j = 0;j < layers;j++) {
         if(size[j] != 0) {
            snprintf((char*)&str,sizeof(str),"#%05d",
                        getTraceFrameID(i,j,frameRate));
         }
         else {
            strcpy((char*)&str,"#-----");
         }
         os << str;
         if(j < layers - 1) {
            os << " ";
         }
      }
      os << ")" << endl;
   }
}




void main(int argc, char* argv[])
{
   loadTraceConfiguration("test.cfg",gConfig);

   cout << "TDTF Trace Generator - Copyright (C) 2000 Thomas Dreibholz" << endl
        << "----------------------------------------------------------" << endl
        << endl;

   cout << gConfig << endl;

int g= 0 ;

   if(g) {
   MPEGTraceArray array(gConfig);



   if(array.load("test2.t",
                 "IBBPBBPBBPBB",
                 gConfig.FrameRate,
                 gConfig.ExtLayers,
                 gConfig.FakeE1,
                 gConfig.FakeE2) == true) {

       array.Frames = 50;

 cout << array << endl;
 //  exit(0);

/*
   double c = array.calculateTrafficCost(0,49);
   printf("C=%lf\n",c);
   exit(0);
*/


      TraceWriter w;
      w.generate("test.out",&array,gConfig);
    }
    else {
       exit(1);
    }
   }


   TraceReader tr;
   if(tr.open("test.out")) {
      const TDTFPrefix* tp = tr.getTDTFPrefix();
      const TDTFSuffix* ts = tr.getTDTFSuffix();
      char str[256];


      cout << "TDTF Trace Printer - Copyright (C) 2000 Thomas Dreibholz" << endl
           << "--------------------------------------------------------" << endl
           << endl;
      cout << "Title            = " << tp->Title     << endl;
      cout << "Copyright        = " << tp->Copyright << endl;

      time_t timeStamp1  = translate64(tp->TimeStamp) / 1000000;
      struct tm* timeptr = localtime(&timeStamp1);
      strftime((char*)&str,256,"%A, %B %d %G, %T",timeptr);
      cout << "Generation Start = " << str << endl;

      time_t timeStamp2  = translate64(ts->TimeStamp) / 1000000;
      timeptr   = localtime(&timeStamp2);
      strftime((char*)&str,256,"%A, %B %d %G, %T",timeptr);
      cout << "Generation End   = " << str << endl;

      time_t diff = timeStamp2 - timeStamp1;
      snprintf((char*)&str,sizeof(str),"%d:%02d:%02d",
                  diff / 3600, (diff / 60) % 60, diff % 60);
      cout << "Duration         = " << str << endl;

      cout << "Frame Rates      = {";
      double frameRate = tr.getMinFrameRate();
      while(frameRate != tr.getMaxFrameRate()) {
         cout << frameRate << " ";
         frameRate = tr.getNextFrameRateForRate(frameRate);
      }
      cout << tr.getMaxFrameRate() << "}" << endl;

      cout << endl;



   //   exit(0);




      tr.print(cout,25.0,true,false);
   }
}
