// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Abstract QoS Description Implementation                          ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "abstractqosdescription.h"


namespace Coral {


// Debug mode: Verify calculated resource/utilization lists.
#define DEBUG


// ###### Constructor #######################################################
AbstractQoSDescription::AbstractQoSDescription()
{
   FrameRate          = HUGE_VAL;
   WantedUtilization  = 1.0;
   MinWantedBandwidth = 0;
   MaxWantedBandwidth = (cardinal)-1;
   StreamPriority     = 0;
   PktMaxSize         = 1500;
   PktHeaderSize      = 28;
   Position           = (card64)-1;
}


// ###### Destructor ########################################################
AbstractQoSDescription::~AbstractQoSDescription()
{
}



// ##########################################################################
// #### Bandwidth methods                                                ####
// ##########################################################################



// ###### Get minimum required total bandwidth ##############################
cardinal AbstractQoSDescription::getMinBandwidth() const
{
   cardinal minBandwidth     = 0;
   const cardinal layers     = getLayers();
   const double minFrameRate = getMinFrameRate();
   for(cardinal i = 0;i < layers;i++) {
      const AbstractLayerDescription* layer = getLayer(i);
      if(!(layer->getFlags() & AbstractLayerDescription::LF_ExtensionLayer)) {
         minBandwidth += layer->frameSizeToBandwidth(
                            minFrameRate, layer->getMinFrameSize(minFrameRate));
      }
   }
   return(minBandwidth);
}


// ###### Get maximum required total bandwidth ##############################
cardinal AbstractQoSDescription::getMaxBandwidth() const
{
   cardinal maxBandwidth     = 0;
   const cardinal layers     = getLayers();
   const double maxFrameRate = getMaxFrameRate();
   for(cardinal i = 0;i < layers;i++) {
      const AbstractLayerDescription* layer = getLayer(i);
      maxBandwidth += layer->frameSizeToBandwidth(
                         maxFrameRate, layer->getMaxFrameSize(maxFrameRate));
   }
   return(maxBandwidth);
}



// ###########################################################################
// #### Get/set resources, calculate utilization for bandwidths           ####
// ###########################################################################



// ###### Get resources ######################################################
double AbstractQoSDescription::getResources(ResourceUtilizationPoint& rup) const
{
   rup.Cost        = 0.0;
   rup.TaskID      = 0;
   rup.FrameRate   = FrameRate;
   rup.Bandwidth   = 0;
   rup.Layers      = getLayers();

   cardinal bandwidth[rup.Layers];
   for(cardinal i = 0;i < rup.Layers;i++) {
      calculateBandwidthInfo(i,rup.LayerBandwidthInfo[i]);
      bandwidth[i] = rup.LayerBandwidthInfo[i].BytesPerSecond;
      rup.Bandwidth += bandwidth[i];
   }

   rup.Utilization = calculateUtilizationForLayerBandwidths(FrameRate,rup.Layers,(cardinal*)bandwidth);
   return(rup.Utilization);
}


// ###### Set resources ######################################################
double AbstractQoSDescription::setResources(const ResourceUtilizationPoint& rup)
{
   cardinal bandwidth[min(rup.Layers,getLayers())];
   setFrameRate(rup.FrameRate);
   for(cardinal i = 0;i < min(rup.Layers,getLayers());i++) {
      AbstractLayerDescription* ald = getLayer(i);
      ald->setBandwidth(FrameRate,rup.LayerBandwidthInfo[i].BytesPerSecond);
      bandwidth[i] = ald->getBandwidth();
   }
   return(calculateUtilizationForLayerBandwidths(FrameRate,rup.Layers,(cardinal*)bandwidth));
}


// ###### Set resources, do buffer delay optimization #######################
double AbstractQoSDescription::setResourcesOptimized(ResourceUtilizationPoint& rup)
{

   cout << "NOT YET IMPLEMENTED!" << endl;
   exit(10);

   return(getResources(rup));
}


// ###### Calculate bandwidth info ##########################################
void AbstractQoSDescription::calculateBandwidthInfo(const cardinal layer,
                                                    BandwidthInfo& bandwidthInfo) const
{
   if(layer >= getLayers()) {
      cerr << "WARNING: AbstractQoSDescription::calculateBandwidthInfo() - " << endl
           << "Invalid parameter " << layer << "!" << endl;
      return;
   }

   const AbstractLayerDescription* ald = getLayer(layer);
   bandwidthInfo.BufferDelay      = ald->getBufferDelay();
   bandwidthInfo.BytesPerSecond   = ald->getBandwidth();
   bandwidthInfo.PacketsPerSecond = ald->getPacketRate(FrameRate);
   bandwidthInfo.MaxTransferDelay = ald->getMaxTransferDelay();
   bandwidthInfo.MaxLossRate      = ald->getMaxLossRate();
   bandwidthInfo.MaxJitter        = ald->getMaxJitter();
}



// ###########################################################################
// #### Resource/Utilization calculation                                  ####
// ###########################################################################


// ###### Calculate utilization for layer bandwidths ########################
double AbstractQoSDescription::calculateUtilizationForLayerBandwidths(
          const double    frameRate,
          const cardinal  layers,
          const cardinal* bandwidth) const
{
   // ====== Initialize =====================================================
   if(layers < getLayers()) {
      return(-1.0);
   }
   double utilization = 0.0;
   double weight      = 0.0;
   bool   ok          = true;


   // ====== Calculate utilization for each layer ===========================
   for(cardinal i = 0;i < getLayers();i++) {
      const AbstractLayerDescription* ald = getLayer(i);
      const cardinal frameSize = ald->bandwidthToFrameSize(frameRate,bandwidth[i]);
      const double w = ald->getFrameSizeUtilizationWeight(frameRate);
      weight      += w;
      if(frameSize < ald->getMinFrameSize(frameRate)) {
         if(!(ald->getFlags() & AbstractLayerDescription::LF_ExtensionLayer)) {
            ok = false;
            break;
         }
         else if(frameSize > 0) {
            cerr << "WARNING: AbstractQoSDescription::calculateUtilizationForLayerBandwidths() -"
                    "Senseless allocation to extension layer?!" << endl;
         }
      }
      else {
         utilization += w *
                        ald->getFrameSizeUtilizationForSize(frameRate,frameSize);
      }
   }


   // ====== Calculate global utilization ===================================
   if(ok == true) {
      const double w = getFrameRateUtilizationWeight(frameRate);
      weight      += w;
      utilization += w * getFrameRateUtilizationForRate(frameRate);
      utilization /= weight;
   }
   else {
      utilization = -1.0;
   }
   return(utilization);
}


// ###### Calculate maximum utilization for given total bandwidths ##########
void AbstractQoSDescription::calculateMaxUtilizationForBandwidthArray(
                                const cardinal*           totalBandwidthArray,
                                ResourceUtilizationPoint* rupArray,
                                const cardinal            points) const
{
   cerr << "WARNING: AbstractQoSDescription::calculateMaxUtilizationForBandwidthArray() - "
           "Needs to be overloaded..." << endl;

}


// ###### Recursive ResourceUtilizationPoint calculation method #############
void AbstractQoSDescription::doResourceUtilizationIteration(
          ResourceUtilizationPoint* rup,
          const cardinal            bwThreshold,
          const double              utThreshold,
          double*                   utilizationCache,
          cardinal*                 bandwidthCache,
          const cardinal            maxPoints,
          const cardinal            maxCachePoints,
          const cardinal            start,
          const cardinal            end,
          const cardinal            startBandwidth,
          const cardinal            endBandwidth,
          const cardinal            level,
          const cardinal            maxLevel,
          cardinal&                 count) const
{
   // ====== Calculate middle cache position and bandwidth ==================
   const cardinal m            = (start + end) / 2;
   const cardinal newBandwidth = (startBandwidth + endBandwidth) / 2;


   // ====== If level is lowest: Calculate new ResourceUtilizationPoint =====
   if(level == maxLevel) {
      utilizationCache[m] = calculateMaxUtilizationForBandwidth(newBandwidth,rup[count]);
      bandwidthCache[m]   = rup[count].Bandwidth;


      // ====== Add point, if utilization and bandwidth is useful ===========
      if(utilizationCache[m] >= 0.0) {
         // ====== Get nearest lower point's utilization ====================
         double   uLeft = utilizationCache[m];
         cardinal bLeft = bandwidthCache[m];
         for(integer i = (integer)m - 1;i >= 0;i--) {
            if(utilizationCache[i] != -HUGE_VAL) {
               uLeft = utilizationCache[i];
               bLeft = bandwidthCache[i];
               break;
            }
         }

         // ====== Get nearest higher point's utilization ===================
         double   uRight = utilizationCache[m];
         cardinal bRight = bandwidthCache[m];
         for(cardinal i = m + 1;i < maxCachePoints;i++) {
            if(utilizationCache[i] != -HUGE_VAL) {
               uRight = utilizationCache[i];
               bRight = bandwidthCache[i];
               break;
            }
         }

         // Add point, if utilization or bandwidth difference is greater than
         // threshold.
         if( ((utilizationCache[m] - uLeft >= utThreshold) &&
              (uRight - utilizationCache[m] >= utThreshold)) ||
             (((integer)bandwidthCache[m] - (integer)bLeft >= (integer)bwThreshold) &&
              ((integer)bRight - (integer)bandwidthCache[m] >= (integer)bwThreshold)) ) {
            count++;
         }
         else {
            utilizationCache[m] = -HUGE_VAL;
         }
      }
      else {
         utilizationCache[m] = -HUGE_VAL;
      }
   }


   // ====== Go into next recursion level ===================================
   else {
      if(count < maxPoints) {
         if(start + 1 < m) {
            doResourceUtilizationIteration(rup,
                                           bwThreshold, utThreshold,
                                           utilizationCache, bandwidthCache,
                                           maxPoints,
                                           maxCachePoints,
                                           start, m,
                                           startBandwidth, newBandwidth,
                                           level + 1, maxLevel, count);
         }
      }
      if(count < maxPoints) {
         if(m + 1 < end) {
            doResourceUtilizationIteration(rup,
                                           bwThreshold, utThreshold,
                                           utilizationCache, bandwidthCache,
                                           maxPoints,
                                           maxCachePoints,
                                           m, end,
                                           newBandwidth, endBandwidth,
                                           level + 1, maxLevel, count);
         }
      }
   }
}


// ###### Calculate ResourceUtilization list ################################
cardinal AbstractQoSDescription::calculateResourceUtilizationList(
            ResourceUtilizationPoint* rup,
            const cardinal            bwThreshold,
            const double              utThreshold,
            const cardinal            maxPoints) const
{
   // ====== Get minimum and maximum RU points ==============================
   cardinal minBandwidth = getMinBandwidth();
   cardinal maxBandwidth = getMaxBandwidth();
   // printf("min=%d  max=%d\n",minBandwidth,maxBandwidth);

   if(calculateMaxUtilizationForBandwidth(minBandwidth,rup[0]) < 0.0) {
      cerr << "WARNING: AbstractQoSDescription::calculateResourceUtilizationList() - "
              "Empty resource/utilization list for minimum bandwidth?!" << endl;
      return(0);
   }
   if(calculateMaxUtilizationForBandwidth(maxBandwidth,rup[1]) < 0.0) {
      cerr << "WARNING: AbstractQoSDescription::calculateResourceUtilizationList() - "
              "Empty resource/utilization list for maximum bandwidth?!" << endl;
      return(0);
   }

/*
   cout << "MIN=" << rup[0] << endl;
   cout << "MAX=" << rup[1] << endl;
   exit(1);
*/

   // ====== Initialize utilization and bandwidth caches ====================
   const cardinal additionalDepth = 2;
   const cardinal iterations =
      additionalDepth + (cardinal)ceil(log((double)maxPoints) / log(2.0));
   const cardinal maxCachePoints =  (1 << (iterations + 1));

   double utilizationCache[maxCachePoints];
   cardinal bandwidthCache[maxCachePoints];
   for(cardinal i = 2;i < maxCachePoints;i++) {
      utilizationCache[i] = -HUGE_VAL;
      bandwidthCache[i]   = (cardinal)-1;
   }
   utilizationCache[0]                  = rup[0].Utilization;
   utilizationCache[maxCachePoints - 1] = rup[1].Utilization;
   bandwidthCache[0]                    = rup[0].Bandwidth;
   bandwidthCache[maxCachePoints - 1]   = rup[1].Bandwidth;
   cardinal count = 2;


   // ====== Execute recursive algorithm ====================================
   for(cardinal i = 0;count < maxPoints;i++) {
      doResourceUtilizationIteration(rup,
         bwThreshold, utThreshold,
         (double*)&utilizationCache, (cardinal*)&bandwidthCache,
         maxPoints,
         maxCachePoints,
         0, maxCachePoints - 1,
         minBandwidth, maxBandwidth,
         0, i, count);

      if(i >= iterations) {
         break;
      }
   }

   // ====== Verify result ==================================================
#ifdef DEBUG
   for(cardinal i = 0;i < count;i++) {
      if((rup[i].Utilization < 0.0) || (rup[i].Utilization > 1.0)) {
         cerr << "ERROR: AbstractQoSDescription::calculateResourceUtilizationList() - "
                 "Resulting utilization out of range [0,1]!" << endl;
         exit(1);
      }
   }
#endif

   return(count);
}


// ##########################################################################
// #### Output operator                                                  ####
// ##########################################################################



// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const AbstractQoSDescription& aqd)
{
   // ====== Get resources ==================================================
   const double frameRate = aqd.getFrameRate() ;
   ResourceUtilizationPoint rup;
   aqd.getResources(rup);


   // ====== Print AbstractQoSDescription ===================================
   os << "Bandwidth        = " << aqd.getMinBandwidth() << " <= "
                               << rup.Bandwidth << " <= "
                               << aqd.getMaxBandwidth()   << endl;
   os << "TotalUtilization = " << rup.Utilization * 100.0 << " [%]" << endl;
   os << "Layers           = " << aqd.getLayers() << endl;


   // ====== Print FrameRateScalability =====================================
   os << "Frame Rate:" << endl;
   os << "   FrameRate        = " << aqd.getMinFrameRate() << " <= "
                               << frameRate             << " <= "
                               << aqd.getMaxFrameRate() << endl;
   os << "   Scalability Cls. = " << aqd.getFrameRateScalabilityClass() << endl;
   os << "   Properties       = ";
   if(aqd.isFrameRateScalable()) os << "<Scalable> "; else os << "<Non-scalable> ";
   os << endl;
   os << "   Scale Factor     = " << aqd.getFrameRateScaleFactor() * 100.0
                               << " [%]" << endl;
   os << "   PrevFrameRate    = " << aqd.getPrevFrameRate() << endl;
   os << "   NextFrameRate    = " << aqd.getNextFrameRate() << endl;


   // ====== Print AbstractLayerDescriptions ================================
   for(cardinal i = 0;i < aqd.getLayers();i++) {
      os << "Layer #" << i << " Frame Size:" << endl;
      const AbstractLayerDescription* ald = aqd.getLayer(i);
      cardinal frameSize = ald->bandwidthToFrameSize(frameRate,ald->getBandwidth());
      if(frameSize > ald->getMaxFrameSize(frameRate)) {
         frameSize = ald->getMaxFrameSize(frameRate);
      }
      os << "   FrameSize        = " << ald->getMinFrameSize(frameRate)  << " <= "
                                     << frameSize << " <= "
                                     << ald->getMaxFrameSize(frameRate)  << "   (Payload: "
                                     << ald->getMinPayloadFrameSizeForDelay(frameRate,ald->getBufferDelay()) << " <= "
                                     << ald->rawToPayload(frameRate,frameSize)    << " <= "
                                     << ald->getMaxPayloadFrameSizeForDelay(frameRate,ald->getBufferDelay()) << ")"
                                     << endl;
      os << "   BytesPerSecond   = " << ald->getBandwidth()           << endl;
      os << "   PacketsPerSecond = " << ald->getPacketRate(frameRate) << " ("
                                     << ald->getPacketCountForSize(frameRate,frameSize)
                                     << " per frame)" << endl;
      os << "   Souce            = " << ald->getSource() << endl;
      os << "   Destination      = " << ald->getDestination() << endl;
      os << "   Scalability Cls. = " << ald->getFrameSizeScalabilityClass() << endl;
      os << "   Properties       = ";
      if(ald->isFrameSizeScalable()) os << "<Scalable> "; else os << "<Non-scalable> ";
      if(ald->isVariableBitrate())   os << "<VBR> ";      else os << "<CBR> ";
      os << endl;
      os << "   Scale Factor     = " << ald->getFrameSizeScaleFactorForSize(frameRate,frameSize) * 100.0
                                     << " [%]" << endl;
      os << "   PrevFrameSize    = " << ald->getPrevFrameSizeForSize(frameRate,frameSize) << " (Payload: "
                                     << ald->rawToPayload(frameRate,ald->getPrevFrameSizeForSize(frameRate,frameSize)) << ")" << endl;
      os << "   NextFrameSize    = " << ald->getNextFrameSizeForSize(frameRate,frameSize) << " (Payload: "
                                     << ald->rawToPayload(frameRate,ald->getNextFrameSizeForSize(frameRate,frameSize)) << ")" << endl;
      os << "   PeakFrameSize    = " << ald->getPeakFrameSizeForSize(frameRate,frameSize) << " (Payload: "
                                     << ald->rawToPayload(frameRate,ald->getPeakFrameSizeForSize(frameRate,frameSize))
                                     << ")" << endl;

      os << "   BufferDelay      = " << ald->getBufferDelay()   << " [Frame(s)]" << endl;
      os << "   MaxTransferDelay = " << ald->getMaxTransferDelay()    << " [ms]" << endl;
      os << "   MaxLossRate      = " << ald->getMaxLossRate() * 100.0 << " [%]"  << endl;
      os << "   MaxJitter        = " << ald->getMaxJitter()           << " [ms]" << endl;
   }
   return(os);
}


}
