// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Bandwidth Manager Implementation                                 ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "bandwidthmanager.h"
#include "streamdescription.h"


namespace Coral {


// Debug mode: Verify utilization in resource/utilization lists
#define DEBUG


// ###### Constructor #######################################################
BandwidthManager::BandwidthManager(ServiceLevelAgreement* sla)
   : TimedThread(13000)
{
   setTimerCorrection(0);
   SLA     = sla;
   Changed = false;
}


// ###### Destructor ########################################################
BandwidthManager::~BandwidthManager()
{
}


// ###### Main loop #########################################################
void BandwidthManager::timerEvent()
{
   doCompleteRemapping();
}


// ###### Add stream to manager #############################################
void BandwidthManager::addStream(ManagedStreamInterface* stream)
{
   StreamDescription* sd = new StreamDescription;
   if(sd != NULL) {
      synchronized();

      sd->QoSDescription = NULL;
      sd->Interface      = stream;
      sd->NextInterval   = 0;
      sd->Quality.reset();

      StreamSet.insert(pair<ManagedStreamInterface*,StreamDescription*>(stream,sd));
      Changed = true;
      unsynchronized();
   }
   else {
      cerr << "ERROR: BandwidthManager::addStream() - Out of memory!" << endl;
   }
}


// ###### Remove stream from manager ########################################
void BandwidthManager::removeStream(ManagedStreamInterface* stream)
{
   synchronized();
   multimap<ManagedStreamInterface*,StreamDescription*>::iterator found =
      StreamSet.find(stream);
   if(found != StreamSet.end()) {
      StreamDescription* sd = found->second;
      if(sd->QoSDescription) {
         delete sd->QoSDescription;
      }
      delete sd;
      StreamSet.erase(stream);
      Changed = true;
   }
   unsynchronized();
}


// ###### Update stream #####################################################
void BandwidthManager::updateStream(ManagedStreamInterface* stream)
{
   cerr << "NOTE: BandwidthManager::updateStream() - Not implemented!"
        << endl;
   return;
}


// ###### Stream's intervall has changed ####################################
bool BandwidthManager::intervalChangeEvent(ManagedStreamInterface* stream, const card64 when)
{
   bool changed = false;
   synchronized();

   multimap<ManagedStreamInterface*,StreamDescription*>::iterator found =
      StreamSet.find(stream);
   if(found != StreamSet.end()) {
      StreamDescription* sd = found->second;
      if((sd->Interface != NULL) && (sd->QoSDescription != NULL)) {
         sd->QoSDescription->setResources(sd->Quality);
         sd->Interface->updateQuality(sd->QoSDescription);
      }
      Changed = true;
      changed = true;
      sd->init(stream);
      sd->NextInterval = when;
   }

   unsynchronized();
   return(changed);
}


// ###### Stream's loss report has been received ############################
void BandwidthManager::lossEvent(ManagedStreamInterface* stream,
                                 const cardinal          layer,
                                 const double            lossRate)
{
   if(lossRate > 0.0) {
      cout << "Loss = " << lossRate << endl;
   }
}






void BandwidthManager::doCompleteRemapping()
{
   // ====== Check, if remapping is necessary ===============================
   synchronized();
   if(Changed == false) {
      unsynchronized();
      return;
   }
   Changed = false;


   // ====== Count streams ==================================================
   // card64 s = getMicroTime();
   cardinal streams = 0;
   multimap<ManagedStreamInterface*,StreamDescription*>::iterator stream = StreamSet.begin();
   while(stream != StreamSet.end()) {
      if(stream->second->QoSDescription != NULL) {
         streams++;
      }
      stream++;
   }


   // ====== Initialize stream descriptions =================================
   StreamDescription* streamDescription[streams];
   stream = StreamSet.begin();
   cardinal i = 0;
   while(stream != StreamSet.end()) {
      streamDescription[i] = stream->second;
      if(streamDescription[i]->QoSDescription != NULL) {
         i++;
      }
      stream++;
   }



   // ====== Initialize available bandwidths ================================
   cardinal totalAvailableBandwidth;
   cardinal classAvailableBandwidthArray[TrafficClassValues::MaxValues];
   totalAvailableBandwidth = SLA->TotalBandwidth;
   for(cardinal i = 0;i < SLA->Classes;i++) {
      classAvailableBandwidthArray[i] = SLA->Class[i].BytesPerSecond;
   }


// ??????????????!!!!!!!!!!!!
   const cardinal maxPoints = 128;
//   const cardinal bwThreshold = 10000;
//   const double   utThreshold = 0.10;


   // ====== Get resource/utilization list ==================================
   ResourceUtilizationPoint* rupList = new ResourceUtilizationPoint[streams * maxPoints];
   cardinal points = 0;
   for(cardinal i = 0;i < streams;i++) {
      ResourceUtilizationPoint rup;
      streamDescription[i]->OldUtilization =
         streamDescription[i]->QoSDescription->getResources(rup);
      rup.Bandwidth = 0;
      for(cardinal j = 0;j < rup.Layers;j++) {
         rup.LayerBandwidthInfo[j].BytesPerSecond   = 0;
         rup.LayerBandwidthInfo[j].PacketsPerSecond = 0;
      }
      streamDescription[i]->Quality = rup;


      // ====== Reset allocated bandwidth to 0 ==============================
      for(cardinal j = 0;j < RTPConstants::RTPMaxQualityLayers;j++) {
         streamDescription[i]->LayerClasses[j]   = 0;
         streamDescription[i]->ClassNumber[j]    = 0;
         streamDescription[i]->ClassBandwidth[j] = 0;
      }


      // ====== Check, if this stream has an RU list ========================
      if(streamDescription[i]->RUEntries > 0) {
         for(cardinal j = 0;j < streamDescription[i]->RUEntries;j++) {
            rupList[points]        = streamDescription[i]->RUList[j];
            rupList[points].TaskID = i;
            points++;
         }

         // ====== Calculate layer -> DiffServ class mapping ================
         if(streamDescription[i]->calculateLayerClassMapping(SLA,streamDescription[i]->RUList) == true) {
            streamDescription[i]->MaximumReached = false;
         }
         else {
            streamDescription[i]->MaximumReached = true;
         }
      }
   }


   // ====== Sort resource/utilization list =================================
   ResourceUtilizationPoint::sortResourceUtilizationList(rupList,0,points - 1);


   // ====== Allocate bandwidth =============================================
   for(cardinal i = 0;i < points;i++) {
      const cardinal taskID = rupList[i].TaskID;

      if(streamDescription[taskID]->MaximumReached == false) {
         const bool ok = streamDescription[taskID]->tryAllocation(
                            totalAvailableBandwidth,
                            (cardinal*)&classAvailableBandwidthArray,
                            &rupList[i]);
         if(ok == false) {
            streamDescription[taskID]->MaximumReached = true;
         }
         else {
            streamDescription[taskID]->Quality = rupList[i];
#ifdef DEBUG
            streamDescription[taskID]->NewUtilization =
               streamDescription[taskID]->QoSDescription->setResources(rupList[i]);
            if(fabs(streamDescription[taskID]->NewUtilization - rupList[i].Utilization) > 0.05) {
               cerr << "WARNING: Utilization from list and calculated value differ!" << endl
                    << "Task ID #" << taskID << ": P" << streamDescription[taskID]->QoSDescription->getPosition() << endl
                    << rupList[i] << endl
                    << "   List Value = " << rupList[i].Utilization << endl
                    << "   Calculated = " << streamDescription[taskID]->NewUtilization << endl;

               exit(0);
            }
#endif
         }
      }
   }


   // ====== Store result to QoS description ================================
   for(cardinal i = 0;i < streams;i++) {
      streamDescription[i]->QoSDescription->setResources(streamDescription[i]->Quality);

cout << streamDescription[i]->Quality << endl;

      for(cardinal j = 0;j < streamDescription[i]->Layers;j++) {
         AbstractLayerDescription* ald = streamDescription[i]->QoSDescription->getLayer(j);

         InternetFlow dest = ald->getDestination();
         dest.setTrafficClass(SLA->Class[streamDescription[i]->ClassNumber[j]].TrafficClass);
         ald->setDestination(dest);

printf("%02x ",SLA->Class[streamDescription[i]->ClassNumber[j]].TrafficClass);
      }
puts("");
fflush(stdout);
   }


   // ====== Free resource/utilization list =================================
   delete rupList;

   // card64 e = getMicroTime();

   unsynchronized();
}


}
