// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### MPEG QoS Description Implementation                              ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "mpegqosdescription.h"


namespace Coral {


// ###### Try to allocate bandwidth to an MPEG layer ########################
bool MPEGQoSDescription::tryAllocation(cardinal& available,
                                       cardinal* requiredArray,
                                       cardinal* bandwidthArray,
                                       cardinal  layers) const
{
   cardinal totalRequired = 0;
   for(cardinal i = 0;i < layers;i++) {
      totalRequired += requiredArray[i];
   }
   if(available >= totalRequired) {
      for(cardinal i = 0;i < layers;i++) {
         bandwidthArray[i] += requiredArray[i];
      }
      available -= totalRequired;
      return(true);
   }
   return(false);
}


// ###### Calculate MPEG's maximum utilization for given total bandwidths ###
void MPEGQoSDescription::calculateMaxUtilizationForBandwidthArray(
                                  const cardinal*           totalBandwidthArray,
                                  ResourceUtilizationPoint* rupArray,
                                  const cardinal            points) const
{
   // ====== Initialize =====================================================
   const cardinal layers = getLayers();
   bool  finished[points];
   for(cardinal i = 0;i < points;i++) {
      rupArray[i].Layers      = layers;
      rupArray[i].Utilization = -1.0;
      rupArray[i].Cost        = 0.0;
      rupArray[i].TaskID      = 0;
      rupArray[i].Bandwidth   = 0;
      for(cardinal j = 0;j < layers;j++) {
         rupArray[i].LayerBandwidthInfo[j].BufferDelay      = Layer[j].getBufferDelay();
         rupArray[i].LayerBandwidthInfo[j].MaxTransferDelay = Layer[j].getMaxTransferDelay();
         rupArray[i].LayerBandwidthInfo[j].MaxLossRate      = Layer[j].getMaxLossRate();
         rupArray[i].LayerBandwidthInfo[j].MaxJitter        = Layer[j].getMaxJitter();
      }
      finished[i] = false;
   }


   // ====== Get minimum bandwidth required =================================
   for(double frameRate = getMinFrameRate();;frameRate = getNextFrameRateForRate(frameRate)) {

      // ====== Initialize iteration ========================================
      cardinal available[points];
      cardinal bandwidth[points][layers];
      double   utilization[points];
      bool     success[points];
      for(cardinal i = 0;i < points;i++) {
         for(cardinal j = 0;j < layers;j++) {
            bandwidth[i][j] = 0;
         }
         utilization[i] = -1.0;
         available[i]   = totalBandwidthArray[i];
         success[i]     = false;
      }


      // ====== First, use MPEG base layer, then extension layers ===========
      for(cardinal level = 0;level < layers;level += 3) {
         const cardinal levelLayers = min(3,layers);


         // ====== Try maximum allocation ===================================
         cardinal maxBandwidth[3];
         bool     allSuccessful = true;
         for(cardinal i = 0;i < levelLayers;i++) {
            maxBandwidth[i] = AbstractLayerDescription::frameSizeToBandwidth(
                                 frameRate, Layer[level + i].getMaxFrameSize(frameRate));
         }


         // ====== Bandwidth sufficient for maximum allocation ==============
         for(cardinal p = 0;p < points;p++) {
            if((!finished[p]) &&
               (tryAllocation(available[p], (cardinal*)&maxBandwidth,
                             (cardinal*)&bandwidth[p][level], levelLayers) == true)) {
               success[p] = true;
            }
            else {
               allSuccessful = false;
            }
         }


         // ====== Bandwidth too low ========================================
         if(allSuccessful == false) {
            // ====== Try minimum allocation ================================
            cardinal minBandwidth[3];
            for(cardinal i = 0;i < levelLayers;i++) {
               minBandwidth[i] = AbstractLayerDescription::frameSizeToBandwidth(
                                    frameRate, Layer[level + i].getMinFrameSize(frameRate));
            }
            for(cardinal p = 0;p < points;p++) {
               if(!finished[p]) {
                  if(tryAllocation(available[p], (cardinal*)&minBandwidth,
                                (cardinal*)&bandwidth[p][level], levelLayers) == true) {
                     success[p] = true;

                     // ====== Divide up remaining bandwidth ================
                     cardinal wanted[3];
                     cardinal totalWanted = 0;
                     for(cardinal i = 0;i < levelLayers;i++) {
                        wanted[i] = maxBandwidth[i] - minBandwidth[i];
                        totalWanted += wanted[i];
                     }

                     const cardinal availableBandwidth = available[p];
                     cardinal portion[3];
                     for(cardinal i = 0;i < levelLayers;i++) {
                        const double fraction  = (double)wanted[i] / (double)totalWanted;
                        portion[i] = (cardinal)floor((double)availableBandwidth * fraction);
                        bandwidth[p][level + i] += portion[i];
                        available[p]            -= portion[i];

                     }
                  }

                  // ====== No bandwidth for extension layer? ===============
                  else {
                     if(level > 0) {
                        success[p]  = true;
                     }
                     finished[p] = true;
                  }
               }
            }
            break;
         }
      }


      // ====== Calculate utilization =======================================
      for(cardinal p = 0;p < points;p++) {
         if(success[p] == true) {
            // ====== Calculate utilization =================================
             utilization[p] = calculateUtilizationForLayerBandwidths(
                                 frameRate, layers,
                                 (cardinal*)&bandwidth[p]);

            // ====== Update RUP, if utilization is highest =================
            if(utilization[p] > rupArray[p].Utilization) {
               rupArray[p].Utilization = utilization[p];
               rupArray[p].FrameRate   = frameRate;
               rupArray[p].Bandwidth   = 0;
               for(cardinal i = 0;i < layers;i++) {
                  rupArray[p].Bandwidth += bandwidth[p][i];
                  rupArray[p].LayerBandwidthInfo[i].BytesPerSecond   = bandwidth[p][i];
                  rupArray[p].LayerBandwidthInfo[i].PacketsPerSecond =
                  Layer[i].frameSizeToPacketRate(
                     rupArray[p].FrameRate,
                     Layer[i].bandwidthToFrameSize(rupArray[p].FrameRate,bandwidth[p][i]));
               }
            }
            finished[p] = false;
         }
         else {
            utilization[p] = -1.0;
            finished[p]    = true;
         }
      }


      // ====== Check, if frame rate is maximum -> exit =====================
      if(frameRate == getMaxFrameRate()) {
         break;
      }
   }
}


}
