// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### RTP Sender Implementation                                        ####
// ####                                                                  ####
// #### Version 1.00  --  September 17, 2000                             ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "rtpsender.h"
#include "rtcpsender.h"
#include "randomizer.h"


#include <signal.h>


namespace Coral {


// Debug mode: Print some debug information
// #define DEBUG


// ###### Constructor #######################################################
RTPSender::RTPSender()
   : TimedThread(1000000)
{
   Encoder      = NULL;
   SenderSocket = NULL;
}


// ###### Constructor #######################################################
RTPSender::RTPSender(const card32      ssrc,
                     EncoderInterface* encoder,
                     Socket*           senderSocket,
                     const cardinal    maxPacketSize
#ifndef USE_TRANSPORTINFO
                     ,BandwidthManager* bwManager = NULL
#endif
                     )
   : TimedThread(1000000)
{
   init(ssrc,encoder,senderSocket,maxPacketSize
#ifndef USE_TRANSPORTINFO
        ,bwManager
#endif
   );
}


// ###### Initialize ########################################################
void RTPSender::init(const card32      ssrc,
                     EncoderInterface* encoder,
                     Socket*           senderSocket,
                     const cardinal    maxPacketSize
#ifndef USE_TRANSPORTINFO
                     ,BandwidthManager* bwManager = NULL
#endif
                     )
{
   Encoder               = encoder;
   SenderSocket          = senderSocket;
#ifndef USE_TRANSPORTINFO
   BandwidthMgr          = bwManager;
#endif
   MaxPacketSize         = maxPacketSize;
   BytesSent             = 0;
   PacketsSent           = 0;
   PayloadPacketsSent    = 0;
   PayloadBytesSent      = 0;
   FramesPerSecond       = 0;
   RenewCounter          = 0;
   Pause                 = false;
   TransmissionError     = false;
   TimeStamp             = getMicroTime();
   SSRC                  = ssrc;
   Randomizer random;

   for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
      // ====== Initialize values for each layer ============================
      SenderSocket->getPeerAddress(Flow[i]);
      SequenceNumber[i]   = random.random16();
   }

#ifdef USE_TRANSPORTINFO

   UnitBytesCheckCounter = 0;
   for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
      UnitBytesSent[i]    = 0;
      UnitPacketsSent[i]  = 0;
      BytesPerSecond[i]   = 0;
      PacketsPerSecond[i] = 0;
   }
   ExtendedTransportInfo ti;
   getTransportInfo(ti);
   setInterval((card64)1000000 / (card64)FramesPerSecond);

#else
   const AbstractQoSDescription* aqd = getQoSDescription(0);
   update(aqd);
   delete aqd;
#endif
}


// ###### Destructor ########################################################
RTPSender::~RTPSender()
{
   stop();
}


#ifdef USE_TRANSPORTINFO

// ###### Get TransportInfo #################################################
void RTPSender::getTransportInfo(ExtendedTransportInfo& transportInfo,
                                 const bool             calculateLevels)
{
   if(Encoder != NULL) {
      synchronized();

      // ====== Get packet and header sizes =================================
      InternetAddress peerAddress;
      SenderSocket->getPeerAddress(peerAddress);
      const cardinal headerSize    = ((peerAddress.isIPv6()) ? IPv6HeaderSize : IPv4HeaderSize) +
                                        UDPHeaderSize +
                                        RTPConstants::RTPDefaultHeaderSize;
      const cardinal maxPacketSize = min(MaxPacketSize, headerSize + RTPConstants::RTPMaxPayloadLimit);

      // ====== Get TransportInfo ===========================================
      Encoder->getTransportInfo(transportInfo,
                                headerSize, maxPacketSize,
                                calculateLevels);

      // ====== Get addresses ===============================================
      InternetAddress localAddress;
      SenderSocket->getSocketAddress(localAddress);
      for(cardinal i = 0;i < min((cardinal)transportInfo.QualityLayers,RTPConstants::RTPMaxQualityLayers);i++) {
         transportInfo.SrcDest[i].Source       = localAddress.getPortableAddress();
         transportInfo.SrcDest[i].Destination  = Flow[i].getPortableAddress();
         transportInfo.SrcDest[i].FlowLabel    = Flow[i].getFlowLabel();
         transportInfo.SrcDest[i].TrafficClass = Flow[i].getTrafficClass();
         transportInfo.SrcDest[i].IsValid      = true;
      }

      update(transportInfo);
      unsynchronized();
   }
}


// ###### Set TransportInfo #################################################
void RTPSender::setTransportInfo(ExtendedTransportInfo& transportInfo,
                                 const bool             calculateLevels)
{
   if(Encoder != NULL) {
      synchronized();
      for(cardinal i = 0;i < min((cardinal)transportInfo.QualityLayers,RTPConstants::RTPMaxQualityLayers);i++) {
         if(transportInfo.SrcDest[i].IsValid == true) {
            Flow[i] = InternetFlow(transportInfo.SrcDest[i].Destination,
                                   transportInfo.SrcDest[i].FlowLabel,
                                   transportInfo.SrcDest[i].TrafficClass);
         }
      }

      InternetAddress peerAddress;
      SenderSocket->getPeerAddress(peerAddress);
      const cardinal headerSize    = ((peerAddress.isIPv6()) ? IPv6HeaderSize : IPv4HeaderSize) +
                                        UDPHeaderSize +
                                        RTPConstants::RTPDefaultHeaderSize;
      const cardinal maxPacketSize = min(MaxPacketSize, headerSize + RTPConstants::RTPMaxPayloadLimit);
      Encoder->setTransportInfo(transportInfo,
                                headerSize, maxPacketSize,
                                calculateLevels);

      update(transportInfo);
      setInterval((card64)1000000 / (card64)FramesPerSecond);
      unsynchronized();
   }
}


// ###### Update bytes, packets and frames per second #######################
void RTPSender::update(const TransportInfo& transportInfo)
{
   bool     changed = false;
   cardinal i;

   // ====== Get bytes per second ===========================================
   for(i = 0;i < min((cardinal)transportInfo.CurrentSetting.QualityLayers,RTPConstants::RTPMaxQualityLayers);i++) {
      if(BytesPerSecond[i] != transportInfo.CurrentSetting.QualityLayer[i].BytesPerSecond) {
         changed = true;
      }
      BytesPerSecond[i] = transportInfo.CurrentSetting.QualityLayer[i].BytesPerSecond;
   }
   for(    ;i < RTPConstants::RTPMaxQualityLayers;i++) {
      BytesPerSecond[i] = 0;
   }

   // ====== Get packets per second =========================================
   for(i = 0;i < min((cardinal)transportInfo.CurrentSetting.QualityLayers,RTPConstants::RTPMaxQualityLayers);i++) {
      if(PacketsPerSecond[i] != transportInfo.CurrentSetting.QualityLayer[i].PacketsPerSecond) {
         changed = true;
      }
      PacketsPerSecond[i] = transportInfo.CurrentSetting.QualityLayer[i].PacketsPerSecond;
   }
   for(    ;i < RTPConstants::RTPMaxQualityLayers;i++) {
      PacketsPerSecond[i] = 0;
   }

   // ====== Get frames per second ==========================================
   FramesPerSecond = transportInfo.CurrentSetting.FramesPerSecond;

   // ====== Reset check values, if necessary ===============================
   if(changed) {
      for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
         UnitBytesSent[i]   = 0;
         UnitPacketsSent[i] = 0;
      }
      UnitBytesCheckCounter = 0;
   }
}


// ###### Quality adaption ##################################################
void RTPSender::adaptQuality(const double fractionLost, const cardinal layer)
{
   if(Encoder != NULL) {
      synchronized();
      Encoder->adaptQuality(fractionLost,layer);

      InternetAddress peerAddress;
      SenderSocket->getPeerAddress(peerAddress);
      const cardinal headerSize    = ((peerAddress.isIPv6()) ? IPv6HeaderSize : IPv4HeaderSize) +
                                        UDPHeaderSize +
                                        RTPConstants::RTPDefaultHeaderSize;
      const cardinal maxPacketSize = min(MaxPacketSize, headerSize + RTPConstants::RTPMaxPayloadLimit);
      TransportInfo transportInfo;
      Encoder->getTransportInfo(transportInfo,
                                headerSize, maxPacketSize);
      update(transportInfo);
      setInterval((card64)1000000 / (card64)FramesPerSecond);

      unsynchronized();
   }
}


#else

// ###### Get QoS description ###############################################
AbstractQoSDescription* RTPSender::getQoSDescription(const card64 offset)
{
   if(Encoder != NULL) {
      synchronized();

      // ====== Get packet and header sizes =================================
      InternetAddress peerAddress;
      SenderSocket->getPeerAddress(peerAddress);
      const cardinal headerSize    = ((peerAddress.isIPv6()) ? IPv6HeaderSize : IPv4HeaderSize) +
                                        UDPHeaderSize +
                                        RTPConstants::RTPDefaultHeaderSize;
      const cardinal maxPacketSize = min(MaxPacketSize, headerSize + RTPConstants::RTPMaxPayloadLimit);

      // ====== Get QoS description =========================================
      AbstractQoSDescription* aqd =
         Encoder->getQoSDescription(headerSize,maxPacketSize,offset);

      // ====== Set addresses ===============================================
      if(aqd != NULL) {
         InternetAddress localAddress;
         SenderSocket->getSocketAddress(localAddress);
         const cardinal layers = min(aqd->getLayers(),RTPConstants::RTPMaxQualityLayers);
         for(cardinal i = 0;i < layers;i++) {
            AbstractLayerDescription* ald = aqd->getLayer(i);
            ald->setSource(localAddress);
            ald->setDestination(Flow[i]);
         }
         update(aqd);
      }

      unsynchronized();
      return(aqd);
   }
   return(NULL);
}


// ###### Update QoS description ############################################
void RTPSender::updateQuality(const AbstractQoSDescription* aqd)
{
   if(Encoder != NULL) {
      synchronized();

      const cardinal layers = min(aqd->getLayers(),RTPConstants::RTPMaxQualityLayers);
      for(cardinal i = 0;i < layers;i++) {
         AbstractLayerDescription* ald = aqd->getLayer(i);
         Flow[i] = ald->getDestination();
      }

      Encoder->updateQuality(aqd);
      update(aqd);

      unsynchronized();
   }
}


// ###### Update frame rate #################################################
void RTPSender::update(const AbstractQoSDescription* aqd)
{
   const double frameRate = aqd->getFrameRate();
   if(frameRate <= 1.0) {
      setInterval(1000000);
      FramesPerSecond = 1;
   }
   else {
      setInterval((card64)(1000000.0 / frameRate));
      FramesPerSecond = (cardinal)ceil(1000000.0 / frameRate);
   }
}


// ###### Lock sender #######################################################
void RTPSender::lock()
{
   synchronized();
}


// ###### Unlock sender #####################################################
void RTPSender::unlock()
{
   unsynchronized();
}

#endif


// ###### Sender loop #######################################################
void RTPSender::timerEvent()
{
   if(Encoder == NULL) {
      cerr << "ERROR: RTPSender::timerEvent() - RTPSender is uninitialized!"
           << endl;
      return;
   }
   synchronized();


   // ==== Renew flow label reservation and send sender report every second =
   RenewCounter++;
   if(RenewCounter >= FramesPerSecond) {
      // ====== Renew flow label ============================================
      SenderSocket->renewFlow(10);
      for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
         SenderSocket->renewFlow(Flow[i],10);
      }
      RenewCounter = 0;

      // ====== Send RTCP Sender Report =====================================
      RTCPSenderReport report(SSRC,0);
      const card64 now = getMicroTime();
      report.setNTPTimeStamp(now);
      report.setRTPTimeStamp(
         (card32)rint((double)(now - TimeStamp) /
                      RTPConstants::RTPMicroSecondsPerTimeStamp) & 0xffffffff);
      report.setOctetsSent(PayloadBytesSent);
      report.setPacketsSent(PayloadPacketsSent);
      if(SenderSocket->send(&report,sizeof(RTCPSenderReport) != sizeof(RTCPSenderReport))) {
#ifdef DEBUG
         cerr << "RTPSender::timerEvent() - Transmission of RTCP SR failed!" << endl;
#endif
         TransmissionError = true;
      }
   }


#ifdef USE_TRANSPORTINFO
   // ====== Reset byte rate of last second =================================
   UnitBytesCheckCounter++;
   if(UnitBytesCheckCounter >= FramesPerSecond) {
      for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
         UnitBytesSent[i]   = 0;
         UnitPacketsSent[i] = 0;
      }
      UnitBytesCheckCounter = 0;
   }
#endif


   // ====== Transmit next frame ============================================
   RTPPacket packet;
   InternetAddress peerAddress;
   SenderSocket->getPeerAddress(peerAddress);
   const cardinal headerSizeTransport =
      ((peerAddress.isIPv6()) ? IPv6HeaderSize : IPv4HeaderSize) +
      UDPHeaderSize;
   const cardinal headerSizeRTP = packet.calculateHeaderSize();
   const cardinal maxPacketSize =
      min(MaxPacketSize, headerSizeTransport + headerSizeRTP + RTPConstants::RTPMaxPayloadLimit);
   const cardinal maxPayloadSize =
      maxPacketSize - (headerSizeTransport + headerSizeRTP);


#ifndef USE_TRANSPORTINFO
   if(!Pause) {
      const card64 nextInterval = Encoder->checkInterval();

      // ====== Check for new interval ======================================
      if(BandwidthMgr != NULL) {
         BandwidthMgr->intervalChangeEvent(this,nextInterval);
      }
   }
#endif


   // ====== Prepare next frame for sending =================================
   if((!Pause) && (Encoder->prepareNextFrame(headerSizeTransport + headerSizeRTP,
                                             maxPacketSize) == true)) {

      // ====== Get packet and transmit packet loop =========================
      cardinal bytesData = 0;
      for(;;) {
         // ====== Get next packet from encoder =============================
         EncoderPacket encoderPacket;
         encoderPacket.Buffer      = packet.getPayloadData();
         encoderPacket.MaxLength   = maxPayloadSize;
         encoderPacket.Layer       = 0;
         encoderPacket.Marker      = false;
         encoderPacket.PayloadType = 0x00;
         bytesData = Encoder->getNextPacket(&encoderPacket);
         if(bytesData > maxPayloadSize) {
            cerr << "WARNING: RTPSender::timerEvent() - Encoder exceeds packet size limit!" << endl;
            bytesData = 0;
         }


         // ====== Check, if frame has been transmitted completely ==========
         if(bytesData > 0) {
#ifdef USE_TRANSPORTINFO
            // ====== Check layer number ====================================
            if(encoderPacket.Layer >= RTPConstants::RTPMaxQualityLayers) {
               cerr << "WARNING: RTPSender::timerEvent() - Invalid layer number!" << endl;
               break;
            }

            // ====== Check, if quality settings have been updated ==========
            if(((headerSizeTransport + headerSizeRTP + bytesData +
                 UnitBytesSent[encoderPacket.Layer]) > BytesPerSecond[encoderPacket.Layer]) ||
              ((UnitPacketsSent[encoderPacket.Layer] + 1) > PacketsPerSecond[encoderPacket.Layer])) {
               ExtendedTransportInfo ti;
               getTransportInfo(ti,false);

               // ====== Check for exceeded byte rate limit ====================
               if((headerSizeTransport + headerSizeRTP + bytesData +
                   UnitBytesSent[encoderPacket.Layer]) > BytesPerSecond[encoderPacket.Layer]) {
                  cerr << "WARNING: RTPSender::timerEvent() - Byte rate exceeds limit:"  << endl;
                  cerr << "   Layer #" << encoderPacket.Layer << ": "
                       << UnitBytesSent[encoderPacket.Layer] << " + "
                       << (headerSizeTransport + headerSizeRTP + bytesData) << " > "
                       << BytesPerSecond[encoderPacket.Layer] << "!" << endl;
                  card64 sum = 0;
                  for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
                     cerr << "Sent over layer #" << i << ": " << UnitBytesSent[i]
                          << ", limit is " << BytesPerSecond[i] << "." << endl;
                     sum += BytesPerSecond[i];
                  }
                  cerr << "Total sent: " << sum << endl;
                  break;
               }

               // ====== Check for exceeded packet rate limit ==================
               if((UnitPacketsSent[encoderPacket.Layer] + 1) > PacketsPerSecond[encoderPacket.Layer]) {
                  cerr << "WARNING: RTPSender::timerEvent() - Packet rate exceeds limit:"  << endl;
                  cerr << "   Layer #" << encoderPacket.Layer << ": "
                       << UnitPacketsSent[encoderPacket.Layer] << " + 1 > "
                       << PacketsPerSecond[encoderPacket.Layer] << "!" << endl;
                  card32 sum = 0;
                  for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
                     cerr << "Sent over layer #" << i << ": " << UnitPacketsSent[i]
                          << ", limit is " << PacketsPerSecond[i] << "." << endl;
                     sum += PacketsPerSecond[i];
                  }
                  cerr << "Total sent: " << sum << endl;
                  break;
               }
            }
#endif


            // ====== Send packet ===========================================
            packet.setMarker(encoderPacket.Marker);
            packet.setPayloadType(encoderPacket.PayloadType);
            packet.setSequenceNumber(SequenceNumber[encoderPacket.Layer]++);
            packet.setSSRC(SSRC);
            packet.setTimeStamp((card32)rint((double)(getMicroTime() - TimeStamp) /
                                RTPConstants::RTPMicroSecondsPerTimeStamp) & 0xffffffff);
            const ssize_t sent = SenderSocket->sendTo(&packet,
                                    packet.calculateHeaderSize() + bytesData,
                                    0,
                                    Flow[encoderPacket.Layer],
                                    Flow[encoderPacket.Layer].getTrafficClass());
            if(sent > 0) {
               PayloadBytesSent += (card32)sent;
               PayloadPacketsSent++;
               BytesSent += sent + headerSizeTransport;
               PacketsSent++;
#ifdef USE_TRANSPORTINFO
               UnitBytesSent[encoderPacket.Layer] += sent + headerSizeTransport;
               UnitPacketsSent[encoderPacket.Layer]++;
#endif
            }
            else {
#ifdef DEBUG
               cerr << "WARNING: RTPSender::timerEvent() - "
                    << "Unable to send " << packet.calculateHeaderSize() + bytesData
                    << " bytes to " << Flow[encoderPacket.Layer] << endl;
#endif
               TransmissionError = true;
               break;
            }
         }
         else {
            break;
         }

         // ====== Check for pending signal =================================
         if(pendingTimerEvent()) {
            // There is a new timer event -> Skip rest of this frame!
#ifdef DEBUG
            cerr << "NOTE: RTPSender::timerEvent() - Pending timer -> "
                 << "Skipping rest of frame!" << endl;
#endif
            break;
         }
      }
   }

   unsynchronized();
}


}
