// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### TDTF Media Reader Implementation                                 ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "tdtfmediareader.h"


namespace Coral {


// ###### Constructor #######################################################
TDTFMediaReader::TDTFMediaReader()
{
   FrameRate   = 0.0;
   Position    = 0;
   MaxPosition = 0;
   Interval    = NULL;
   ErrorCode   = ME_NoMedia;
}


// ###### Open ##############################################################
bool TDTFMediaReader::open(const char* name,
                           const bool  readWrite,
                           const bool  ruRequired)
{
   bool ok = TDTFReader::open(name,readWrite);
   if(ok == true) {
      if((!ruRequired) || (hasResourceUtilizationLists())) {
         Interval    = NULL;
         FrameRate   = MaxFrameRate;
         Position    = 0;
         MaxPosition = framePositionToPosition(MaxFrameRate,getFrames(MaxFrameRate));
         if(MaxPosition < 1000000000) {
            close();
            return(false);
         }
         MaxPosition -= 1000000000;
         ErrorCode   = ME_NoError;
      }
      else {
         close();
         ErrorCode = ME_BadMedia;
         cerr << "WARNING: TDTFMediaReader::open() - File <"
                  << name << "> does not contain resource/utilization lists!" << endl;
         ok = false;
      }
   }
   else {
      ErrorCode = ME_BadMedia;
   }
   return(ok);
}


// ###### Close #############################################################
void TDTFMediaReader::close()
{
   TDTFReader::close();
   Interval    = NULL;
   FrameRate   = 0.0;
   Position    = 0;
   MaxPosition = 0;
   ErrorCode   = ME_NoError;
}


// ###### Get MediaInfo #####################################################
void TDTFMediaReader::getMediaInfo(MediaInfo& mediaInfo) const
{
   mediaInfo.reset();
   if(ready()) {
      mediaInfo.StartTimeStamp = 0;
      mediaInfo.EndTimeStamp   = MaxPosition;
      const TDTFPrefix* tp = getTDTFPrefix();
      if(tp != NULL) {
         snprintf((char*)&mediaInfo.Title,sizeof(mediaInfo.Title),
                  "%s",tp->Title);
         snprintf((char*)&mediaInfo.Artist,sizeof(mediaInfo.Artist),
                  "%s",tp->Copyright);
         snprintf((char*)&mediaInfo.Comment,sizeof(mediaInfo.Comment),
                  "%s",tp->Comment);
      }
   }
}


// ###### Check, if new interval is reached #################################
card64 TDTFMediaReader::checkInterval()
{
   // ====== Get interval header ============================================
   if(MainIndex == NULL) {
      return((card64)-1);
   }
   const cardinal framePosition = positionToFramePosition(FrameRate,getPosition());
   const IntervalHeader* ih = NULL;
   if(framePosition < getFrames(FrameRate)) {
      ih = getIntervalHeader(framePosition,FrameRate);
   }

   const card64 nextPosition = framePositionToPosition(FrameRate,translate32(ih->Position) + translate32(ih->Length));
   if((int64)nextPosition - (int64)getPosition() < 0) {
      cerr << "ERROR: Negative interval size?!" << endl;
      exit(1);
   }

   return(nextPosition - getPosition());
}


// ###### Read next trace block #############################################
cardinal TDTFMediaReader::getNextBlock(const cardinal layers,
                                       cardinal*      frameIDArray,
                                       cardinal*      frameSizeArray)
{
   // ====== Prepare next frame, if not end of trace ========================
   if(Position < MaxPosition) {
      // ====== Get frame data ==============================================
      const cardinal mediaLayers = min(layers,getLayers(FrameRate));

      for(cardinal i = 0;i < mediaLayers;i++) {
         frameSizeArray[i] = getTraceFrameSize(Position,i,FrameRate);
         frameIDArray[i]   = getTraceFrameID(Position,i,FrameRate);
      }

      Position += (card64)ceil((double)PositionStepsPerSecond / FrameRate);
      Position = framePositionToPosition(MaxFrameRate,positionToFramePosition(MaxFrameRate,Position));
      Position = min(Position,MaxPosition);

      return(mediaLayers);
   }

   // ====== End of trace reached ===========================================
   return(0);
}


}
