// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### TDTF Reader Implementation                                       ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "tdtfreader.h"


#include <sys/mman.h>
#include <fcntl.h>
#include <limits.h>


namespace Coral {


// ###### Constructor #######################################################
TDTFReader::TDTFReader()
{
   InputFile   = -1;
   InputMemory = NULL;

   // Use close() method to reset attributes!
   close();
}


// ###### Destructor ########################################################
TDTFReader::~TDTFReader()
{
   close();
}


// ###### Check, if trace file access is within valid range #################
bool TDTFReader::checkAccess(const cardinal position, const cardinal length) const
{
   if(position + length <= InputLength) {
      return(true);
   }

   cerr << "WARNING: TDTFReader::checkAccess() - "
           "Illegal access to " << position << ", length " << length << "!" << endl;
   return(false);
}


// ###### Close TDTF trace file #############################################
void TDTFReader::close()
{
   if(InputMemory != NULL) {
      const int ok = munmap(InputMemory,InputLength);
      if(ok != 0) {
         cerr << "ERROR: TDTFReader::close() - munmap() failed!" << endl;
         exit(1);
      }
      InputMemory = NULL;
   }
   if(InputFile != -1) {
      ::close(InputFile);
      InputFile = -1;
   }
   InputLength      = 0;
   MainIndex        = NULL;
   MainIndexEntries = 0;
   RULIndex         = NULL;
   RULIndexEntries  = 0;
   MinFrameRate     = 0.0;
   MaxFrameRate     = 0.0;
}


// ###### Open TDTF trace file ##############################################
bool TDTFReader::open(const char* name, const bool readWrite)
{
   // ====== Open file ======================================================
   close();
   InputFile = ::open(name,(readWrite == true) ? O_RDWR : O_RDONLY);
   if(InputFile <= 0) {
      cerr << "ERROR: TDTFReader::open() - Unable to open <"
           << name << ">!" << endl;
      return(false);
   }
   off_t len = lseek(InputFile,0,SEEK_END);
   if(len < 0) {
      cerr << "ERROR: TDTFReader::open() - lseek() failed!" << endl;
      return(false);
   }
   InputLength = (cardinal)len;


   // ====== Map file to memory =============================================
   InputMemory = (char*)mmap(NULL, InputLength,
                             (readWrite == true) ? PROT_READ|PROT_WRITE : PROT_READ,
                             MAP_SHARED,
                             InputFile, 0);
   if(InputMemory == NULL) {
      cerr << "ERROR: TDTFReader::open() - mmap() failed!" << endl;
      return(false);
   }


   // ====== Check format ===================================================
   const TDTFPrefix* tp = (TDTFPrefix*)&InputMemory[0];
   if((tp->TDTF_ID[0] != 'T') && (tp->TDTF_ID[1] != 'D') && (tp->TDTF_ID[2] != 'T') && (tp->TDTF_ID[3] != 'F')) {
      cerr << "ERROR: TDTFReader::open() - No TDTF header!" << endl;
      return(false);
   }
   if(translate32(tp->Version) != TDTFPrefix::CurrentVersion) {
      cerr << "ERROR: TDTFReader::open() - Incompatible trace version!" << endl;
      return(false);
   }
   const TDTFSuffix* ts = (TDTFSuffix*)&InputMemory[InputLength - sizeof(TDTFSuffix)];
   if((ts->TDTF_ID[0] != 'T') && (ts->TDTF_ID[1] != 'D') && (ts->TDTF_ID[2] != 'T') && (ts->TDTF_ID[3] != 'F')) {
      cerr << "ERROR: TDTFReader::open() - Corrupted TDTF file -- Missing TDTF suffix!" << endl;
      return(false);
   }


   // ====== Read main index ================================================
   const cardinal indexMainIndex = translate32(ts->MainIndex);
   if(checkAccess(translate32(ts->MainIndex),sizeof(MainIndexHeader)) == false) {
      return(false);
   }
   MainIndex        = (MainIndexHeader*)&InputMemory[indexMainIndex];
   MainIndexEntries = translate32(MainIndex->Entries);
   if(checkAccess(indexMainIndex,MainIndexEntries * sizeof(MainIndexEntry)) == false) {
      cerr << "ERROR: TDTFReader::open() - Bad main index!" << endl;
      return(false);
   }
   const cardinal indexRULIndex = translate32(ts->ResourceUtilizationIndex);
   if(indexRULIndex != 0) {
      RULIndex        = (ResourceUtilizationListIndexHeader*)&InputMemory[indexRULIndex];
      RULIndexEntries = translate32(RULIndex->Entries);
      if(checkAccess(indexRULIndex,RULIndexEntries * sizeof(ResourceUtilizationListIndexEntry)) == false) {
         cerr << "ERROR: TDTFReader::open() - Bad RU list index!" << endl;
         return(false);
      }
   }


   // ====== Initialize =====================================================
   // Note: Frame rates are in *de*creasing order!
   MinFrameRate = translateToDouble(MainIndex->Entry[MainIndexEntries - 1].FrameRate);
   MaxFrameRate = translateToDouble(MainIndex->Entry[0].FrameRate);


   // ====== Do some integrity checks =======================================
   for(cardinal i = 0;i < MainIndexEntries;i++) {
      if(checkAccess(translate32(MainIndex->Entry[i].Trace),sizeof(TraceHeader)) == false) {
         cerr << "ERROR: TDTFReader::open() - Bad main index reference to trace!" << endl;
         return(false);
      }
      if(checkAccess(translate32(MainIndex->Entry[i].Intervals),sizeof(PositionLengthIntervalIndexHeader)) == false) {
         cerr << "ERROR: TDTFReader::open() - Bad main index reference to interval!" << endl;
         return(false);
      }
   }

   return(true);
}


// ###### Check, if frame rate is valid #####################################
bool TDTFReader::isValidFrameRate(const double frameRate) const
{
   if(MainIndex == NULL) {
      return(false);
   }
   for(cardinal i = 0;i < MainIndexEntries;i++) {
      const double f = translateToDouble(MainIndex->Entry[i].FrameRate);
      if(f == frameRate) {
         return(true);
      }
   }
   return(false);
}


// ###### Get nearest lower valid frame rate ################################
double TDTFReader::getNearestValidFrameRate(const double frameRate) const
{
   // ====== Check limits ===================================================
   if(MainIndex == NULL) {
      return(0.0);
   }
   else if(frameRate <= MinFrameRate) {
      return(MinFrameRate);
   }
   else if(frameRate >= MaxFrameRate) {
      return(MaxFrameRate);
   }

   // ====== Search for frame rate ==========================================
   // Note: Frame rates are in *de*creasing order!
   for(cardinal i = 0;i < MainIndexEntries;i++) {
      const double f = translateToDouble(MainIndex->Entry[i].FrameRate);
      if(f <= frameRate) {
         return(f);
      }
   }
   return(MinFrameRate);
}


// ###### Get next higher frame rate ########################################
double TDTFReader::getNextFrameRateForRate(const double frameRate) const
{
   // ====== Check limits ===================================================
   if(MainIndex == NULL) {
      return(0.0);
   }
   else if(frameRate >= MaxFrameRate) {
      return(MaxFrameRate);
   }
   else if(frameRate < MinFrameRate) {
      return(MinFrameRate);
   }

   // ====== Search for frame rate ==========================================
   // Note: Frame rates are in *de*creasing order!
   for(integer i = (integer)MainIndexEntries - 1;i >= 0;i--) {
      const double f = translateToDouble(MainIndex->Entry[i].FrameRate);
      if(f > frameRate) {
         return(f);
      }
   }
   return(MaxFrameRate);
}


// ###### Get next lower frame rate #########################################
double TDTFReader::getPrevFrameRateForRate(const double frameRate) const
{
   // ====== Check limits ===================================================
   if(MainIndex == NULL) {
      return(0.0);
   }
   else if(frameRate > MaxFrameRate) {
      return(MaxFrameRate);
   }
   else if(frameRate <= MinFrameRate) {
      return(MinFrameRate);
   }

   // ====== Search for frame rate ==========================================
   // Note: Frame rates are in *de*creasing order!
   for(cardinal i = 0;i < MainIndexEntries;i++) {
      const double f = translateToDouble(MainIndex->Entry[i].FrameRate);
      if(f < frameRate) {
         return(f);
      }
   }
   return(MinFrameRate);
}



// ###### Get main index entry ##############################################
const MainIndexEntry* TDTFReader::getMainIndexEntry(const double frameRate) const
{
   if(MainIndex == NULL) {
      return(NULL);
   }
   cardinal start = 0;
   cardinal end   = MainIndexEntries - 1;

   // ====== Check limits ===================================================
   if(frameRate <= translateToDouble(MainIndex->Entry[end].FrameRate)) {
      return(&MainIndex->Entry[end]);
   }
   else if(frameRate >= translateToDouble(MainIndex->Entry[0].FrameRate)) {
      return(&MainIndex->Entry[0]);
   }

   // ====== Do binary search ===============================================
   else {
      cardinal m = 0;
      do {
         m = (start + end) / 2;
         const double f = translateToDouble(MainIndex->Entry[m].FrameRate);
         if(frameRate > f) {
            end = m;
         }
         else if(frameRate < f) {
            start = m;
         }
         else {
            return(&MainIndex->Entry[m]);
         }
      } while(end - start > 1);
   }
   return(NULL);
}


// ###### Get TDTF prefix ###################################################
const TDTFPrefix* TDTFReader::getTDTFPrefix() const
{
   if(InputMemory != NULL) {
      return((TDTFPrefix*)InputMemory);
   }
   return(NULL);
}


// ###### Get TDTF suffix ###################################################
const TDTFSuffix* TDTFReader::getTDTFSuffix() const
{
   if(InputMemory != NULL) {
      return((TDTFSuffix*)&InputMemory[InputLength - sizeof(TDTFSuffix)]);
   }
   return(NULL);
}


// ###### Get trace header ##################################################
const TraceHeader* TDTFReader::getTraceHeader(const double frameRate) const
{
   // ====== Get main index entry ===========================================
   const MainIndexEntry* ie = getMainIndexEntry(frameRate);
   if(ie == NULL) {
      return(NULL);
   }
   const cardinal indexTrace = translate32(ie->Trace);

   // ====== Check reference integrity ======================================
   const TraceHeader* th = (const TraceHeader*)&InputMemory[indexTrace];
   if(th == NULL) {
      return(NULL);
   }
   if(checkAccess(indexTrace,sizeof(TraceHeader)) == false) {
      return(NULL);
   }
   if(checkAccess(indexTrace,sizeof(TraceHeader) +
                     translate32(th->Frames) * th->Layers * sizeof(FrameDescription)) == false) {
      return(NULL);
   }
   return(th);
}


// ###### Get PositionLengthInterval index header ###########################
const PositionLengthIntervalIndexHeader* TDTFReader::getIPLIHeader(
                                            const double frameRate) const
{
   // ====== Get main index entry ===========================================
   const MainIndexEntry* ie = getMainIndexEntry(frameRate);
   if(ie == NULL) {
      return(NULL);
   }
   const cardinal indexIPLI = translate32(ie->Intervals);

   // ====== Check reference integrity ======================================
   const PositionLengthIntervalIndexHeader* ipli = (const PositionLengthIntervalIndexHeader*)&InputMemory[indexIPLI];
   if(ipli == NULL) {
      return(NULL);
   }
   if(checkAccess(indexIPLI,sizeof(PositionLengthIntervalIndexHeader)) == false) {
      return(NULL);
   }
   if(checkAccess(indexIPLI,sizeof(PositionLengthIntervalIndexHeader) +
                     translate32(ipli->Entries) * sizeof(PositionLengthIntervalIndexEntry)) == false) {
      return(NULL);
   }
   return(ipli);
}


// ###### Get interval header ###############################################
const IntervalHeader* TDTFReader::getIntervalHeader(
                                      const cardinal position,
                                      const double   frameRate) const
{
   // ====== Get PositionLengthInterval index header ========================
   const PositionLengthIntervalIndexHeader* ipli = getIPLIHeader(frameRate);
   if(ipli == NULL) {
      return(NULL);
   }
   cardinal indexInterval = (cardinal)-1;
   cardinal start         = 0;
   cardinal end           = translate32(ipli->Entries) - 1;


   // ====== Check limits ===================================================
   if(position <= translate32(ipli->Entry[0].Position)) {
      indexInterval = translate32(ipli->Entry[0].Interval);
   }
   else if(position >= translate32(ipli->Entry[end].Position)) {
      indexInterval = translate32(ipli->Entry[end].Interval);
   }


   // ====== Do binary search ===============================================
   else {
      cardinal m = 0;
      do {
         m = (start + end) / 2;
         const cardinal p0 = translate32(ipli->Entry[m].Position);
         const cardinal p1 = p0 + translate32(ipli->Entry[m].Length) - 1;
         if(position < p0) {
            end = m;
         }
         else if(position > p1) {
            start = m;
         }
         else {
            start = m;
            end   = m;
         }
      } while(end - start > 1);
      indexInterval = translate32(ipli->Entry[start].Interval);
   }


   // ====== Check reference integrity ================================
   const IntervalHeader* ih = (const IntervalHeader*)&InputMemory[indexInterval];
   if(checkAccess(indexInterval,sizeof(IntervalHeader)) == false) {
      return(NULL);
   }
   if(checkAccess(indexInterval,sizeof(PositionLengthIntervalIndexHeader) +
                                   translate32(ih->IntervalDescriptionSize)) == false) {
      return(NULL);
   }
   if((position < translate32(ih->Position)) ||
      (position > translate32(ih->Position) + translate32(ih->Length) - 1)) {
      cerr << "WARNING: TDTFReader::getIntervalHeader() - Position check failed!" << endl;
      printf("%d  p=%d l=%d\n",position,translate32(ih->Position),translate32(ih->Length));
      return(NULL);
   }
   return(ih);
}


// ###### Get layer header ##################################################
const LayerHeader* TDTFReader::getLayerHeader(
                                   const cardinal position,
                                   const cardinal layer,
                                   const double   frameRate) const
{
   // ====== Get Interval index header ======================================
   const IntervalHeader* ih = getIntervalHeader(position,frameRate);
   if(ih == NULL) {
      cerr << "WARNING: TDTFReader::getLayerHeader() - Invalid parameters "
           << "P=" << position << ", FR=" << frameRate << "!" << endl;
      return(NULL);
   }


   // ====== Get layer header ===============================================
   if(layer >= ih->Layers) {
      cerr << "WARNING: TDTFReader::getLayerHeader() - Invalid parameter "
           << "L=" << layer << endl;
      return(NULL);
   }
   const cardinal offset   = ih->getOffsetLH(layer);
   const cardinal lhOffset = (cardinal)translate32(ih->Offset[offset]);
   const LayerHeader* lh   = (const LayerHeader*)((long)ih + (long)lhOffset);

   // ====== Check reference integrity ================================
   if(checkAccess((long)lh - (long)InputMemory,sizeof(LayerHeader)) == false) {
      return(NULL);
   }

   return(lh);
}


// ###### Get empirical envelope ############################################
const EmpiricalEnvelope* TDTFReader::getEmpiricalEnvelope(
                                         const cardinal position,
                                         const cardinal layer,
                                         const double   frameRate,
                                         const bool     frameCount) const
{
   // ====== Get Interval index header ======================================
   const IntervalHeader* ih = getIntervalHeader(position,frameRate);
   if(ih == NULL) {
      cerr << "ERROR: TDTFReader::getEmpiricalEnvelope() - Invalid parameters "
           << "P=" << position << ", FR=" << frameRate << "!" << endl;
      return(NULL);
   }

   // ====== Get empirical envelope =========================================
   if(layer >= ih->Layers) {
      cerr << "ERROR: TDTFReader::getEmpiricalEnvelope() - Invalid parameter "
           << "L=" << layer << endl;
      return(NULL);
   }
   cardinal offset;
   if(frameCount) {
      offset = ih->getOffsetEEFC(layer);
   }
   else {
      offset = ih->getOffsetEEBR(layer);
   }
   const cardinal eeOffset = (cardinal)translate32(ih->Offset[offset]);
   const EmpiricalEnvelope* ee = (const EmpiricalEnvelope*)((long)ih + (long)eeOffset);

   // ====== Check reference integrity ================================
   if(checkAccess((long)ee - (long)InputMemory,sizeof(EmpiricalEnvelope)) == false) {
      return(NULL);
   }
   if(checkAccess((long)ee - (long)InputMemory,sizeof(EmpiricalEnvelope) +
                                translate16(ee->Pairs) * sizeof(EmpiricalEnvelopePair)) == false) {
      return(NULL);
   }
   return(ee);
}


// ###### Get maximum byte count ############################################
cardinal TDTFReader::getMaxByteCountForDelay(
                         const card64   rtpPosition,
                         const cardinal layer,
                         const double   frameRate,
                         const cardinal bufferDelay) const
{
   // ====== Get empirical envelope =========================================
   const cardinal position = positionToFramePosition(frameRate,rtpPosition);
   if(position >= getFrames(frameRate)) {
      return(0);
   }
   const EmpiricalEnvelope* ee = getEmpiricalEnvelope(position,layer,frameRate);
   if(ee == NULL) {
      return(0);
   }
   return(ee->getConstraint(bufferDelay));
}


// ###### Get maximum frame count ###########################################
cardinal TDTFReader::getMaxFrameCountForDelay(const card64   rtpPosition,
                                              const cardinal layer,
                                              const double   frameRate,
                                              const cardinal bufferDelay) const
{
   // ====== Get empirical envelope =========================================
   const cardinal position = positionToFramePosition(frameRate,rtpPosition);
   if(position >= getFrames(frameRate)) {
      return(0);
   }
   const EmpiricalEnvelope* fc = getEmpiricalEnvelope(position,layer,frameRate,true);
   if(fc == NULL) {
      return(0);
   }
   return(fc->getConstraint(bufferDelay));
}


// ###### Get layer header ##################################################
double TDTFReader::getLayerScalability(
                      const card64   rtpPosition,
                      const cardinal layer,
                      const double   frameRate) const
{
   // ====== Get layer header ===============================================
   const cardinal position = positionToFramePosition(frameRate,rtpPosition);
   const LayerHeader* lh = getLayerHeader(position,layer,frameRate);
   if(lh == NULL) {
      return(1.0);
   }

   return((double)translate32(lh->Scalability) / pow(2.0,30.0));
}


// ###### Get layer flags ###################################################
cardinal TDTFReader::getLayerFlags(
                        const card64   rtpPosition,
                        const cardinal layer,
                        const double   frameRate) const
{
   // ====== Get layer header ===============================================
   const cardinal position = positionToFramePosition(frameRate,rtpPosition);
   const LayerHeader* lh = getLayerHeader(position,layer,frameRate);
   if(lh == NULL) {
      return(0);
   }
   return(lh->Flags);
}


// ###### Get number of frames ##############################################
cardinal TDTFReader::getFrames(const double frameRate) const
{
   const TraceHeader* trace = getTraceHeader(frameRate);
   if(trace != NULL) {
      return(translate32(trace->Frames));
   }
   return(0);
}


// ###### Get number of layers ##############################################
cardinal TDTFReader::getLayers(const double frameRate) const
{
   const TraceHeader* trace = getTraceHeader(frameRate);
   if(trace != NULL) {
      return(trace->Layers);
   }
   return(0);
}


// ###### Get frame size for position from trace ############################
cardinal TDTFReader::getTraceFrameSize(const card64   rtpPosition,
                                       const cardinal layer,
                                       const double   frameRate) const
{
   const cardinal position = positionToFramePosition(frameRate,rtpPosition);
   if(position >= getFrames(frameRate)) {
      cerr << "ERROR: TDTFReader::getTraceFrameSize() - Invalid position!"
           << endl;
      exit(1);
   }

   const TraceHeader* trace = getTraceHeader(frameRate);
   if(trace != NULL) {
      const cardinal frames = translate32(trace->Frames);
      if(position < frames) {
         return(translate32(trace->Frame[layer * frames + position].Size));
      }
   }
   return(0);
}


// ###### Get frame ID for position from trace ############################
cardinal TDTFReader::getTraceFrameID(const card64   rtpPosition,
                                     const cardinal layer,
                                     const double   frameRate) const
{
   const cardinal position = positionToFramePosition(frameRate,rtpPosition);
   if(position >= getFrames(frameRate)) {
      cerr << "ERROR: TDTFReader::getTraceFrameID() - Invalid position!"
           << endl;
      exit(1);
   }

   const TraceHeader* trace = getTraceHeader(frameRate);
   if(trace != NULL) {
      const cardinal frames = translate32(trace->Frames);
      if(position < frames) {
         return(translate32(trace->Frame[layer * frames + position].ID));
      }
   }
   return(0);
}


// ###### Get maximum buffer delay ##########################################
cardinal TDTFReader::getMaxBufferDelay(const card64 rtpPosition,
                                       const double frameRate) const
{
   const cardinal position = positionToFramePosition(frameRate,rtpPosition);
   if(position >= getFrames(frameRate)) {
      return(1);
   }

   const cardinal layers = getLayers(frameRate);
   cardinal minDelay = (cardinal)-1;
   for(cardinal j = 0;j < layers;j++) {
      const EmpiricalEnvelope* ee = getEmpiricalEnvelope(position,j,frameRate);
      if(ee == NULL) {
         cerr << "ERROR: TDTFReader::getMaxBufferDelay() - Empirical envelope header not found!" << endl;
         exit(1);
      }
      const cardinal pairs = translate16(ee->Pairs);
      if(translate16(ee->Pair[pairs - 1].Interval) < minDelay) {
         minDelay = translate16(ee->Pair[pairs - 1].Interval);
      }
   }
   return(minDelay);
}


// ###### Get utilization constants for frame size ##########################
const ResourceUtilizationHeader* TDTFReader::getResourceUtilizationHeader(
                                                const cardinal position,
                                                const double   frameRate) const
{
   // ====== Get RU list index header =======================================
   if(RULIndex == NULL) {
      return(NULL);
   }
   cardinal indexList = (cardinal)-1;
   cardinal length    = 0;
   cardinal start     = 0;
   cardinal end       = RULIndexEntries - 1;


   // ====== Check limits ===================================================
   if(position <= translate32(RULIndex->Entry[0].Position)) {
      indexList = translate32(RULIndex->Entry[0].List);
      length    = translate32(RULIndex->Entry[0].Length);
   }
   else if(position >= translate32(RULIndex->Entry[end].Position)) {
      indexList = translate32(RULIndex->Entry[end].List);
      length    = translate32(RULIndex->Entry[end].Length);
   }


   // ====== Do binary search ===============================================
   else {
      cardinal m = 0;
      do {
         m = (start + end) / 2;
         const cardinal p0 = translate32(RULIndex->Entry[m].Position);
         const cardinal p1 = p0 + translate32(RULIndex->Entry[m].Length) - 1;
         if(position < p0) {
            end = m;
         }
         else if(position > p1) {
            start = m;
         }
         else {
            start = m;
            end   = m;
         }
      } while(end - start > 1);
      indexList = translate32(RULIndex->Entry[start].List);
      length    = translate32(RULIndex->Entry[start].Length);
   }


   // ====== Get resource/utilization header ================================
   const ResourceUtilizationHeader* uh =
      (const ResourceUtilizationHeader*)&InputMemory[indexList];
   if(checkAccess((long)uh - (long)InputMemory,sizeof(ResourceUtilizationHeader)) == false) {
      return(NULL);
   }
   if(checkAccess((long)uh - (long)InputMemory,
                  sizeof(ResourceUtilizationHeader) +
                     (translate16(uh->Entries) * sizeof(ResourceUtilizationEntry))) == false) {
      return(NULL);
   }
   if((position < translate32(uh->Position)) ||
      (position > translate32(uh->Position) + length - 1)) {
      cerr << "ERROR: TDTFReader::getResourceUtilizationHeader() - Position check failed!" << endl;
      printf("%d  p=%d l=%d\n",position,translate32(uh->Position),length);
//      exit(1);
      return(NULL);
   }
   return(uh);
}


// ###### Get utilization constants for frame size ##########################
const UtilizationHeader* TDTFReader::getFrameSizeUtilizationHeader(
                                        const cardinal position,
                                        const cardinal layer,
                                        const double   frameRate) const
{
   // ====== Get layer header ===============================================
   const LayerHeader* lh = getLayerHeader(position,layer,frameRate);
   if(lh == NULL) {
      return(NULL);
   }

   // ====== Get utilization header =========================================
   const cardinal uhOffset     = (cardinal)translate32(lh->FrameSizeUtilizationOffset);
   const UtilizationHeader* uh = (const UtilizationHeader*)((long)lh + (long)uhOffset);
   if(checkAccess((long)uh - (long)InputMemory,sizeof(UtilizationHeader)) == false) {
      return(NULL);
   }
   if(checkAccess((long)uh - (long)InputMemory,
                  sizeof(UtilizationHeader) +
                     ((cardinal)uh->Constants * sizeof(card64))) == false) {
      return(NULL);
   }
   return(uh);
}


// ###### Get utilization constants for frame rate ##########################
const UtilizationHeader* TDTFReader::getFrameRateUtilizationHeader(
                                        const cardinal position,
                                        const double   frameRate) const
{
   // ====== Get interval header ============================================
   const IntervalHeader* ih = getIntervalHeader(position,frameRate);
   if(ih == NULL) {
      return(NULL);
   }

   // ====== Get utilization header =========================================
   const cardinal uhOffset     = (cardinal)translate32(ih->Offset[IntervalHeader::offsetUH]);
   const UtilizationHeader* uh = (const UtilizationHeader*)((long)ih + (long)uhOffset);
   if(checkAccess((long)uh - (long)InputMemory,sizeof(UtilizationHeader)) == false) {
      return(NULL);
   }
   if(checkAccess((long)uh - (long)InputMemory,
                  sizeof(UtilizationHeader) +
                     ((cardinal)uh->Constants * sizeof(card64))) == false) {
      return(NULL);
   }
   return(uh);
}


// ###### Print empirical envelope data for given frame rate ################
void TDTFReader::printEmpiricalEnvelope(ostream&       os,
                                        const cardinal position,
                                        const double   frameRate,
                                        const bool     frameCount,
                                        const bool     compact) const
{
   const cardinal           layers = getLayers(frameRate);
   const EmpiricalEnvelope* ee[layers];
   char                     str[256];


   // ====== Get empirical envelope for each layer ==========================
   cardinal maxPairs       = 0;
   cardinal maxBufferDelay = 0;
   for(cardinal j = 0;j < layers;j++) {
      ee[j] = getEmpiricalEnvelope(position,j,frameRate,frameCount);
      if(ee[j] == NULL) {
         cerr << "ERROR: TDTFReader::printEmpiricalEnvelope() - Empirical envelope header not found!" << endl;
         exit(1);
      }
      const cardinal pairs = translate16(ee[j]->Pairs);
      if(pairs > maxPairs) {
         maxPairs = pairs;
      }
      if(maxBufferDelay < translate16(ee[j]->Pair[pairs - 1].Interval)) {
         maxBufferDelay = translate16(ee[j]->Pair[pairs - 1].Interval);
      }
   }


   // ====== Print empirical envelope for each layer ========================
   os << " ";
   for(cardinal j = 0;j < layers;j++) {
      snprintf((char*)&str,sizeof(str),"Layer #%d       ",j);
      os << str;
   }
   os << endl;



   // ====== Do verification of empirical envelope ==========================
   for(cardinal j = 1;j <= maxBufferDelay;j++) {
      for(cardinal k = 0;k < layers;k++) {
         cardinal sum;
         if(frameCount == false) {
            sum = getMaxByteCountForDelay(framePositionToPosition(frameRate,position),k,frameRate,j);
         }
         else {
            sum = getMaxFrameCountForDelay(framePositionToPosition(frameRate,position),k,frameRate,j);
         }

         for(cardinal i = 0;i < translate16(ee[k]->Pairs);i++) {
            if(j == translate16(ee[k]->Pair[i].Interval)) {
               if(sum != translate32(ee[k]->Pair[i].Sum)) {
                  cerr << "INTERNAL ERROR: TDTFReader::printEmpiricalEnvelope() - Verification assertion for empirical envelope!"
                       << endl;
                  exit(1);
               }
               break;
            }
         }
      }
   }


   // ====== Verify traffic constraint ======================================
   const IntervalHeader* ih = getIntervalHeader(position,frameRate);
   if(ih == NULL) {
      cerr << "INTERNAL ERROR: TDTFReader::printEmpiricalEnvelope() - Verification assertion for interval header!"
           << endl;
      exit(1);
   }
   const cardinal start  = translate32(ih->Position);
   const cardinal end = start + translate32(ih->Length) - 1;
   for(cardinal l = 0;l < layers;l++) {
      for(cardinal d = 0;d < translate16(ee[l]->Pairs);d++) {
         const cardinal delay = translate16(ee[l]->Pair[d].Interval);
         const cardinal limit = translate32(ee[l]->Pair[d].Sum);
         const cardinal traceStart = 0;
         const cardinal traceEnd   = getFrames(frameRate) - 1;
         cardinal istart;
         cardinal iend;

         if(start - traceStart >= delay - 1) {
            istart = traceStart + start - (delay - 1);
         }
         else {
            istart = traceStart;
         }

         if(end + delay - 1 < traceEnd) {
             iend = end + delay - 1;
         }
         else {
             iend = traceEnd;
         }

         cardinal max = 0;
         for(cardinal i = istart;i <= iend;i++) {
            cardinal sum = 0;
            for(cardinal j = i;j <= min(i + (delay - 1),iend);j++) {
               const cardinal size = getTraceFrameSize(
                  framePositionToPosition(frameRate,j),l,frameRate);
               if(!frameCount) {
                  sum += size;
               }
               else {
                  if(size > 0) {
                     sum++;
                  }
               }
            }
            if(sum > max) {
               max = sum;
            }
         }

         if(max != limit) {
            cerr << "INTERNAL ERROR: TDTFReader::printEmpiricalEnvelope() - Verification assertion for traffic constraint!"
                 << endl;
            exit(1);
         }
      }
   }


   // ====== Full view, with EE approximations ==============================
   if(!compact) {
      for(cardinal j = 1;j <= maxBufferDelay;j++) {
         os << " ";
         for(cardinal k = 0;k < layers;k++) {
            // ====== Get value =============================================
            cardinal sum;
            if(frameCount == false) {
               sum = getMaxByteCountForDelay(framePositionToPosition(frameRate,position),k,frameRate,j);
            }
            else {
               sum = getMaxFrameCountForDelay(framePositionToPosition(frameRate,position),k,frameRate,j);
            }

            // ====== Print value ===========================================
            bool found = false;
            for(cardinal i = 0;i < translate16(ee[k]->Pairs);i++) {
               if(j == translate16(ee[k]->Pair[i].Interval)) {
                  found = true;
                  break;
               }
            }
            if(found == false) {
               snprintf((char*)&str,sizeof(str),"%2d -> %6d   ",j,sum);
            }
            else {
               snprintf((char*)&str,sizeof(str),"%2d +> %6d   ",j,sum);
            }
            os << str;
         }
         os << endl;
      }
   }

   // ====== Compact view, without EE approximations ===========================
   else {
      for(cardinal j = 0;j < maxPairs;j++) {
         os << " ";
         for(cardinal k = 0;k < layers;k++) {
            // ====== Check, if there are pairs to print =======================
            if(j < translate16(ee[k]->Pairs)) {
               // ====== Print entry ===========================================
               const cardinal interval = translate16(ee[k]->Pair[j].Interval);
               const cardinal sum      = translate32(ee[k]->Pair[j].Sum);
               snprintf((char*)&str,sizeof(str),"%2d +> %6d   ",interval,sum);
               os << str;
            }

            // ====== No more pairs to print in this layer =====================
            else {
               os << "               ";
            }
         }
         os << endl;
      }
   }
}


// ###### Print TDTF trace for given frame rate #############################
void TDTFReader::print(ostream&     os,
                       const double printFrameRate,
                       const bool   printEE,
                       const bool   compactEE,
                       const bool   printRUL,
                       const bool   printUC) const
{
   // ====== Initialize =====================================================
   if(MainIndex == NULL) {
      os << "(no file opened)" << endl;
      return;
   }
   const IntervalHeader* ih;
   const double   frameRate = getNearestValidFrameRate(printFrameRate);
   const cardinal frames    = getFrames(frameRate);
   const cardinal layers    = getLayers(frameRate);
   cardinal       intLength = 0;
   cardinal       intCount  = 0;
   char           str[256];


   // ====== Get interval count =============================================
   const PositionLengthIntervalIndexHeader* ipli = getIPLIHeader(frameRate);
   if(ipli == NULL) {
      cerr << "ERROR: TDTFReader::print() - PositionLengthInterval header not found!" << endl;
      exit(1);
   }
   const cardinal intMaxCount = translate32(ipli->Entries);
   const TraceHeader* trace = getTraceHeader(frameRate);
   if(trace == NULL) {
      cerr << "ERROR: TDTFReader::print() - Trace header not found!" << endl;
      exit(1);
   }


   // ====== Print all frames ===============================================
   os << endl;
   os << "####################################" << endl;
   snprintf((char*)&str,sizeof(str),"#### Trace for Frame Rate %1.2f",frameRate);
   os << str;
   for(cardinal i = strlen(str);i <32;i++) {
      os << " ";
   }
   os << "####" << endl;
   os << "####################################" << endl << endl;
   for(cardinal i = 0;i < frames;i++) {
      // ====== Update interval description =================================
      if(intLength <= 0) {
         // ====== Get interval header ======================================
         ih = getIntervalHeader(i,frameRate);
         if(ih == NULL) {
            cerr << "ERROR: TDTFReader::print() - Interval header not found!" << endl;
            exit(1);
         }
         intLength = translate32(ih->Length);
         intCount++;

         snprintf((char*)&str,sizeof(str),
                  "==== Interval %05d of %05d, length %04d =============",
                  intCount,intMaxCount,intLength);
         os << endl << str;
         for(cardinal j = 3;j < layers;j++) {
            os << "==============";
         }
         os << endl;


         // ====== Print scalability ========================================
         for(cardinal j = 0;j < ih->Layers;j++) {
            const LayerHeader* lh = getLayerHeader(i,j,frameRate);
            if(lh == NULL) {
               cerr << "ERROR: TDTFReader::print() - Layer header not found!" << endl;
               exit(1);
            }
            const double scalability = (double)translate32(lh->Scalability) / pow(2.0,30.0);
            snprintf((char*)&str, sizeof(str), "Layer #%d Flags = $%02x  Scalability = %1.3f",
                                               j, lh->Flags,
                                               scalability);
            os << str << endl;

            // ====== Do verification =======================================
            const double check =
               getLayerScalability(framePositionToPosition(frameRate,i),j,frameRate);
            if(check != scalability) {
               cerr << "INTERNAL ERROR: TDTFReader::print() - Verification assertion for scalability!"
                    << endl;
               exit(1);
            }
         }

         // ====== Print frame rate utilization =============================
         if(printUC == true) {
            os << "Utilization Constants:" << endl;
            const UtilizationHeader* uh = getFrameRateUtilizationHeader(i,frameRate);
            if(uh == NULL) {
               cerr << "ERROR: TDTFReader::print() - Missing utilization header for frame rate!"
                    << endl;
               exit(1);
            }
            char str[64];
            snprintf((char*)&str,sizeof(str),"$%04x",translate16(uh->Type));
            os << " Frame Rate: Type " << str
               << ", Weight " << translateToDouble(uh->Weight) << " - {";
            for(cardinal j = 0;j < uh->Constants;j++) {
               os << translateToDouble(uh->Constant[j]);
               if((integer)j < (integer)uh->MaxConstants - 1) {
                  os << ", ";
               }
            }
            for(integer j = uh->Constants;j < uh->MaxConstants;j++) {
               os << "(empty)";
               if((integer)j < (integer)uh->MaxConstants - 1) {
                  os << ", ";
               }
            }
            os << "}" << endl;


            // ====== Print frame size utilization =============================
            for(cardinal j = 0;j < ih->Layers;j++) {
               const UtilizationHeader* uh = getFrameSizeUtilizationHeader(i,j,frameRate);
               snprintf((char*)&str,sizeof(str),"$%04x",translate16(uh->Type));
               os << " Layer #" << j << " Frame Size: Type " << str
                  << ", Weight " << translateToDouble(uh->Weight) << " - {";
               if(uh == NULL) {
                  cerr << "ERROR: TDTFReader::print() - Missing utilization header for frame rate!"
                       << endl;
                  exit(1);
               }
               for(cardinal j = 0;j < uh->Constants;j++) {
                  os << translateToDouble(uh->Constant[j]);
                  if((integer)j < (integer)uh->MaxConstants - 1) {
                     os << ", ";
                  }
               }
               for(integer j = uh->Constants;j < uh->MaxConstants;j++) {
                  os << "(empty)";
                  if((integer)j < (integer)uh->MaxConstants - 1) {
                     os << ", ";
                  }
               }
               os << "}" << endl;
            }
         }

         // ====== Print resource/utilization list ==========================
         if(printRUL == true) {
            os << "Resource/Utilization List:" << endl;
            const ResourceUtilizationHeader* ruh = getResourceUtilizationHeader(i,frameRate);
            if(ruh == NULL) {
               cerr << "ERROR: TDTFReader::print() - Resource/Utilization header not found!" << endl;
               exit(1);
            }
            const cardinal entries = translate16(ruh->Entries);
            if(entries > 0) {
               os << " Point  Util.    Total  ";
               for(cardinal k = 0;k < ih->Layers;k++) {
                  os << "Layer #" << k << " ";
               }
               os << " Rate" << endl;
               const ResourceUtilizationEntry* entry = &ruh->Entry[0];
               for(cardinal k = 0;k < entries;k++) {
                  snprintf((char*)&str,sizeof(str),
                           " %3d    %1.3f   ",k,
                           (double)translate32(entry->Utilization) / (double)pow(2.0,30.0));
                  os << str;
                  for(cardinal j = 0;j < (cardinal)(ih->Layers + 1);j++) {
                     snprintf((char*)&str,sizeof(str),
                              "%7d  ",translate32(entry->Bandwidth[j]));
                     os << str;
                  }
                  os << translateToDouble(entry->FrameRate) << endl;

                  entry = (ResourceUtilizationEntry*)
                     ((long)entry +
                      (long)(sizeof(ResourceUtilizationEntry) +
                            ((ih->Layers + 1) * sizeof(card32))));
               }
            }
            else {
               os << " (empty list)" << endl;
            }
         }


         // ===== Print empirical envelopes =================================
         if(printEE == true) {
            os << "Byterate Empirical Envelope:" << endl;
            printEmpiricalEnvelope(os,i,frameRate,false,compactEE);
            os << "Frame Count Empirical Envelope:" << endl;
            printEmpiricalEnvelope(os,i,frameRate,true,compactEE);
         }
         if((printEE == true) || (printRUL == true)) {
            cout << "Trace:" << endl;
         }
      }
      intLength--;


      // ====== Print trace =================================================
      snprintf((char*)&str,sizeof(str),"#%06d: ",i);
      os << str;
      cardinal size[layers];
      for(cardinal j = 0;j < layers;j++) {
         // ====== Print frame size =========================================
         size[j] = getTraceFrameSize(
                      framePositionToPosition(frameRate,i),j,frameRate);
         snprintf((char*)&str,sizeof(str),"%6d ",size[j]);
         os << str;

         // ====== Do verification ==========================================
         if(size[j] != translate32(trace->Frame[j * frames + i].Size)) {
            cerr << "INTERNAL ERROR: TDTFReader::print() - Verification assertion for trace frame size!" << endl;
            exit(1);
         }
         const EmpiricalEnvelope* ee =
            getEmpiricalEnvelope(i,j,frameRate,false);
         if((ee == NULL) ||
            (translate16(ee->Pair[0].Interval) != 1) ||
            (translate32(ee->Pair[0].Sum) < size[j])) {
            cerr << "INTERNAL ERROR: TDTFReader::print() - Verification assertion for traffic constraint!" << endl;
            exit(1);
         }
      }
      os << "   (";
      for(cardinal j = 0;j < layers;j++) {
         // ====== Print frame ID ===========================================
         cardinal id = (cardinal)-1;
         if(size[j] != 0) {
            id = getTraceFrameID(framePositionToPosition(frameRate,i),j,frameRate);
            snprintf((char*)&str,sizeof(str),"#%05d",id);
         }
         else {
            strcpy((char*)&str,"#-----");
         }

         // ====== Do verification ==========================================
         if(id != translate32(trace->Frame[j * frames + i].ID)) {
            cerr << "INTERNAL ERROR: TDTFReader::print() - Verification assertion for trace frame ID!" << endl;
            exit(1);
         }

         // ====== Print space ==============================================
         os << str;
         if(j < layers - 1) {
            os << " ";
         }
      }
      os << ")" << endl;
   }
}


}
