// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### TDTF Writer Implementation                                       ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "tdtfwriter.h"
#include "tools.h"


namespace Coral {


// Verbose mode: Print some information
// #define VERBOSE


// ###### Constructor #######################################################
TDTFWriter::TDTFWriter()
{
}


// ###### Destructor ########################################################
TDTFWriter::~TDTFWriter()
{
}


// ###### Write trace #######################################################
cardinal TDTFWriter::writeTrace()
{
   cardinal filePosition = ftell(OutputFile);

   // ====== Write trace header =============================================
#ifdef VERBOSE
   cout << "   Writing trace..." << endl;
#endif
   TraceHeader th;
   th.FrameRate = translateToBinary(TArray->FrameRate);
   th.Frames    = translate32(TArray->Frames);
   th.Layers    = (card8)TArray->Layers;
   th.pad01     = 0;
   th.pad02     = 0;

   if(fwrite((char*)&th,sizeof(th),1,OutputFile) != 1) {
      cerr << "ERROR: TDTFWriter::writeTrace() - Write error!" << endl;
      exit(1);
   }


   // ====== Write trace ====================================================
   for(cardinal j = 0;j < TArray->Layers;j++) {
      FrameDescription frames[TArray->Frames];
      for(cardinal i = 0;i < TArray->Frames;i++) {
         frames[i].ID   = translate32(TArray->LayerTrace[j]->Frame[i].ID);
         frames[i].Size = translate32(TArray->LayerTrace[j]->Frame[i].Size);
      }
      if(fwrite((char*)&frames,sizeof(frames),1,OutputFile) != 1) {
         cerr << "ERROR: Write error!" << endl;
         exit(1);
      }
   }

   return(filePosition);
}


// ###### Write intervals ###################################################
cardinal TDTFWriter::writeIntervals()
{
   // ====== Compute intervals ==============================================
#ifdef VERBOSE
   cout << "   Computing intervals..." << endl;
#endif
   const cardinal minLength =
      max(1,(cardinal)ceil(((double)Config.MinIntervalLength * (double)Config.FrameRate)
                             / 1000000.0));
   const cardinal maxLength =
      max(minLength + 1,(cardinal)ceil(((double)Config.MaxIntervalLength * (double)Config.FrameRate)
                                          / 1000000.0));

   double   costArray[TArray->Frames];
   cardinal lengthArray[TArray->Frames];
   TArray->calculateOptimalIntervals((double*)&costArray,
                                     (cardinal*)&lengthArray,
                                     minLength,
                                     maxLength);


   // ====== Count intervals ================================================
#ifdef VERBOSE
   cout << "   Writing intervals..." << endl;
#endif
   cardinal pos       = 0;
   cardinal intervals = 0;
   while(pos < TArray->Frames) {
      pos += lengthArray[pos];
      intervals++;
   }


   // ====== Generate traffic description ===================================
   PositionLengthIntervalIndexEntry index[TArray->Frames];
   cardinal indexPosition = 0;
   pos = 0;
   while(pos < TArray->Frames) {
      // ====== Add entry to index ==========================================
      // printf("POS=%d  LEN=%d  LAYERS=%d\n",pos,lengthArray[pos],TArray->Layers);
      index[indexPosition].Position = translate32(pos);
      index[indexPosition].Length   = translate32(lengthArray[pos]);
      index[indexPosition].Interval = translate32(ftell(OutputFile));
      indexPosition++;

/*
      for(cardinal ii = 0;ii < indexPosition;ii++) {
         printf("P=%d L=%d\n",translate32(index[ii].Position),translate32(index[ii].Length));
      }
*/


      // ====== Write interval header =======================================
      char ihBuffer[IntervalHeader::getIntervalHeaderSize(TArray->Layers)];
      IntervalHeader* ih = (IntervalHeader*)&ihBuffer;
      ih->Position = translate32(pos);
      ih->Length   = translate32(lengthArray[pos]);
      ih->Flags    = 0x00;
      ih->Layers   = (card8)TArray->Layers;
      ih->pad      = 0;
      const cardinal indexIntervalHeader = ftell(OutputFile);
      /* Important note:
         This header will be overwritten with updated IntervalHeader
         including offsets for Layer and EmpiricalEnvelope headers!
      */
      ih->IntervalDescriptionSize = sizeof(ihBuffer);
      if(fwrite(ihBuffer,sizeof(ihBuffer),1,OutputFile) != 1) {
         cerr << "ERROR: TDTFWriter::writeIntervals() - Write error!" << endl;
         exit(1);
      }


      // ====== Write frame rate utilization ================================
      char frUhBuffer[UtilizationHeader::getUtilizationSize(Config.FrameRateUtilizationMaxConstants)];
      UtilizationHeader* uh = (UtilizationHeader*)&frUhBuffer;
      uh->MaxConstants = Config.FrameRateUtilizationMaxConstants;
      uh->Constants    = Config.FrameRateUtilizationConstants;
      uh->Type         = translate16(Config.FrameRateUtilizationType);
      uh->Weight       = translateToBinary(Config.FrameRateUtilizationWeight);
      for(cardinal i = Config.FrameRateUtilizationConstants;i < Config.FrameRateUtilizationMaxConstants;i++) {
         uh->Constant[i] = translateToBinary(0.0);
      }
      for(cardinal i = 0;i < Config.FrameRateUtilizationConstants;i++) {
         uh->Constant[i] = translateToBinary(Config.FrameRateUtilizationConstant[i]);
      }
      if(fwrite((char*)uh,sizeof(frUhBuffer),1,OutputFile) != 1) {
         cerr << "ERROR: TDTFWriter::writeIntervals() - Write error!" << endl;
         exit(1);
      }
      ih->Offset[IntervalHeader::offsetUH] = translate32(ih->IntervalDescriptionSize);
      ih->IntervalDescriptionSize += sizeof(frUhBuffer);


      // ====== Write layer-specific information ============================
      for(cardinal i = 0;i < TArray->Layers;i++) {
         // ====== Write layer header =======================================
         LayerHeader lh;
         lh.Scalability = translate32((cardinal)rint(Config.Layer[i].Scalability * pow(2.0,30.0)));
         lh.FrameSizeUtilizationOffset = translate32(sizeof(lh));
         lh.Flags = (card8)Config.Layer[i].LayerFlags;
         lh.pad   = 0x00;
         if(fwrite((char*)&lh,sizeof(lh),1,OutputFile) != 1) {
            cerr << "ERROR: TDTFWriter::writeIntervals() - Write error!" << endl;
            exit(1);
         }
         ih->Offset[ih->getOffsetLH(i)] = translate32(ih->IntervalDescriptionSize);
         ih->IntervalDescriptionSize += sizeof(lh);


         // ====== Write frame size utilization =============================
         char fsUhBuffer[UtilizationHeader::getUtilizationSize(Config.Layer[i].FrameSizeUtilizationMaxConstants)];
         UtilizationHeader* uh = (UtilizationHeader*)&fsUhBuffer;
         uh->MaxConstants = Config.Layer[i].FrameSizeUtilizationMaxConstants;
         uh->Constants    = Config.Layer[i].FrameSizeUtilizationConstants;
         uh->Type         = translate16(Config.Layer[i].FrameSizeUtilizationType);
         uh->Weight       = translateToBinary(Config.Layer[i].FrameSizeUtilizationWeight);
         for(cardinal j = Config.Layer[i].FrameSizeUtilizationConstants;j < Config.Layer[i].FrameSizeUtilizationMaxConstants;j++) {
            uh->Constant[j] = translateToBinary(0.0);
         }
         for(cardinal j = 0;j < Config.Layer[i].FrameSizeUtilizationConstants;j++) {
            uh->Constant[j] = translateToBinary(Config.Layer[i].FrameSizeUtilizationConstant[j]);
         }
         if(fwrite((char*)uh,sizeof(fsUhBuffer),1,OutputFile) != 1) {
            cerr << "ERROR: TDTFWriter::writeIntervals() - Write error!" << endl;
            exit(1);
         }
         ih->IntervalDescriptionSize += sizeof(fsUhBuffer);


         // ====== Calculate empirical envelope =============================
         char eeBuffer[EmpiricalEnvelope::getSize(Config.Layer[i].ByterateEmpiricalEnvelopePairs)];
         EmpiricalEnvelope* ee =
            TArray->initEmpiricalEnvelope(eeBuffer,Config.Layer[i].ByterateEmpiricalEnvelopePairs,false);
         TArray->calculateEmpiricalEnvelope(
                        i, pos, pos + lengthArray[pos] - 1,
                        0, TArray->Frames - 1, ee,
                        false);

         // ====== Write empirical envelope =================================
         for(cardinal j = 0;j < ee->Pairs;j++) {
            ee->Pair[j].Sum      = translate32(ee->Pair[j].Sum);
            ee->Pair[j].Interval = translate16(ee->Pair[j].Interval);
         }
         ee->Pairs = translate16(ee->Pairs);
         if(fwrite((char*)&eeBuffer,sizeof(eeBuffer),1,OutputFile) != 1) {
            cerr << "ERROR: TDTFWriter::writeIntervals() - Write error!" << endl;
            exit(1);
         }
         ih->Offset[ih->getOffsetEEBR(i)] = translate32(ih->IntervalDescriptionSize);
         ih->IntervalDescriptionSize += sizeof(eeBuffer);


         // ====== Calculate frame counts ===================================
         char fcBuffer[EmpiricalEnvelope::getSize(Config.Layer[i].FrameCountEmpiricalEnvelopePairs)];
         EmpiricalEnvelope* fc =
            TArray->initEmpiricalEnvelope(fcBuffer,Config.Layer[i].FrameCountEmpiricalEnvelopePairs,true);
         TArray->calculateEmpiricalEnvelope(
                        i, pos, pos + lengthArray[pos] - 1,
                        0, TArray->Frames - 1, fc,
                        true);


         // ====== Write frame counts =======================================
         for(cardinal j = 0;j < fc->Pairs;j++) {
            fc->Pair[j].Sum      = translate32(fc->Pair[j].Sum);
            fc->Pair[j].Interval = translate16(fc->Pair[j].Interval);
         }
         fc->Pairs = translate16(fc->Pairs);
         if(fwrite((char*)&fcBuffer,sizeof(fcBuffer),1,OutputFile) != 1) {
            cerr << "ERROR: TDTFWriter::writeIntervals() - Write error!" << endl;
            exit(1);
         }
         ih->Offset[ih->getOffsetEEFC(i)] = translate32(ih->IntervalDescriptionSize);
         ih->IntervalDescriptionSize += sizeof(fcBuffer);
      }


      // ====== Update header ===============================================
      const cardinal oldPosition = ftell(OutputFile);
      fseek(OutputFile,indexIntervalHeader,SEEK_SET);
      ih->IntervalDescriptionSize = translate32(ih->IntervalDescriptionSize);
      if(fwrite(ihBuffer,sizeof(ihBuffer),1,OutputFile) != 1) {
         cerr << "ERROR: Write error!" << endl;
         exit(1);
      }
      fseek(OutputFile,oldPosition,SEEK_SET);


      // ====== Get next position ===========================================
      pos += lengthArray[pos];
   }


   // ====== Write index ====================================================
   cardinal filePosition = ftell(OutputFile);
   PositionLengthIntervalIndexHeader iplih;
   iplih.Entries = translate32(indexPosition);
   if(fwrite((char*)&iplih,sizeof(iplih),1,OutputFile) != 1) {
      cerr << "ERROR: TDTFWriter::writeIntervals() - Write error!" << endl;
      exit(1);
   }
   if(fwrite((char*)&index,sizeof(PositionLengthIntervalIndexEntry),indexPosition,OutputFile) != indexPosition) {
      cerr << "ERROR: TDTFWriter::writeIntervals() - Write error!" << endl;
      exit(1);
   }
   if(translate32(index[0].Position) != 0) {
      cerr << "INTERNAL ERROR: TDTFWriter::writeIntervals() - Buffer overflow?!"
           << endl;
   }

/*
   puts("Index:");
   for(cardinal i = 0;i < translate32(iplih.Entries);i++) {
      printf("P=%d L=%d\n",translate32(index[i].Position),translate32(index[i].Length));
   }
   exit(1);
*/
   return(filePosition);
}


// ###### Generate TDTF from trace ##########################################
void TDTFWriter::generate(const char*               name,
                          TraceArray*               originalTraceArray,
                          const TraceConfiguration& config)
{
   // ====== Create file, initialize ========================================
   OutputFile=fopen(name,"w+");
   if(OutputFile == NULL) {
      cerr << "ERROR: TDTFWriter::generator() - Unable to create file <"
           << name << ">!" << endl;
      exit(1);
   }
   TArray = originalTraceArray;
   Config = config;


   // ====== Writing prefix =================================================
   TDTFPrefix tp;
   tp.TraceTimeStamp       = translate64(getMicroTime());
   tp.UtilizationTimeStamp = 0;
   tp.TDTF_ID[0] = 'T';
   tp.TDTF_ID[1] = 'D';
   tp.TDTF_ID[2] = 'T';
   tp.TDTF_ID[3] = 'F';
   tp.Version           = translate32(TDTFPrefix::CurrentVersion);
   tp.MinIntervalLength = translate64(config.MinIntervalLength);
   tp.MaxIntervalLength = translate64(config.MaxIntervalLength);
   strncpy((char*)&tp.Title,(char*)&config.Title,TDTFPrefix::MaxTitleLength);
   strncpy((char*)&tp.Copyright,(char*)&config.Copyright,TDTFPrefix::MaxCopyrightLength);
   strncpy((char*)&tp.Comment,(char*)&config.Comment,TDTFPrefix::MaxCommentLength);
   strncpy((char*)&tp.URL,(char*)&config.URL,TDTFPrefix::MaxURLLength);
   for(cardinal i = strlen(tp.Title);i < TDTFPrefix::MaxTitleLength;i++) {
      tp.Title[i] = 0x00;
   }
   for(cardinal i = strlen(tp.Copyright);i < TDTFPrefix::MaxCopyrightLength;i++) {
      tp.Copyright[i] = 0x00;
   }
   for(cardinal i = strlen(tp.Comment);i < TDTFPrefix::MaxCommentLength;i++) {
      tp.Comment[i] = 0x00;
   }
   for(cardinal i = strlen(tp.URL);i < TDTFPrefix::MaxURLLength;i++) {
      tp.URL[i] = 0x00;
   }
   if(fwrite((char*)&tp,sizeof(tp),1,OutputFile) != 1) {
      cerr << "ERROR: TDTFWriter::generator() - Write error!" << endl;
      exit(1);
   }


   // ====== Write last extension header ====================================
   TDTFPrefixExtensionHeader eh;
   eh.ExtID[0]  = 0x00;
   eh.ExtID[1]  = 0x00;
   eh.ExtID[2]  = 0x00;
   eh.ExtID[3]  = 0x00;
   eh.ExtLength = 0;
   if(fwrite((char*)&eh,sizeof(eh),1,OutputFile) != 1) {
      cerr << "ERROR: TDTFWriter::generator() - Write error!" << endl;
      exit(1);
   }


   // ====== Write trace and intervals for each frame rate ==================
   const cardinal traceStart = ftell(OutputFile);

   // Limit of possible frame rates: 256
   const cardinal maxFrameRates = 256;
   double   indexFrameRate[maxFrameRates];
   cardinal indexTrace[maxFrameRates];
   cardinal indexIntervals[maxFrameRates];
   cardinal frameRates = 0;

   do {
      cout << "Computing intervals for frame rate " << TArray->FrameRate
           << "." << endl;

      indexFrameRate[frameRates] = TArray->FrameRate;
      indexTrace[frameRates]     = writeTrace();
      indexIntervals[frameRates] = writeIntervals();
      frameRates++;

      TArray = TArray->decreaseFrameRate();
   } while((TArray != NULL) && (frameRates < maxFrameRates));


   // ====== Write main index ===============================================
#ifdef VERBOSE
   cout << "Writing index..." << endl;
#endif
   const cardinal indexMainIndex = ftell(OutputFile);
   MainIndexHeader ih;
   ih.Entries = translate32(frameRates);
   if(fwrite((char*)&ih,sizeof(ih),1,OutputFile) != 1) {
      cerr << "ERROR: TDTFWriter::generator() - Write error!" << endl;
      exit(1);
   }
   for(cardinal i = 0;i < frameRates;i++) {
      MainIndexEntry ie;
      ie.FrameRate = translateToBinary(indexFrameRate[i]);
      ie.Trace     = translate32(indexTrace[i]);
      ie.Intervals = translate32(indexIntervals[i]);
      if(fwrite((char*)&ie,sizeof(ie),1,OutputFile) != 1) {
         cerr << "ERROR: TDTFWriter::generator() - Write error!" << endl;
         exit(1);
      }
   }


   // ====== Write suffix ===================================================
   TDTFSuffix ts;
   ts.TDTF_ID[0] = 'T';
   ts.TDTF_ID[1] = 'D';
   ts.TDTF_ID[2] = 'T';
   ts.TDTF_ID[3] = 'F';
   ts.TraceStart               = translate32(traceStart);
   ts.ResourceUtilizationStart = translate32(ftell(OutputFile));
   ts.MainIndex                = translate32(indexMainIndex);
   ts.ResourceUtilizationIndex = 0;
   ts.UtilizationTimeStamp     = 0;
   ts.TraceTimeStamp           = translate64(getMicroTime());
   if(fwrite((char*)&ts,sizeof(ts),1,OutputFile) != 1) {
      cerr << "ERROR: TDTFWriter::generator() - Write error!" << endl;
      exit(1);
   }

   fclose(OutputFile);
   OutputFile = NULL;
   cout << endl << "Congratulations! Your trace file is complete!" << endl;
}


}
