// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Client Implementation                                      ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "socket.h"
#include "tracedecoder.h"
#include "tracedecoderrepository.h"
#include "rtpreceiver.h"
#include "rtcppacket.h"
#include "rtcpsender.h"
#include "set.h"
#include "tools.h"
#include "strings.h"
#include "randomizer.h"

#include "traceclientapppacket.h"
#include "traceclient.h"


namespace Coral {


// Debug mode: Display some information on UDP connections.
// #define DEBUG


// ###### Constructor #######################################################
TraceClient::TraceClient(const char* receiverName)
{
   // ====== Default Settings ===============================================
   Status.reset();
   Sender          = NULL;
   Receiver        = NULL;
   OldPosition     = (card64)-1;
   ChangeTimeStamp = 0;
   IsPlaying       = false;
   ReceiverAddress = InternetAddress(receiverName);

   // ====== Initialize decoders ============================================
   TraceDecoder* traceDecoder = new TraceDecoder();
   if(traceDecoder == NULL) {
      cerr << "ERROR: TraceClient::TraceClient() - Out of memory!" << endl;
      exit(1);
   }
   bool a1 = Decoders.addDecoder(traceDecoder);
   DecoderSet.insert(pair<const cardinal,TraceDecoderInterface*>(0,traceDecoder));
   if(!a1) {
      cerr << "ERROR: TraceClient::TraceClient() - Out of memory!" << endl;
      exit(1);
   }
   Decoders.activate();
   Status.Encoding = Decoders.getTypeID();
}


// ###### Destructor ########################################################
TraceClient::~TraceClient()
{
   stop();

   // ====== Delete decoders ================================================
   while(DecoderSet.begin() != DecoderSet.end()) {
      multimap<const cardinal,TraceDecoderInterface*>::iterator decoderIterator =
         DecoderSet.begin();
      TraceDecoderInterface* decoder = decoderIterator->second;
      DecoderSet.erase(decoderIterator);
      Decoders.removeDecoder(decoder);
      delete decoder;
   }
}


// ###### Change media ######################################################
void TraceClient::change(const char* mediaName)
{
   if(Sender != NULL) {
      strncpy((char*)&Status.MediaName,mediaName,sizeof(Status.MediaName));
      Status.StartPosition   = 0;
      Status.RestartPosition = 0;
      ChangeTimeStamp        = getMicroTime();
      sendCommand(false);
   }
   else {
      cerr << "ERROR: TraceClient::change() - No connection!" << endl;
   }
}


// ###### Start playing #####################################################
bool TraceClient::play(const char*  server,
                       const char*  mediaName,
                       const card32 sessionDescriptor)
{
   if(Sender == NULL) {
      // ====== Set default settings ========================================
      Status.FormatID          = TraceClientAppPacket::TraceClientFormatID;
      Status.SequenceNumber    = 0;
      Status.PosChgSeqNumber   = 0;
      Status.Status            = TraceClientAppPacket::TCAS_Play;
      Status.StartPosition     = (card64)-1;
      Status.RestartPosition   = 0;
      Status.SessionDescriptor = sessionDescriptor;
      strncpy((char*)&Status.MediaName,mediaName,sizeof(Status.MediaName));
      OldPosition     = (card64)-1;
      ChangeTimeStamp = 0;
      Randomizer random;
      OurSSRC = random.random32();
      Decoders.reset();


      // ====== Get server address ==========================================
      ServerAddress = InternetAddress(server);
      if(!ServerAddress.isValid()) {
         stop();
         return(false);
      }
      if(ServerAddress.getPort() == 0) {
         ServerAddress.setPort(TraceClientAppPacket::RTPTraceDefaultPort);
      }


      // ====== Create sockets ==============================================
      SenderSocket.create(Socket::IP,Socket::UDP);
      if(!SenderSocket.ready()) {
         cerr << "ERROR: TraceClient::play() - "
                 "Unable to create socket for RTCPSender!" << endl;
         stop();
         return(false);
      }
      ReceiverSocket.create(Socket::IP,Socket::UDP);
      if(!ReceiverSocket.ready()) {
         cerr << "ERROR: TraceClient::play() - "
                 "Unable to connect socket for RTPReceiver!" << endl;
         stop();
         return(false);
      }
      if(!Socket::bindInternetSocketPair(SenderSocket,ReceiverSocket,ReceiverAddress)) {
         cerr << "ERROR: TraceClient::play() - Unable to bind socket pair!" << endl;
         stop();
         return(false);
      }


      // ====== Connect sender socket to server =============================
      Flow = SenderSocket.allocFlow(ServerAddress);
      if(Flow.getFlowLabel() != 0) {
         if(SenderSocket.connect(Flow,TraceClientDefaultTrafficClass) == 0) {
            if(SenderSocket.connect(ServerAddress,TraceClientDefaultTrafficClass) == false) {
               cerr << "ERROR: TraceClient::play() - Unable to connect socket for RTCPSender!" << endl;
               stop();
               return(false);
            }
         }
      }
      else {
         if(SenderSocket.connect(ServerAddress,TraceClientDefaultTrafficClass) == false) {
            cerr << "ERROR: TraceClient::play() - Unable to connect socket for RTCPSender!" << endl;
            stop();
            return(false);
         }
      }


      // ====== Create RTPReceiver ==========================================
      Receiver = new RTPReceiver(&Decoders,&ReceiverSocket);
      if(Receiver == NULL) {
         cerr << "ERROR: TraceClient::play() - Out of memory!" << endl;
         stop();
         return(false);
      }
      if(Receiver->start() == false) {
         cerr << "ERROR: TraceClient::play() - Unable to start RTP receiver thread!" << endl;
         stop();
         return(false);
      }


      // ====== Create RTCPSender ===========================================
      Sender = new RTCPSender(OurSSRC,&SenderSocket,Receiver,3000);
      if(Sender == NULL) {
         cerr << "ERROR: TraceClient::play() - Out of memory!" << endl;
         stop();
         return(false);
      }


      // ====== Add SDES items ==============================================
      char host[100];
      if(gethostname((char*)&host,100) == -1) {
         host[0] = 0x00;
      }
      char user[100];
      if(cuserid((char*)&user) == NULL) {
         strcpy((char*)&user,"?");
      }
      char cname[300];
      snprintf((char*)&cname,sizeof(cname),"%s@%s",user,host);
      if(Sender->addSDESItem(RTCP_SDES_CNAME,cname) == false) {
         cerr << "ERROR: TraceClient::play() - Out of memory!" << endl;
         stop();
         return(false);
      }

/*
      Sender->addSDESItem(RTCP_SDES_PHONE,"+49 1234 56789");
      Sender->addSDESItem(RTCP_SDES_LOC,"Wiehl-Forst, Germany");
      Sender->addSDESItem(RTCP_SDES_EMAIL,"email@pop3.domain.xy");
*/

      Sender->sendSDES();
      if(Sender->start() == false) {
         cerr << "ERROR: TraceClient::play() - Unable to start RTCP sender thread!" << endl;
         stop();
         return(false);
      }
      IsPlaying = true;
      sendCommand();


      // ====== Get client's local address =====================================
      // Get our local address from the server's view.
      SenderSocket.getSocketAddress(OurAddress);
      InternetAddress receiverAddress;
      ReceiverSocket.getSocketAddress(receiverAddress);
      OurAddress.setPort(receiverAddress.getPort());


      // ====== Print information ==============================================
#ifdef DEBUG
      cout << "Connecting to trace server at " << ServerAddress
           << "." << endl;
      char str[128];
      if(SenderSocket.getSendFlowLabel()) {
         snprintf((char*)&str,sizeof(str),"$%05x, traffic class $%02x.",
                 SenderSocket.getSendFlowLabel(),SenderSocket.getSendTrafficClass());
         cout << "   => IPv6 flow label is " << str << endl;
      }
      else if(SenderSocket.getSendTrafficClass()) {
         snprintf((char*)&str,sizeof(str),"$%02x.",SenderSocket.getSendTrafficClass());
         cout << "   => TOS is " << str << endl;
      }
#endif
   }
   return(IsPlaying);
}


// ###### Stop playing ######################################################
void TraceClient::stop()
{
   IsPlaying = false;
   if(Sender != NULL) {
      Sender->sendBye();
      Sender->stop();
      if(Flow.getFlowLabel() !=0) {
         SenderSocket.freeFlow(Flow);
      }
      Sender = NULL;
   }
   if(Receiver != NULL) {
      Receiver->stop();
      delete Receiver;
      Receiver = NULL;
   }
   SenderSocket.close();
   ReceiverSocket.close();
   OurAddress.reset();
   ServerAddress.reset();
   Decoders.reset();
   OldPosition     = (card64)-1;
   ChangeTimeStamp = 0;
}


// ###### Begin/end pause mode ##############################################
void TraceClient::setPause(const bool on)
{
   if(on) {
      Status.Status = TraceClientAppPacket::TCAS_Pause;
   }
   else {
      Status.Status = TraceClientAppPacket::TCAS_Play;
   }
   if(IsPlaying) sendCommand();
}


// ###### Send command to server ############################################
void TraceClient::sendCommand(const bool updateRestartPosition)
{
   if(IsPlaying) {
      Status.SequenceNumber  = Status.SequenceNumber + 1;

      if(updateRestartPosition == true)
         getPosition(); // This will update RestartPosition!

      TraceClientAppPacket app;
      app = Status;
      app.translate();
      Sender->sendApp("HELO",(void *)&app,sizeof(TraceClientAppPacket));
      if(Sender->addSDESItem(RTCP_SDES_PRIV,(char*)&app,sizeof(TraceClientAppPacket)) == false) {
         cerr << "ERROR: Unable to add SDES - Out of memory!" << endl;
      }
   }
}


// ###### Get current frame number ##########################################
card64 TraceClient::getPosition()
{
   if(IsPlaying) {
      card64 position = Decoders.getPosition();

      if(position != OldPosition) {
         if((ChangeTimeStamp == 0) || (getMicroTime() - ChangeTimeStamp > RestartPositionUpdateDelay)) {
            OldPosition     = position;
            ChangeTimeStamp = 0;
            Status.RestartPosition   = position;
            TraceClientAppPacket app = Status;
            app.translate();
            if(Sender->addSDESItem(RTCP_SDES_PRIV,(char*)&app,sizeof(TraceClientAppPacket)) == false) {
               cerr << "ERROR: Unable to add SDES - Out of memory!" << endl;
            }
         }
      }
      return(position);
   }
   return(0);
}


// ###### Set encoder #######################################################
void TraceClient::setEncoding(const cardinal index)
{
   multimap<const cardinal,TraceDecoderInterface*>::iterator decoderIterator =
      DecoderSet.find(index);
   if(decoderIterator != DecoderSet.end()) {
      TraceDecoderInterface* decoder = decoderIterator->second;
      Status.Encoding = decoder->getTypeID();
      if(Receiver) Receiver->synchronized();
      Decoders.selectDecoderForTypeID(decoder->getTypeID());
      if(Receiver) Receiver->unsynchronized();
      sendCommand();
   }
}


// ###### Get media info ####################################################
MediaInfo TraceClient::getMediaInfo() const
{
   MediaInfo mediaInfo;
   if(IsPlaying) {
      Decoders.getMediaInfo(mediaInfo);
   }
   else {
      mediaInfo.reset();
   }
   return(mediaInfo);
}


// ###### Get server SSRC ###################################################
card32 TraceClient::getServerSSRC(const cardinal layer) const
{
   if(IsPlaying) {
      SourceStateInfo ssi = Receiver->getSSI(layer);
      return(ssi.getSSRC());
   }
   return(0);
}


// ###### Get packets lost ##################################################
card64 TraceClient::getPacketsLost(const cardinal layer) const
{
   if(IsPlaying) {
      SourceStateInfo ssi = Receiver->getSSI(layer);
      return(ssi.getPacketsLost());
   }
   return(0);
}


// ###### Get fraction lost #################################################
double TraceClient::getFractionLost(const cardinal layer) const
{
   if(IsPlaying) {
      SourceStateInfo ssi = Receiver->getSSI(layer);
      return(ssi.getFractionLost());
   }
   return(0.0);
}


// ###### Get jitter ########################################################
double TraceClient::getJitter(const cardinal layer) const
{
   if(IsPlaying) {
      SourceStateInfo ssi = Receiver->getSSI(layer);
      return(ssi.getJitter());
   }
   return(0.0);
}


// ###### Get server address ################################################
String TraceClient::getServerAddressString(InternetAddress::PrintFormat format) const
{
   if(IsPlaying) {
      InternetAddress address = ServerAddress;
      address.setPrintFormat(format);
      return(address.getAddressString());
   }
   else {
      return("N/A");
   }
}


// ###### Get client address ################################################
String TraceClient::getOurAddressString(InternetAddress::PrintFormat format) const
{
   if(IsPlaying) {
      InternetAddress address = OurAddress;
      address.setPrintFormat(format);
      return(address.getAddressString());
   }
   else {
      return("N/A");
   }
}


// ###### Get encoding name #################################################
const char* TraceClient::getEncodingName(const cardinal index)
{
   multimap<const cardinal,TraceDecoderInterface*>::iterator decoderIterator =
      DecoderSet.find(index);
   if(decoderIterator != DecoderSet.end()) {
      TraceDecoderInterface* decoder = decoderIterator->second;
      return(decoder->getTypeName());
   }
   return(NULL);
}


}
