// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Decoder Implementation                                     ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "tracedecoder.h"
#include "tracepacket.h"
#include "seqnumvalidator.h"
#include "tools.h"


namespace Coral {


// Debug mode: Print received frames
// #define DEBUG


// ###### Constructor #######################################################
TraceDecoder::TraceDecoder()
   : Synchronizable()
{
   // SeqNumValidator with maxMisorder = 1
   //      => all packets not in sequence will be dropped!
   // minSequential = 1 => Check is done by RTPReceiver -> unnecessary here!
   for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
      SeqNumber[i] = SeqNumValidator(1,1);
      FrameID[i]   = (cardinal)-1;
      FrameSize[i] = 0;
   }
   reset();
}


// ###### Destructor ########################################################
TraceDecoder::~TraceDecoder()
{
   deactivate();
}


// ###### Activate decoder ##################################################
void TraceDecoder::activate()
{
   reset();
}


// ###### Deactivate decoder ################################################
void TraceDecoder::deactivate()
{
   reset();
}


// ###### Reset decoder #####################################################
void TraceDecoder::reset()
{
   FrameRate       = 0.0;
   Utilization     = -HUGE_VAL;
   StreamPriority  = -128;
   SessionPriority = -128;
   Position        = 0;
   MaxPosition     = 0;
   ErrorCode       = 0;
   Media.reset();
   for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
      SeqNumber[i].reset();
      FrameID[i]   = (cardinal)-1;
      FrameSize[i] = 0;
   }
}


// ###### Get MediaInfo #####################################################
void TraceDecoder::getMediaInfo(MediaInfo& mediaInfo) const
{
   ((TraceDecoder*)this)->synchronized();
   mediaInfo = Media;
   ((TraceDecoder*)this)->unsynchronized();
}


// ###### Get error code ####################################################
card8 TraceDecoder::getErrorCode() const
{
   ((TraceDecoder*)this)->synchronized();
   const card8 errorCode = ErrorCode;
   ((TraceDecoder*)this)->unsynchronized();
   return(errorCode);
}


// ###### Get frame rate ####################################################
double TraceDecoder::getFrameRate() const
{
   ((TraceDecoder*)this)->synchronized();
   const double frameRate = FrameRate;
   ((TraceDecoder*)this)->unsynchronized();
   return(frameRate);
}

// ###### Get utilization ###################################################
double TraceDecoder::getUtilization() const
{
   ((TraceDecoder*)this)->synchronized();
   const double utilization = Utilization;
   ((TraceDecoder*)this)->unsynchronized();
   return(utilization);
}


// ###### Get bandwidth #####################################################
Range<cardinal> TraceDecoder::getBandwidth() const
{
   ((TraceDecoder*)this)->synchronized();
   const Range<cardinal> bandwidthRange(MinBandwidth,Bandwidth,MaxBandwidth);
   ((TraceDecoder*)this)->unsynchronized();
   return(bandwidthRange);
}


// ###### Get stream priority ###############################################
int8 TraceDecoder::getStreamPriority() const
{
   ((TraceDecoder*)this)->synchronized();
   const int8 priority = StreamPriority;
   ((TraceDecoder*)this)->unsynchronized();
   return(priority);
}

// ###### Get session priority ###############################################
int8 TraceDecoder::getSessionPriority() const
{
   ((TraceDecoder*)this)->synchronized();
   const int8 priority = SessionPriority;
   ((TraceDecoder*)this)->unsynchronized();
   return(priority);
}


// ###### Get position ######################################################
card64 TraceDecoder::getPosition() const
{
   ((TraceDecoder*)this)->synchronized();
   const card64 position = Position;
   ((TraceDecoder*)this)->unsynchronized();
   return(position);
}


// ###### Get maximum position ##############################################
card64 TraceDecoder::getMaxPosition() const
{
   ((TraceDecoder*)this)->synchronized();
   const card64 maxPosition = MaxPosition;
   ((TraceDecoder*)this)->unsynchronized();
   return(maxPosition);
}


// ###### Get decoder's type ID #############################################
const card16 TraceDecoder::getTypeID() const
{
   return(TracePacket::TraceTypeID);
}


// ###### Get decoder's type name ###########################################
const char* TraceDecoder::getTypeName() const
{
   return((const char*)&TracePacket::TraceTypeName);
}


// ###### Check next packet #################################################
bool TraceDecoder::checkNextPacket(DecoderPacket* decoderPacket)
{
   // ====== Check packet ===================================================
   TracePacket* packet = (TracePacket*)decoderPacket->Buffer;
   if(translate32(packet->FormatID) != TracePacket::TraceFormatID) {
      return(false);
   }

   cardinal layer = packet->Layer;
   cardinal layers = packet->Layers;
   if((layer >= RTPConstants::RTPMaxQualityLayers)  ||
      (layers >= RTPConstants::RTPMaxQualityLayers) ||
      (layer >= layers)) {
      cerr << "WARNING: TraceDecoder::checkNextPacket() - Invalid layer values!"
           << endl;
      return(false);
   }

   // ====== Check, if packet is in sequence ================================
   SeqNumValidator::ValidationResult valid =
      SeqNumber[layer].validate((card64)decoderPacket->SequenceNumber);
   if(valid != SeqNumValidator::Valid) {
      return(false);
   }

   // ====== Set layers and layer ===========================================
   decoderPacket->Layer  = (card8)layer;
   decoderPacket->Layers = (card8)layers;
   return(true);
}


// ###### Handle next packet ################################################
void TraceDecoder::handleNextPacket(const DecoderPacket* decoderPacket)
{
   synchronized();

   // ====== Translate packet's byte order ==================================
   TracePacket* packet = (TracePacket*)decoderPacket->Buffer;
   packet->translate();

   ErrorCode   = packet->ErrorCode;
   Position    = packet->Position;
   MaxPosition = packet->MaxPosition;

   // ====== Check, if packet contains additional information ===============
   if(packet->Flags == TracePacket::TF_Information) {
      TracePacketData* tpd = (TracePacketData*)&packet->Data[0];
      tpd->Information.translate();
      Media           = tpd->Information;
      FrameRate       = translateToDouble(tpd->FrameRate);
      Utilization     = translateToDouble(tpd->Utilization);
      Bandwidth       = translate32(tpd->Bandwidth);
      MinBandwidth    = translate32(tpd->MinBandwidth);
      MaxBandwidth    = translate32(tpd->MaxBandwidth);
      StreamPriority  = tpd->StreamPriority;
      SessionPriority = tpd->SessionPriority;
   }

   if(ErrorCode == 0) {
      if(FrameID[decoderPacket->Layer] == packet->FrameID) {
         FrameSize[decoderPacket->Layer] += decoderPacket->Length -
                                               sizeof(TracePacket);
      }
      else {
#ifdef DEBUG
         if(FrameID[decoderPacket->Layer] != 0) {
            char str[256];
            const card64 seconds  = packet->Position / PositionStepsPerSecond;
            snprintf((char*)&str,sizeof(str),"#%d/%02Ld.%02Ld.%02Ld:   %6d   #%05d",
               decoderPacket->Layer,
               (seconds / 60),
               (seconds % 60),
               (packet->Position % PositionStepsPerSecond) / (PositionStepsPerSecond / 100),
               (unsigned int)FrameSize[decoderPacket->Layer],
               (unsigned int)FrameID[decoderPacket->Layer]);
            cerr << str << endl;
         }
#endif
         FrameID[decoderPacket->Layer]   = packet->FrameID;
         FrameSize[decoderPacket->Layer] = decoderPacket->Length -
                                              sizeof(TracePacket);
      }
   }

   unsynchronized();
}


}
