// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Frame Size Scalability                                     ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "traceframesizescalability.h"
#include "utilityfunction.h"


namespace Coral {


// ###### Constructor #######################################################
TraceFrameSizeScalability::TraceFrameSizeScalability()
{
   TraceReader = NULL;
}


// ###### Desstructor #######################################################
TraceFrameSizeScalability::~TraceFrameSizeScalability()
{
}


// ###### Initialize ########################################################
void TraceFrameSizeScalability::initFrameSizeScalability(
                                   TDTFReader*    traceReader,
                                   const card64   position,
                                   const cardinal layer,
                                   const double   scaleFactor)
{
   if((scaleFactor < 0.0) || (scaleFactor > 1.0)) {
      cerr << "ERROR: TraceFrameSizeScalability::initFrameSizeScalability() - " << endl
           << "Invalid parameter " << scaleFactor << "!" << endl;
      exit(1);
   }

   TraceReader = traceReader;
   Position    = position;
   Layer       = layer;
   ScaleFactor = scaleFactor;
}


// ###### Get frame size scalability class name #############################
const char* TraceFrameSizeScalability::getFrameSizeScalabilityClass() const
{
   return((const char*)"Trace Frame-size Scalability");
}


// ###### Check, if frame size is scalable ##################################
bool TraceFrameSizeScalability::isFrameSizeScalable() const
{
   return(true);
}


// ###### Check, if frame size is variable bit rate #########################
bool TraceFrameSizeScalability::isVariableBitrate() const
{
   return(true);
}


// ###### Get minimum frame size ############################################
cardinal TraceFrameSizeScalability::getMinPayloadFrameSizeForDelay(
                                         const double   frameRate,
                                         const cardinal bufferDelay) const
{
   const cardinal max = getMaxPayloadFrameSizeForDelay(frameRate,bufferDelay);
   return((cardinal)ceil((double)max * ScaleFactor));
}


// ###### Get maxmum frame size ############################################
cardinal TraceFrameSizeScalability::getMaxPayloadFrameSizeForDelay(
                                         const double   frameRate,
                                         const cardinal bufferDelay) const
{
   return(TraceReader->getPayloadFrameSizeForDelay(
             Position, Layer, frameRate, bufferDelay));
}


// ###### Get frame size utilization ########################################
double TraceFrameSizeScalability::getPayloadFrameSizeUtilizationForDelayAndSize(
                                               const double   frameRate,
                                               const cardinal bufferDelay,
                                               const cardinal frameSize) const
{
   // ====== Get utilization header =========================================
   const UtilizationHeader* uh =
      TraceReader->getFrameSizeUtilizationHeader(
         TraceReader->positionToFramePosition(frameRate,Position),
         Layer, frameRate);
   if(uh == NULL) {
      return(0.0);
   }

   const double scaleFactor = getPayloadFrameSizeScaleFactorForDelayAndSize(frameRate,bufferDelay,frameSize);
   return(evaluateUtilityFunctionTranslated(translate16(uh->Type),
                                            scaleFactor,
                                            (card64*)&uh->Constant,
                                            uh->Constants));
}


// ###### Get frame rate utilization ########################################
double TraceFrameSizeScalability::getFrameSizeUtilizationWeight(const double frameRate) const
{
   // ====== Get utilization header =========================================
   const UtilizationHeader* uh =
      TraceReader->getFrameSizeUtilizationHeader(
         TraceReader->positionToFramePosition(frameRate,Position),
         Layer, frameRate);
   if(uh == NULL) {
      return(0.0);
   }

   return(translateToDouble(uh->Weight));
}


// ###### Get maximum frame count ###########################################
cardinal TraceFrameSizeScalability::getMaxFrameCountForDelay(
                                         const double   frameRate,
                                         const cardinal bufferDelay) const
{
   return(TraceReader->getMaxFrameCountForDelay(
             Position, Layer, frameRate, bufferDelay));
}


// ###### Get maximum buffer delay ##########################################
cardinal TraceFrameSizeScalability::getMaxBufferDelay(const double frameRate) const
{
   return(TraceReader->getMaxBufferDelay(Position,frameRate));
}


}
