// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### MPEG Trace Array Implementation                                  ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "mpegtracearray.h"
#include "randomizer.h"


#include <fstream.h>


namespace Coral {


// ###### Constructor #######################################################
MPEGTraceArray::MPEGTraceArray(const TraceConfiguration& config)
   : TraceArray(config)
{
}


// ###### Load MPEG trace ###################################################
bool MPEGTraceArray::load(const char*    name,
                          const char*    framePattern,
                          const cardinal frameRate,
                          const cardinal extLayers,
                          const double   fakeE1,
                          const double   fakeE2)
{
   // ====== Check pattern ==================================================
   const cardinal patternSize = strlen(framePattern);
   if(patternSize < 1) {
      cerr << "ERROR: MPEGTraceArray::load() - Empty frame pattern!" << endl;
      return(false);
   }
   cardinal       pattern[patternSize + 1];
   const cardinal layers = 3 + (3 * extLayers);

   for(cardinal i = 0;i < patternSize;i++) {
      switch(toupper(framePattern[i])) {
         case 'I':
            pattern[i] = LayerMPEGBaseI;
          break;
         case 'P':
            pattern[i] = LayerMPEGBaseP;
          break;
         case 'B':
            pattern[i] = LayerMPEGBaseB;
          break;
         default:
            cerr << "ERROR: MPEGTraceArray::load() - Illegal frame type " << pattern[i] << "!" << endl;
            return(false);
          break;
      }
   }
   pattern[patternSize] = (cardinal)-1;


   // ====== Open trace file, allocate memory ===============================
   ifstream is(name);
   if(is.good()) {
      // ====== Count frames ================================================
      cardinal frameCount  = 0;
      cardinal id          = 0;
      cardinal frameSizeB  = 0;
      cardinal frameSizeE1 = 0;
      cardinal frameSizeE2 = 0;
      cardinal idB  = 0;
      cardinal idE1 = 0;
      cardinal idE2 = 0;
      while(!is.eof()) {
         is >> frameSizeB;
         if(extLayers > 0) {
            if(fakeE1 == 0.0) {
               is >> frameSizeE1;
            }
            if(extLayers > 1) {
               if(fakeE2 == 0.0) {
                  is >> frameSizeE1;
               }
            }
         }
         frameCount++;
      }
      is.close();
      is.open(name);

      // ====== Initialize ==================================================
      cardinal patternPos = 0;
      init(layers,frameCount);
      FrameRate = frameRate;

      // ====== Read frame sizes ============================================
      for(;;) {
         if(Frames >= MaxFrames) {
            cerr << "ERROR: MPEGTraceArray::load() - Too many frames!" << endl;
            return(false);
         }

         // ====== Read frame size ==========================================
         is >> frameSizeB;
         if(is.eof()) {
            break;
         }
         idB = id++;
         if(extLayers > 0) {
            if(fakeE1 != 0.0) {
               frameSizeE1 = (cardinal)ceil(fakeE1 * (double)frameSizeB);
            }
            else {
               is >> frameSizeE1;
               if(is.eof()) {
                  break;
               }
            }
            idE1 = id++;
            if(extLayers > 1) {
               if(fakeE2 != 0.0) {
                  frameSizeE2 = (cardinal)ceil(fakeE2 * (double)frameSizeB);
               }
               else {
                  is >> frameSizeE2;
                  if(is.eof()) {
                     break;
                  }
               }
               idE2 = id++;
            }
         }

         // ====== Determine frame type =====================================
         if(pattern[patternPos] == (cardinal)-1) {
            patternPos = 0;
         }
         const cardinal frameLayer = (cardinal)pattern[patternPos];

         // ====== Write trace into array ===================================
         for(cardinal i = 0;i < MaxLayers;i++) {
            LayerTrace[i]->Frame[Frames].ID   = (cardinal)-1;
            LayerTrace[i]->Frame[Frames].Size = 0;
            LayerTrace[i]->Frames++;
         }
         LayerTrace[frameLayer]->Frame[Frames].ID   = idB;
         LayerTrace[frameLayer]->Frame[Frames].Size = frameSizeB;
         if(extLayers > 0) {
            LayerTrace[frameLayer + 3]->Frame[Frames].ID   = idE1;
            LayerTrace[frameLayer + 3]->Frame[Frames].Size = frameSizeE1;
            if(extLayers > 1) {
               LayerTrace[frameLayer + 6]->Frame[Frames].ID   = idE2;
               LayerTrace[frameLayer + 6]->Frame[Frames].Size = frameSizeE2;
            }
         }

         Frames++;
         patternPos++;
      }

      // ====== Ensure correct alignment ====================================
      const cardinal diff = Frames % (cardinal)rint(FrameRate);
      Frames -= diff;
      return(Frames > 0);
   }
   else {
      cerr << "ERROR: MPEGTraceArray::load() - Unable to load trace " << name << "!" << endl;
      return(false);
   }
}


// ###### Decrease frame rate of MPEG trace by 1 frame/s ####################
TraceArray* MPEGTraceArray::decreaseFrameRate() const
{
   // Set random seed to constant value; this will cause all invokations
   // of the funtion to generate the same output for the same input!
   Randomizer random;
   random.setSeed(12345678);


   // ====== Initialize =====================================================
   const cardinal frameRate = (cardinal)rint(FrameRate);
   if(frameRate <= 1.0) {
      return(NULL);
   }
   const cardinal frames    = Frames;
   const cardinal layers    = Layers;
   bool skipArray[frames];
   for(cardinal i = 0;i < frames;i++) {
      skipArray[i] = false;
   }


   // ====== Step 1: Drop B-frames ==========================================
   cardinal newFrames = frames;
   for(cardinal i = 0;i < frames;i += frameRate) {
      for(cardinal j = i;j < min(frames,i + frameRate);j++) {
         if(LayerTrace[LayerMPEGBaseB]->Frame[j].Size > 0) {
            skipArray[j] = true;
            newFrames--;
            break;
         }
      }
   }

   // ====== Step 1b: Not enough B-frames -> Drop random P-frames ===========
   if(newFrames < frames) {
      const cardinal calculatedFrames = frames - (frames / frameRate);
      integer diff = (integer)newFrames - (integer)calculatedFrames;
      if(diff > 0) {
         cardinal index[frames];
         cardinal counter  = 0;
         cardinal nextI    = (cardinal)-1;
         bool     newBlock = false;
         for(integer i = (integer)newFrames - 1;i >= 0;i--) {
            if(LayerTrace[LayerMPEGBaseI]->Frame[i].Size > 0) {
               nextI = i;
            }
            if(((i + 1) % frameRate) == 0) {
               newBlock = true;
            }
            if(newBlock == true) {
               if(LayerTrace[LayerMPEGBaseP]->Frame[i].Size > 0) {
                  index[counter] = i;
                  counter++;
                  newBlock = false;
               }
            }
         }

         do {
            cardinal r = random.random32() % counter;
            if(skipArray[index[r]] == false) {
               skipArray[index[r]] = true;
               newFrames--;
            }
         } while(newFrames > calculatedFrames);
      }
   }


   // ====== Step 2: Drop P-frames ==========================================
   else {
      cardinal nextI    = (cardinal)-1;
      bool     newBlock = false;

      for(integer i = (integer)frames - 1;i >= 0;i--) {
         if(LayerTrace[LayerMPEGBaseI]->Frame[i].Size > 0) {
            nextI = i;
         }
         if(((i + 1) % frameRate) == 0) {
            newBlock = true;
         }

         if(newBlock == true) {
            if(LayerTrace[LayerMPEGBaseP]->Frame[i].Size > 0) {
               skipArray[i] = true;
               newFrames--;
               newBlock     = false;
            }
         }
      }


      // ====== Step 2b: Not enough P-frames -> Drop random I-frames ========
      if(newFrames < frames) {
         const cardinal calculatedFrames = frames - (frames / frameRate);
         integer diff = (integer)newFrames - (integer)calculatedFrames;
         if(diff > 0) {
            cardinal index[frames];
            cardinal counter = 0;
            for(cardinal i = 0;i < frames;i++) {
               if(LayerTrace[LayerMPEGBaseI]->Frame[i].Size > 0) {
                  index[counter] = i;
                  counter++;
               }
            }

            do {
               cardinal r = random.random32() % counter;
               if(skipArray[index[r]] == false) {
                  skipArray[index[r]] = true;
                  newFrames--;
               }
            } while(newFrames > calculatedFrames);
         }
      }


      // ====== Step 3: Drop I-frames =======================================
      else if(newFrames == frames) {
         for(cardinal i = 0;i < frames;i += frameRate) {
            for(cardinal j = i;j < min(frames,i + frameRate);j++) {
               if(LayerTrace[LayerMPEGBaseI]->Frame[j].Size > 0) {
                  skipArray[j] = true;
                  newFrames--;
                  break;
               }
            }
         }
      }
   }


   // ====== Allocate trace array ===========================================
   TraceArray* newTraceArray = new MPEGTraceArray(Config);
   if(newTraceArray == NULL) {
      cerr << "ERROR: MPEGTraceArray::decreaseFrameRate() - Out of memory!" << endl;
      exit(1);
   }
   newTraceArray->init(layers,newFrames);
   newTraceArray->Layers    = layers;
   newTraceArray->Frames    = newFrames;
   newTraceArray->FrameRate = frameRate - 1;


   // ====== Copy selected elements =========================================
   cardinal j = 0;
   for(cardinal i = 0;i < frames;i++) {
      if(skipArray[i] == false) {
         for(cardinal k = 0;k < layers;k++) {
            newTraceArray->LayerTrace[k]->Frame[j].ID   = LayerTrace[k]->Frame[i].ID;
            newTraceArray->LayerTrace[k]->Frame[j].Size = LayerTrace[k]->Frame[i].Size;
         }
         j++;
      }
   }

   return(newTraceArray);
}


}
