// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### RTCP Receiver Implementation                                     ####
// ####                                                                  ####
// #### Version 1.00  --  September 17, 2000                             ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "rtcpreceiver.h"
#include "internetflow.h"
#include "tools.h"


namespace Coral {


// Debug mode: Print corrupted RTCP packets
#define DEBUG


// ###### Constructor #######################################################
RTCPReceiver::RTCPReceiver()
   : Thread()
{
   Server         = NULL;
   ReceiverSocket = NULL;
}


// ###### Constructor #######################################################
RTCPReceiver::RTCPReceiver(RTCPAbstractServer* server, Socket* receiverSocket)
   : Thread()
{
   init(server,receiverSocket);
}


// ###### Initialize ########################################################
void RTCPReceiver::init(RTCPAbstractServer* server, Socket* receiverSocket)
{
   Server         = server;
   ReceiverSocket = receiverSocket;
}


// ###### Destructor ########################################################
RTCPReceiver::~RTCPReceiver()
{
   stop();
}


// ###### Receiver loop #####################################################
void RTCPReceiver::run()
{
   if(ReceiverSocket == NULL) {
      cerr << "ERROR: RTCPReceiver::run() - RTCPReceiver is uninitialized!"
           << endl;
      return;
   }

   InternetFlow flow;
   for(;;) {

      // ====== Read RTCP packet header from socket =========================
      RTCPCommonHeader header;
      if(ReceiverSocket->receive(&header,sizeof(RTCPCommonHeader),MSG_PEEK) != sizeof(RTCPCommonHeader)) {
         cerr << "WARNING: RTCPReceiver::run() - Received bad RTCP header" << endl;
         continue;
      }

      // ====== Read complete RTCP packet from socket =======================
      cardinal packetSize = header.getLength();
      char     packetData[packetSize];

/*
      cout << "RTCP Common Header\n";
      cout << "   Version     : " << header.getVersion()    << "\n";
      cout << "   Padding     : " << header.getPadding()    << "\n";
      cout << "   Count       : " << header.getCount()      << "\n";
      cout << "   Packet Type : " << header.getPacketType() << "\n";
      cout << "   Length      : " << header.getLength()     << endl;                 
*/

      // ====== Verify RTCP packet ==========================================
      integer receivedPacketSize =
         ReceiverSocket->receiveFrom((void*)&packetData,packetSize,flow);
      if(receivedPacketSize != (integer)packetSize) {
#ifdef DEBUG
         cerr << "RTCP packet: receivedPacketSize != packetSize: "
              << receivedPacketSize << " <=> " << packetSize << endl;
#endif
         continue;
      }

      if(header.getVersion() != RTPConstants::RTPVersion) {
#ifdef DEBUG
         cerr << "RTCP packet: Invalid RTP version: " << header.getVersion() << endl;
#endif
         continue;
      }

      RTCPCommonHeader* r    = (RTCPCommonHeader*)&packetData;
      RTCPCommonHeader* rend = (RTCPCommonHeader*)(long)r + packetSize;
      do {
         r = (RTCPCommonHeader*)((long*)r + (long)r->getLength());
      } while((r < rend) && (r->getVersion() == RTPConstants::RTPVersion));
      if(r != rend) {
#ifdef DEBUG
         cerr << "RTCP packet: Length check failed!" << endl;
#endif
         continue;
      }


      // ====== Get type and invoke server function =========================
      synchronized();
      switch(header.getPacketType()) {

         // ====== Packet is a Receiver Report ==============================
         case RTCP_RR:
            {
               RTCPReceiverReport* receiverReport = (RTCPReceiverReport*)&packetData;
               cardinal bytes = sizeof(RTCPReceiverReport);
               cardinal layer = 0;
               card32   ssrc  = 0;
               for(cardinal i = 0;i < receiverReport->getCount();i++) {
                  if((bytes + sizeof(RTCPReceptionReportBlock)) <= packetSize) {
                     if(receiverReport->rr[i].getSSRC() == ssrc) {
                        layer++;
                     }
                     else {
                        layer = 0;
                        ssrc  = receiverReport->getSSRC();
                     }
                     Server->receivedReceiverReport(
                        flow, receiverReport->getSSRC(),
                        &receiverReport->rr[i], layer);
                  }
                  else {
#ifdef DEBUG
                     cerr << "RTCP packet: Invalid receiver report length!" << endl;
#endif
                     break;
                  }
                  bytes += sizeof(RTCPReceptionReportBlock);
               }
            }
          break;

         // ====== Packet is a Sender Report ================================
         case RTCP_SR:
            {
               RTCPSenderReport* senderReport = (RTCPSenderReport*)&packetData;
               cardinal bytes = (long)&senderReport->rr[0] - (long)senderReport;
               cardinal layer = 0;
               card32   ssrc  = 0;
               for(cardinal i = 0;i < senderReport->getCount();i++) {
                  if((bytes + sizeof(RTCPReceptionReportBlock)) <= packetSize) {
                     if(senderReport->rr[i].getSSRC() == ssrc) {
                        layer++;
                     }
                     else {
                        layer = 0;
                        ssrc  = senderReport->getSSRC();
                     }
                     Server->receivedSenderReport(
                        flow, senderReport->getSSRC(),
                        &senderReport->rr[i], layer);
                  }
                  else {
#ifdef DEBUG
                     cerr << "RTCP packet: Invalid sender report length!" << endl;
#endif
                     break;
                  }
                  bytes += sizeof(RTCPReceptionReportBlock);
               }
            }
          break;

         // ====== Packet is a Source Description ===========================
         case RTCP_SDES:
            {
               RTCPSourceDescription* sdes          = (RTCPSourceDescription*)&packetData;
               const RTCPSourceDescriptionItem* end = (RTCPSourceDescriptionItem*)((long)sdes + sdes->getLength());
               RTCPSourceDescriptionChunk* sd       = &sdes->Chunk[0];
               RTCPSourceDescriptionItem* rsp;
               RTCPSourceDescriptionItem* rspn;
               integer count = sdes->getCount();
               while(--count >= 0) {
                  rsp = &sd->Item[0];
                  if(rsp >= end) {
                     break;
                  }
                  for ( ;rsp->Type;rsp = rspn) {
                     rspn = (RTCPSourceDescriptionItem*)((char*)rsp + rsp->Length + 2);
                     if(rspn >= end) {
                        rsp = rspn;
                        break;
                     }
                     Server->receivedSourceDescription(
                         flow, sd->SRC, rsp->Type, rsp->Data, rsp->Length);
                  }
                  rsp = (RTCPSourceDescriptionItem*)((long)sd + (((char*)rsp - (char*)sd) >> 2) + 1);
               }
            }
          break;

         // ====== Packet is a Bye message ==================================
         case RTCP_BYE:
            {
               RTCPBye* bye = (RTCPBye*)&packetData;
               for(cardinal i = 0;i < bye->getCount();i++) {
                  Server->receivedBye(flow, bye->getSource(i),
                                      RTCPAbstractServer::DeleteReason_UserBye);
               }
            }
           break;

         // ====== Packet is an App message =================================
         case RTCP_APP:
            {
               RTCPApp* app = (RTCPApp*)&packetData;
               Server->receivedApp(flow,
                                   app->getSource(),
                                   app->getName(),
                                   (void*)app->getData(),
                                   packetSize - sizeof(RTCPCommonHeader) - 8);
            }
           break;

         // ====== Packet type is unknown ===================================
         default:
            receivedPacketSize = 0;
#ifdef DEBUG
            cerr << "RTCP packet: Unknown SDES type "
                 << header.getPacketType() << endl;
#endif
          break;

      }
      unsynchronized();
      AverageRTCPSize = (1.0/16.0) * receivedPacketSize + (15.0/16.0) * AverageRTCPSize;
   }
}


}
