// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Array Implementation                                       ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "tracearray.h"


namespace Coral {


// ###### Constructor #######################################################
TraceArray::TraceArray(const TraceConfiguration& config)
{
   LayerTrace = NULL;
   Config     = config;
   Layers     = 0;
   Frames     = 0;
   MaxLayers  = 0;
   MaxFrames  = 0;
}


// ###### Destructor ########################################################
TraceArray::~TraceArray()
{
   if(LayerTrace != NULL) {
      for(cardinal i = 0;i < MaxLayers;i++) {
         if(LayerTrace[i] != NULL) {
            delete LayerTrace[i];
         }
      }
      delete LayerTrace;
   }
   LayerTrace = NULL;
}


// ###### Initialize ########################################################
void TraceArray::init(const cardinal maxLayers,
                      const cardinal maxFrames)
{
   // ====== Remove old trace ===============================================
   if(LayerTrace != NULL) {
      for(cardinal i = 0;i < MaxLayers;i++) {
         if(LayerTrace[i] != NULL) {
            delete LayerTrace[i];
         }
      }
      delete LayerTrace;
   }

   // ====== Initialize =====================================================
   Layers    = maxLayers;
   Frames    = 0;
   MaxLayers = maxLayers;
   MaxFrames = maxFrames;

   // ====== Allocate memory for new trace ==================================
   LayerTrace = new Trace*[MaxLayers];
   if(LayerTrace == NULL) {
      cerr << "ERROR: TraceArray::init() - Out of memory!" << endl;
      exit(1);
   }
   for(cardinal i = 0;i < MaxLayers;i++) {
      LayerTrace[i] = NULL;
   }

   for(cardinal i = 0;i < MaxLayers;i++) {
      LayerTrace[i] = (Trace*)new char[sizeof(Trace) + maxFrames * sizeof(FrameDescription)];
      if(LayerTrace[i] == NULL) {
         cerr << "ERROR: TraceArray::init() - Out of memory!" << endl;
         exit(1);
      }
   }
}


// ###### Decrease frame rate by one step ###################################
TraceArray* TraceArray::decreaseFrameRate() const
{
   return(NULL);
}


// ###### Calculate one empirical envelope point ############################
cardinal TraceArray::calculateEmpiricalEnvelopePoint(
                        const cardinal layer,
                        const cardinal start,
                        const cardinal end,
                        const cardinal traceStart,
                        const cardinal traceEnd,
                        const cardinal delay,
                        const bool     frameCount = false) const
{
   cardinal istart;
   cardinal iend;

   if(start - traceStart >= delay - 1) {
      istart = traceStart + start - (delay - 1);
   }
   else {
      istart = traceStart;
   }

   if(end + delay - 1 < traceEnd) {
       iend = end + delay - 1;
   }
   else {
       iend = traceEnd;
   }

   cardinal max = 0;
   for(cardinal i = istart;i <= iend;i++) {
      cardinal sum = 0;
      for(cardinal j = i;j <= min(i + (delay - 1),iend);j++) {
         if(!frameCount) {
            sum += LayerTrace[layer]->Frame[j].Size;
         }
         else {
            if(LayerTrace[layer]->Frame[j].Size > 0) {
               sum++;
            }
         }
      }
      if(sum > max) {
         max = sum;
      }
   }

   return(max);
}


// ###### Calculate empirical envelope ######################################
void TraceArray::calculateEmpiricalEnvelope(
                    const cardinal     layer,
                    const cardinal     start,
                    const cardinal     end,
                    const cardinal     traceStart,
                    const cardinal     traceEnd,
                    EmpiricalEnvelope* ee,
                    const bool         frameCount = false) const
{
   for(cardinal i = 0;i < ee->Pairs;i++) {
      ee->Pair[i].Sum = calculateEmpiricalEnvelopePoint(
                           layer,
                           start, end,
                           traceStart, traceEnd,
                           ee->Pair[i].Interval,
                           frameCount);
   }
}


// ###### Initialize intervals of empirical envelope ########################
EmpiricalEnvelope* TraceArray::initEmpiricalEnvelope(
                                  char*          buffer,
                                  const cardinal eePairs,
                                  const bool     frameCount = false) const
{
   // ====== Initialize =====================================================
   const cardinal delay =
      max(1,(cardinal)ceil(((double)Config.MaxBufferDelay * Config.FrameRate)
                             / 1000000.0));
   const double step = (double)delay / (double)(eePairs - 1);
   EmpiricalEnvelope* ee = (EmpiricalEnvelope*)buffer;

   // ====== Calculate intervals using MaxBufferDelay =======================
   ee->Pair[0].Interval = 1;
   ee->Pairs            = 1;
   for(cardinal i = 1;i < (cardinal)(eePairs - 1);i++) {
      ee->Pair[ee->Pairs].Interval = (card16)ceil((double)i * step);
      if(ee->Pair[ee->Pairs].Interval > ee->Pair[ee->Pairs - 1].Interval) {
         ee->Pairs++;
      }
   }
   ee->Pair[ee->Pairs].Interval = delay;
   if(ee->Pair[ee->Pairs].Interval > ee->Pair[ee->Pairs - 1].Interval) {
      ee->Pairs++;
   }
   return(ee);
}


// ###### Calculate optimal intervals using Gumbrich's algorithm ############
void TraceArray::calculateOptimalIntervals(
                    double*         costArray,
                    cardinal*       lengthArray,
                    const cardinal  minLength,
                    const cardinal  maxLength) const
{
   // ====== Initialize =====================================================
   char str[256];
   for(cardinal i = 0;i < Frames;i++) {
      costArray[i]   = HUGE_VAL;
      lengthArray[i] = 0;
   }


   // ====== Calculate last interval ========================================
   const cardinal last = (cardinal)((integer)max((integer)Frames - (integer)minLength,(integer)0));
   costArray[last]   = calculateTrafficCost(Frames - 1,Frames - last);
   lengthArray[last] = Frames - last;
   for(cardinal i = last + 1;i < Frames;i++) {
      lengthArray[i] = Frames - i;
      costArray[i]   = calculateTrafficCost(Frames - 1,Frames - i);
   }

   // printf("<Min=%d  Max=%d>\n",minLength,maxLength);


   // ====== Execute Gumbrich's algorithm ([Gumb98]) ========================
   cardinal count = 0;
   cout << "     0 -> ";
   cout.flush();
   for(integer i = (integer)Frames - 1;i >= 0;i--) {
      const cardinal startLength =
         (cardinal)min((integer)minLength,(integer)Frames - 1 - i);
      for(cardinal length = startLength;(cardinal)i + length < Frames;length++) {
         // ====== Compare cost =============================================
         const double cost =
            Config.RemappingCost           +
            calculateTrafficCost(i,length) +
            costArray[i + length];

         // ====== Use new length, if it is cheapest setting ================
         // Use "<=". This will generate intervals as long as possible!
         if(cost <= costArray[i]) {
            costArray[i]   = cost;
            lengthArray[i] = length;
         }

         // ====== Check, if maximum length is reached ======================
         if(length >= maxLength) {
            break;
         }
      }

      // ====== Print progress ==============================================
      count++;
      if((count % 10) == 0) {
         cout << ".";
         cout.flush();
         if((count % 100) == 0) {
            if((count % 500) != 0) {
               cout << " ";
               cout.flush();
            }
            else {
               snprintf((char*)&str,sizeof(str)," [%1.1f%%]",
                           (double)count * 100.0 / Frames);
               cout << str << endl;
               snprintf((char*)&str,sizeof(str),"%6d -> ",count);
               cout << str;
               cout.flush();
            }
         }
      }
   }
   cout << endl;

/*
   for(cardinal i = 0;i < Frames-1;i++) {
      printf("I%05dd:  %4d  %f\n",i,lengthArray[i],costArray[i]);
   }
   // exit(1);
*/
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const TraceArray& traceArray)
{
   char str[256];

   printf("L=%d   %d\n",traceArray.Layers,traceArray.Frames);


   for(cardinal i = 0;i < traceArray.Frames;i++) {
      snprintf((char*)&str,sizeof(str),"#%06d:   ",i);
      os << str;
      for(cardinal j = 0;j < traceArray.Layers;j++) {
         snprintf((char*)&str,sizeof(str),"%7d  ",traceArray.LayerTrace[j]->Frame[i].Size);
         os << str;
      }
      os << endl;
   }
   return(os);
}


}
