// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Configuration Implementation                               ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "traceconfiguration.h"


namespace Coral {


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const TraceConfiguration& config)
{
   config.print(os,false);
   return(os);
}


// ###### Print TraceConfiguration ##########################################
void TraceConfiguration::print(ostream&   os,
                               const bool utilizationOnly) const
{
   if(utilizationOnly != true) {
      cout << "Frame Rate               = " << FrameRate        << endl;
      cout << "Frame Pattern            = " << FramePattern     << endl;
      cout << "Min Interval Length      = " << (MinIntervalLength / 1000000.0) << " [s]" << endl;
      cout << "Max Interval Length      = " << (MaxIntervalLength / 1000000.0) << " [s]" << endl;
      cout << "Extension Layers         = " << ExtLayers        << endl;
      cout << "Fake 1st Ext. Layer      = " << FakeE1           << endl;
      cout << "Fake 2nd Ext. Layer      = " << FakeE2           << endl;
      cout << "Max Buffer Delay         = " << (MaxBufferDelay / 1000000.0) << " [s]" << endl;
      cout << "Remapping Cost           = " << RemappingCost    << endl;
   }
   cout << "Resource/Utilization Pts = " << RUPoints                   << endl;
   cout << "Bandwidth Threshold      = " << BandwidthThreshold         << endl;
   cout << "Utilization Threshold    = " << UtilizationThreshold       << endl;
   char str[256];
   snprintf((char*)&str,sizeof(str),"$%04x",FrameRateUtilizationType);
   cout << "FR Utilization Type      = " << str                        << endl;
   cout << "FR Utilization Weight    = " << FrameRateUtilizationWeight << endl;
   cout << "FR Utilization Constants = {";
   for(cardinal i = 0;i < FrameRateUtilizationConstants;i++) {
      cout << FrameRateUtilizationConstant[i];
      if((integer)i < (integer)FrameRateUtilizationMaxConstants - 1) {
         cout << ", ";
      }
   }
   for(cardinal j = FrameRateUtilizationConstants;j < FrameRateUtilizationMaxConstants;j++) {
      cout << "(empty)";
      if((integer)j < (integer)FrameRateUtilizationMaxConstants - 1) {
         cout << ", ";
      }
   }
   cout << "}" << endl;
   if((!utilizationOnly) || (Title[0] != 0x00)) {
      cout << "Title                    = " << Title     << endl;
   }
   if((!utilizationOnly) || (Copyright[0] != 0x00)) {
      cout << "Copyright                = " << Copyright << endl;
   }
   if((!utilizationOnly) || (Comment[0] != 0x00)) {
      cout << "Comment                  = " << Comment   << endl;
   }
   if((!utilizationOnly) || (URL[0] != 0x00)) {
      cout << "URL                      = " << URL       << endl;
   }
   const cardinal layers = 3 + (ExtLayers * 3);
   for(cardinal i = 0;i < layers;i++) {
      cout << "Layer #" << i << ":" << endl;
      if(utilizationOnly != true) {
         cout << "   Scalability                          = " << Layer[i].Scalability                      << endl;
         cout << "   Cost Factor                          = " << Layer[i].CostFactor                       << endl;
         cout << "   Byterate Empirical Envelope Pairs    = " << Layer[i].ByterateEmpiricalEnvelopePairs   << endl;
         cout << "   Frame Count Empirical Envelope Pairs = " << Layer[i].FrameCountEmpiricalEnvelopePairs << endl;
         snprintf((char*)&str,sizeof(str),"$%02x",Layer[i].LayerFlags);
         cout << "   Flags                                = " << str                                       << endl;
      }
      snprintf((char*)&str,sizeof(str),"$%04x",Layer[i].FrameSizeUtilizationType);
      cout << "   Frame Size Utilization Type          = " << str                                 << endl;
      cout << "   Frame Size Utilization Weight        = " << Layer[i].FrameSizeUtilizationWeight << endl;
      cout << "   Frame Size Utilization Constants     = {";
      for(cardinal j = 0;j < Layer[i].FrameSizeUtilizationConstants;j++) {
         cout << Layer[i].FrameSizeUtilizationConstant[j];
         if((integer)j < (integer)Layer[i].FrameSizeUtilizationMaxConstants - 1) {
            cout << ", ";
         }
      }
      for(cardinal j = Layer[i].FrameSizeUtilizationConstants;j < Layer[i].FrameSizeUtilizationMaxConstants;j++) {
         cout << "(empty)";
         if((integer)j < (integer)Layer[i].FrameSizeUtilizationMaxConstants - 1) {
            cout << ", ";
         }
      }
      cout << "}" << endl;
   }
};


// ###### Load configuration from file ######################################
bool TraceConfiguration::load(const char* fileName,
                              const bool  utilizationOnly)
{
   // ====== Open file ======================================================
   FILE* inputFD = fopen(fileName,"r");
   if(inputFD == NULL) {
      cerr << "ERROR: Unable to open input file <" << fileName << ">!" << endl;
      return(false);
   }

   // ====== Read configuration from file ===================================
   char str[256];
   char* result = fgets((char*)&str,256,inputFD);
   cardinal line = 0;
   while(!feof(inputFD)) {
      line++;
      const cardinal inputLength = strlen((char*)&str);
      if(inputLength > 1) {
         str[inputLength - 1] = 0x00;
         switch(str[0]) {
            // ====== Line is a comment =====================================
            case '#':
             break;

            // ====== Line is a setting =====================================
            default:
               const String input(str);
               String name;
               String value;
               if(input.scanSetting(name,value)) {
                  if(name == "FRAME RATE UTILIZATION") {
                     unsigned int type,max;
                     cardinal count =
                        sscanf(value.getData(),
                               "T$%x M%d W%lf %lf %lf %lf %lf %lf %lf %lf %lf",
                               &type, &max,
                               &FrameRateUtilizationWeight,
                               &FrameRateUtilizationConstant[0],
                               &FrameRateUtilizationConstant[1],
                               &FrameRateUtilizationConstant[2],
                               &FrameRateUtilizationConstant[3],
                               &FrameRateUtilizationConstant[4],
                               &FrameRateUtilizationConstant[5],
                               &FrameRateUtilizationConstant[6],
                               &FrameRateUtilizationConstant[7]);
                     if(count < 3) {
                        cerr << "ERROR: Bad frame rate utilization setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Frame Rate Utilization = T<type> M<max> W<weight> {c1} {c2} ..." << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     FrameRateUtilizationType         = (card16)type;
                     FrameRateUtilizationMaxConstants = (card16)max;
                     FrameRateUtilizationConstants    = (card16)(count - 3);
                     if(FrameRateUtilizationMaxConstants < FrameRateUtilizationConstants) {
                        FrameRateUtilizationMaxConstants = FrameRateUtilizationConstants;
                     }
                  }
                  else if(name == "LAYER FRAME SIZE UTILIZATION") {
                     unsigned int layer,type,max;
                     double       weight;
                     double a[8];
                     cardinal count =
                        sscanf(value.getData(),
                               "#%d T$%x M%d W%lf %lf %lf %lf %lf %lf %lf %lf %lf",
                               &layer, &type, &max, &weight,
                               &a[0], &a[1], &a[2], &a[3],
                               &a[4], &a[5], &a[6], &a[7]);
                     if(count < 4) {
                        cerr << "ERROR: Bad layer frame size utilization setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Layer Frame Size Utilization = #<layer> T<type> M<max> W<weight> {c1} {c2} ..." << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(layer >= MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     Layer[layer].FrameSizeUtilizationType         = (card16)type;
                     Layer[layer].FrameSizeUtilizationWeight       = (card16)weight;
                     Layer[layer].FrameSizeUtilizationMaxConstants = (card16)max;
                     Layer[layer].FrameSizeUtilizationConstants    = (card16)(count - 4);
                     for(cardinal j = 0;j < 8;j++) {
                        Layer[layer].FrameSizeUtilizationConstant[j] = a[j];
                     }
                     if(Layer[layer].FrameSizeUtilizationMaxConstants < Layer[layer].FrameSizeUtilizationConstants) {
                        Layer[layer].FrameSizeUtilizationMaxConstants = Layer[layer].FrameSizeUtilizationConstants;
                     }
                  }
                  else if(name == "RESOURCE/UTILIZATION POINTS") {
                     unsigned int points;
                     cardinal count = sscanf(value.getData(),
                                        "%d",
                                        &points);
                     if(count < 1) {
                        cerr << "ERROR: Bad resource/utilization point setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Resource/Utilization Points = <count>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(points > 65535) {
                        points = 65535;
                     }
                     else if(points < 4) {
                        points = 4;
                     }
                     RUPoints = (card16)points;
                  }
                  else if(name == "LAYER COST FACTOR") {
                     unsigned int layer;
                     double        cost;
                     if(sscanf(value.getData(),"#%d %lf",&layer,&cost) != 2) {
                        cerr << "ERROR: Bad layer cost factor setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Layer Cost Factor = #<layer> <cost>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(layer >= MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     Layer[layer].CostFactor = cost;
                  }
                  else if(name == "LAYER SCALABILITY") {
                     unsigned int layer;
                     double        scalability;
                     if(sscanf(value.getData(),"#%d %lf",&layer,&scalability) != 2) {
                        cerr << "ERROR: Bad layer scalability setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Layer Scalability = #<layer> <scalability>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(layer >= MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(scalability < 0.0) {
                        scalability = 0.0;
                     }
                     else if(scalability > 1.0) {
                        scalability = 1.0;
                     }
                     Layer[layer].Scalability = scalability;
                  }
                  else if(name == "LAYER FLAGS") {
                     unsigned int layer;
                     unsigned int flags;
                     if(sscanf(value.getData(),"#%d F$%x",&layer,&flags) != 2) {
                        cerr << "ERROR: Bad layer flags setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Layer Flags = #<layer> F$<flags>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(layer >= MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     Layer[layer].LayerFlags = (flags & 0xff);
                  }
                  else if(name == "REMAPPING COST") {
                     double cost;
                     if(sscanf(value.getData(),"%lf",&cost) != 1) {
                        cerr << "ERROR: Bad remapping cost setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Remapping Cost = <cost>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     RemappingCost = cost;
                  }
                  else if(name == "BYTERATE EMPIRICAL ENVELOPE PAIRS") {
                     unsigned int layer;
                     unsigned int pairs;
                     if(sscanf(value.getData(),"#%d %d",&layer,&pairs) != 2) {
                        cerr << "ERROR: Bad byterate empirical envelope pairs setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Byterate Empirical Envelope Pairs = #<layer> <pairs>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(layer >= MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(pairs < 2) {
                        pairs = 2;
                     }
                     if(pairs > 256) {
                        pairs = 256;
                     }
                     Layer[layer].ByterateEmpiricalEnvelopePairs = pairs;
                  }
                  else if(name == "FRAME COUNT PAIRS") {
                     unsigned int layer;
                     unsigned int pairs;
                     if(sscanf(value.getData(),"#%d %d",&layer,&pairs) != 2) {
                        cerr << "ERROR: Bad frame count pairs setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Frame Count Pairs = #<layer> <pairs>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(layer >= MaxLayers) {
                        cerr << "ERROR: Bad layer number, "
                                "line " << line << "!" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(pairs < 2) {
                        pairs = 2;
                     }
                     if(pairs > 256) {
                        pairs = 256;
                     }
                     Layer[layer].FrameCountEmpiricalEnvelopePairs = pairs;
                  }
                  else if(name == "MAX BUFFER DELAY") {
                     unsigned int delay;
                     if(sscanf(value.getData(),"%d",&delay) != 1) {
                        cerr << "ERROR: Bad max buffer delay setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Max Buffer Delay = <delay [s]>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(delay < 10000) {
                        delay = 10000;
                     }
                     if(delay > 30000000) {
                        delay = 30000000;
                     }
                     MaxBufferDelay = (cardinal)delay;
                  }
                  else if(name == "MIN INTERVAL LENGTH") {
                     if(sscanf(value.getData(),"%Ld",&MinIntervalLength) != 1) {
                        cerr << "ERROR: Bad minimum interval length setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Min Interval Length = <length [s]>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                  }
                  else if(name == "MAX INTERVAL LENGTH") {
                     if(sscanf(value.getData(),"%Ld",&MaxIntervalLength) != 1) {
                        cerr << "ERROR: Bad maximum interval length setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Max Interval Length = <length [s]>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                  }
                  else if(name == "BANDWIDTH THRESHOLD") {
                     if(sscanf(value.getData(),"%d",&BandwidthThreshold) != 1) {
                        cerr << "ERROR: Bad bandwidth threshold setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: bandwidth threshold = <value>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(BandwidthThreshold < 1) {
                        BandwidthThreshold = 1;
                     }
                  }
                  else if(name == "UTILIZATION THRESHOLD") {
                     if(sscanf(value.getData(),"%lf",&UtilizationThreshold) != 1) {
                        cerr << "ERROR: Bad utilization threshold setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: utilization threshold = <value>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(UtilizationThreshold < 0.0) {
                        UtilizationThreshold = 0.0;
                     }
                     else if(UtilizationThreshold > 1.0) {
                        UtilizationThreshold = 1.0;
                     }
                  }
                  else if(name == "EXTENSION LAYERS") {
                     unsigned int layers;
                     if(sscanf(value.getData(),"%d",&layers) != 1) {
                        cerr << "ERROR: Bad extension layers setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Extension Layers = <0|1|2>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(layers > 2) {
                        layers = 2;
                     }
                     ExtLayers = layers;
                  }
                  else if(name == "FAKE EXTENSION LAYER") {
                     unsigned int layer;
                     double       factor;
                     if(sscanf(value.getData(),"E%d %lf",&layer,&factor) != 2) {
                        cerr << "ERROR: Bad fake extension layer setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Fake Extension Layer = E<layer> <factor>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(factor < 0.0) {
                        factor = 0.0;
                     }
                     if(layer <= 1) {
                        FakeE1 = factor;
                     }
                     if(layer >= 2) {
                        FakeE2 = factor;
                     }
                  }
                  else if(name == "TITLE") {
                     snprintf((char*)&Title,sizeof(Title),"%s",
                              value.getData());
                  }
                  else if(name == "COPYRIGHT") {
                     snprintf((char*)&Copyright,sizeof(Copyright),"%s",
                              value.getData());
                  }
                  else if(name == "COMMENT") {
                     snprintf((char*)&Comment,sizeof(Comment),"%s",
                              value.getData());
                  }
                  else if(name == "URL") {
                     snprintf((char*)&URL,sizeof(URL),"%s",
                              value.getData());
                  }
                  else if(name == "INPUT") {
                     snprintf((char*)&InputName,sizeof(InputName),"%s",
                              value.getData());
                  }
                  else if(name == "FRAME PATTERN") {
                     snprintf((char*)&FramePattern,sizeof(FramePattern),"%s",
                              value.getData());
                  }
                  else if(name == "FRAME RATE") {
                     FrameRate = atof(value.getData());
                     if(FrameRate < 1.0) {
                        FrameRate = 1.0;
                     }
                     else if(FrameRate > 150.0) {
                        FrameRate = 150.0;
                     }
                  }
                  else {
                     cerr << "ERROR: Unknown option <"
                          << name << " = " << value << ">, "
                             "line " << line << "!" << endl;
                     fclose(inputFD);
                     return(false);
                  }
               }
             break;
         }
      }
      result = fgets((char*)&str,256,inputFD);
   }
   fclose(inputFD);
   return(true);
}


}
