#// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Encoder Implementation                                     ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "traceencoder.h"
#include "tracepacket.h"
#include "mediainfo.h"
#include "tools.h"
#include "traceqosdescription.h"


// Debug mode: Print frames to send
// #define DEBUG


namespace Coral {


// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// ALLES IN UNTERKLASSE TraceTraceEncoder!!!!!!!!!!!!!!!!!!!!


// ###### Get QoS description ###############################################
AbstractQoSDescription* TraceEncoder::getQoSDescription(
                           const cardinal pktHeaderSize,
                           const cardinal pktMaxSize,
                           const card64   offset)
{
  TraceQoSDescription* aqd = new TraceQoSDescription();
   if(aqd != NULL) {
      aqd->initTraceDescription(
         Source,
         Source->getPosition() + offset,
         Source->getMaxPosition(),
         Source->getFrameRate());
      aqd->updateDescription(pktHeaderSize + sizeof(TracePacket),pktMaxSize);
      aqd->setStreamPriority(StreamPriority);
      aqd->setWantedUtilization(WantedUtilization);
      aqd->setMinWantedBandwidth(MinWantedBandwidth);
      aqd->setMaxWantedBandwidth(MaxWantedBandwidth);
   }
   return(aqd);
}


// ###### Update quality from QoS description ###############################
void TraceEncoder::updateQuality(const AbstractQoSDescription* aqd)
{
   const cardinal layers = min(RTPConstants::RTPMaxQualityLayers,aqd->getLayers());

   // ====== Get manager's frame rate and scale factors =====================
   ManagerFrameRate = aqd->getFrameRate();

   bool bad = false;
   do {
      bad = false;
      Source->setFrameRate(ManagerFrameRate);
      for(cardinal i = 0;i < layers;i++) {
         const AbstractLayerDescription* ald = aqd->getLayer(i);
         const double alpha =
            ald->getFrameSizeScaleFactorForSize(
               ManagerFrameRate,
               AbstractLayerDescription::bandwidthToFrameSize(
                  ManagerFrameRate,ald->getBandwidth()));
         const double min = (double)ald->getMinFrameSize(ManagerFrameRate);
         const double max = (double)ald->getMaxFrameSize(ManagerFrameRate);

         ManagerScaleFactor[i]    = (min + alpha * (max - min)) / max;
         ManagerFrameSizeLimit[i] = (cardinal)floor(max * ManagerScaleFactor[i]);
      }
   } while(bad == true);


   // ====== Calculate utilization ==========================================
   ResourceUtilizationPoint rup;
   CurrentUtilization = aqd->getResources(rup);
   CurrentBandwidth   = rup.Bandwidth;
}


// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!??????????????????????????????????????
// ALLES IN UNTERKLASSE TraceTraceEncoder!!!!!!!!!!!!!!!!!!!!





// ###### Constructor #######################################################
TraceEncoder::TraceEncoder(TDTFMediaReader* traceReader)
{
   Source = traceReader;
   reset();
}


// ###### Destructor ########################################################
TraceEncoder::~TraceEncoder()
{
}


// ###### Activate encoder ##################################################
void TraceEncoder::activate()
{
   reset();
}


// ###### Deactivate encoder ################################################
void TraceEncoder::deactivate()
{
   reset();
}


// ###### Reset encoder #####################################################
void TraceEncoder::reset()
{
   ErrorCode = 0;
   SendError = 0;
   Scaled    = false;
   for(cardinal i = 0;i< RTPConstants::RTPMaxQualityLayers;i++) {
      FrameSize[i]             = 0;
      ManagerScaleFactor[i]    = 1.0;
      ManagerFrameSizeLimit[i] = 0;
   }
   CurrentUtilization = -HUGE_VAL;
   CurrentBandwidth    = 0;

   ManagerFrameRate   = 0.0;
   setWantedUtilization(1.0);
   setMaxWantedDelay(HUGE_VAL);
   setMinWantedBandwidth(0);
   setMaxWantedBandwidth((cardinal)-1);
   setStreamPriority(0);
}


// ###### Get encoder's type ID #############################################
const card16 TraceEncoder::getTypeID() const
{
   return(TracePacket::TraceTypeID);
}


// ###### Get encoder's type name ###########################################
const char* TraceEncoder::getTypeName() const
{
   return((const char*)&TracePacket::TraceTypeName);
}


// ###### Get utilization ###################################################
double TraceEncoder::getWantedUtilization() const
{
   return(WantedUtilization);
}


// ###### Set utilization ###################################################
void TraceEncoder::setWantedUtilization(const double utilization)
{
   if(utilization < 0.0) {
      WantedUtilization = 0.0;
   }
   else if(utilization > 1.0) {
      WantedUtilization = 1.0;
   }
   else {
      WantedUtilization = utilization;
   }
}


// ###### Get stream priority ###############################################
inline int8 TraceEncoder::getStreamPriority() const
{
   return(StreamPriority);
}


// ###### Set stream priority ###############################################
inline void TraceEncoder::setStreamPriority(const int8 priority)
{
   StreamPriority = priority;
}


// ###### Get maximum wanted delay ##########################################
double TraceEncoder::getMaxWantedDelay() const
{
   return(MaxWantedDelay);
}


// ###### Set maximum wanted delay ##########################################
void TraceEncoder::setMaxWantedDelay(const double delay)
{
   if(delay < 0.0) {
      MaxWantedDelay = 0.0;
   }
   else {
      MaxWantedDelay = delay;
   }
}


// ###### Get minimum bandwidth  ############################################
cardinal TraceEncoder::getMinWantedBandwidth() const
{
   return(MinWantedBandwidth);
}


// ###### Get maximum bandwidth  ############################################
cardinal TraceEncoder::getMaxWantedBandwidth() const
{
   return(MaxWantedBandwidth);
}


// ###### Set minimum bandwidth #############################################
void TraceEncoder::setMinWantedBandwidth(const cardinal bandwidth)
{
   MinWantedBandwidth = bandwidth;
}


// ###### Set maximum bandwidth #############################################
void TraceEncoder::setMaxWantedBandwidth(const cardinal bandwidth)
{
   MaxWantedBandwidth = bandwidth;
}


// ###### Check, if new interval is reached #################################
card64 TraceEncoder::checkInterval()
{
   const card64 newInterval = Source->checkInterval();
   return(newInterval);
}


// ###### Prepare frame for packaging #######################################
bool TraceEncoder::prepareNextFrame(const cardinal headerSize,
                                    const cardinal maxPacketSize,
                                    const cardinal flags)
{
   // ====== Initialize =====================================================
   for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
      FrameID[i]   = 0;
      FrameSize[i] = 0;
      Offset[i]    = 0;
   }

   FrameFrameRate = ManagerFrameRate;
   Source->setFrameRate(FrameFrameRate);


   // ====== Get next trace =================================================
   FramePosition    = Source->getPosition();
   FrameMaxPosition = Source->getMaxPosition();
   FrameUtilization = CurrentUtilization;
   FrameBandwidth   = CurrentBandwidth;
   Layers = Source->getNextBlock(RTPConstants::RTPMaxQualityLayers,
                                 (cardinal*)&FrameID,
                                 (cardinal*)FrameSize);
   if(Layers > 0) {
#ifdef DEBUG
      char str[256];
      const card64 seconds  = FramePosition / PositionStepsPerSecond;
      snprintf((char*)&str,sizeof(str),"%2Ld:%02Ld.%02Ld/%05d: ",
         (seconds / 60),
         (seconds % 60),
         (FramePosition % PositionStepsPerSecond) / (PositionStepsPerSecond / 100),
         Source->positionToFramePosition(FrameFrameRate,FramePosition));
      cout << str;

      for(cardinal j = 0;j < Layers;j++) {
         snprintf((char*)&str,sizeof(str),"%6d ",
            (cardinal)floor((double)FrameSize[j] * ManagerScaleFactor[j]));
         cout << str;
      }
      cout << "   (";
      for(cardinal j = 0;j < Layers;j++) {
         if(FrameSize[j] != 0) {
            snprintf((char*)&str,sizeof(str),"#%05d",FrameID[j]);
         }
         else {
            strcpy((char*)&str,"#-----");
         }
         cout << str;
         if(j < Layers - 1) {
            cout << " ";
         }
      }

      cout << ")   Rate=" << FrameFrameRate << endl;
#endif

      Scaled    = false;
      ErrorCode = ME_NoError;
      SendError = 0;
      return(true);
   }

   // ====== End of trace reached ===========================================
   else {
#ifdef DEBUG
      char str[256];
      const card64 seconds  = FramePosition / PositionStepsPerSecond;
      snprintf((char*)&str,sizeof(str),"%2Ld:%02Ld.%02Ld/%05d: ",
         (seconds / 60),
         (seconds % 60),
         (FramePosition % PositionStepsPerSecond) / (PositionStepsPerSecond / 100),
         Source->positionToFramePosition(FrameFrameRate,FramePosition));
      cerr << str << " --- End of File ---   Rate=" << FrameFrameRate << endl;
#endif


      // Source has an error -> send error message
      card8 error = (card8)Source->getErrorCode();
      if(error == 0) {
         if(Source->getPosition() >= Source->getMaxPosition()) {
            error = ME_EOF;
         }
      }
      if(error > 0) {
         if(SendError == 0) {
            ErrorCode    = error;
            Scaled       = true;
            FrameSize[0] = 4;
            Layers       = 1;
            SendError    = (cardinal)ceil(FrameFrameRate);
            return(true);
         }
         if(SendError > 0) {
            SendError--;
         }
      }
      return(false);
   }
   return(false);
}


// ###### Get next packet from current frame ################################
cardinal TraceEncoder::getNextPacket(EncoderPacket* encoderPacket)
{
   // ====== Check, if data has to be scaled ================================
   if(Scaled == false) {
      for(cardinal i = 0;i < Layers;i++) {
         // ====== Scale down frame size ====================================
         const cardinal old = FrameSize[i];
         FrameSize[i] = (cardinal)floor((double)FrameSize[i] * ManagerScaleFactor[i]);

         // ====== Verify result ============================================
         if((FrameSize[i] == 0) && (old > 0)) {
            cerr << "ERROR: TraceEncoder::getNextPacket() - Scale down to 0!" << endl;
            return(0);
         }
         else if(FrameSize[i] > ManagerFrameSizeLimit[i]) {
            cerr << "ERROR: TraceEncoder::getNextPacket() - Frame size exceeds limit by "
                 << FrameSize[i] - ManagerFrameSizeLimit[i]
                 << endl;
            return(0);
         }
      }
      Scaled = true;
   }


   // ====== Create packet ==================================================
   TracePacket* packet        = (TracePacket*)encoderPacket->Buffer;
   card8* data                = (card8*)&packet->Data[0];
   encoderPacket->Layer       = 0;
   encoderPacket->PayloadType = (card8)(TracePacket::TraceTypeID & 0x00ff);

   packet->FormatID     = TracePacket::TraceFormatID;
   packet->Position     = FramePosition;
   packet->MaxPosition  = FrameMaxPosition;
   packet->Layers       = (card8)Layers;
   packet->ErrorCode    = ErrorCode;
   packet->Flags        = TracePacket::TF_None;


   // ====== Get payload data ===============================================
   for(cardinal i = 0;i < Layers;i++) {
      if(FrameSize[i] > 0) {
         // ====== Calculate payload bytes ==================================
         cardinal bytes = min(encoderPacket->MaxLength - sizeof(TracePacket),
                              FrameSize[i]);

         // ====== Fill payload with dummy bytes ============================
         for(cardinal j = 0;j < bytes;j++) {
            data[j] = (card8)(j & 0xff);
         }
         FrameSize[i] -= bytes;

         // ====== Check, if payload is sufficient to store info block ======
         if(bytes >= sizeof(TracePacketData)) {
            TracePacketData *tpd = (TracePacketData*)data;
            tpd->FrameRate    = translateToBinary(FrameFrameRate);
            tpd->Utilization  = translateToBinary(FrameUtilization);
            tpd->Bandwidth    = translate32(FrameBandwidth);
            tpd->MinBandwidth = translate32(MinWantedBandwidth);
            tpd->MaxBandwidth = translate32(MaxWantedBandwidth);
            tpd->StreamPriority  = StreamPriority;
            tpd->SessionPriority = -128;
            Source->getMediaInfo(tpd->Information);
            tpd->Information.translate();
            packet->Flags = TracePacket::TF_Information;
         }

         // ====== Complete packet information ==============================
         packet->FrameID      = FrameID[i];
         packet->Layer        = (card8)i;
         encoderPacket->Layer = (card8)i;
         packet->translate();
         return(bytes + sizeof(TracePacket));
      }
   }
   return(0);
}


}
